/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	The implementation of the ARM processor.

****************************************************************************/

#include "gba.h"

u32 g_retiredCheck;
u32 g_retiredBP;
u32 g_branchBP;
u32 g_branchDepth;
u32 g_nextExpected;		// Next expected instruction.
static U8 s_scratchDecode[ARM::LINE_SIZE * 2];
u32 g_executeOffset;	// To avoid hitting a breakpoint on the current line.
static ARM::CacheRecord * s_LRU[ARM::LINES];
U16 ARM::s_cacheTable[MEM_TOTAL_SIZE];
U16 * ARM::s_cacheArray[0x10];
Sticky < Bool > g_skipStartup("Skip startup", FALSE);

/////////////////////////////////////////////////////////////////////////////
//  Convert the indicated address to a cache entry.
inline U16 & ARM::GetCacheFromAddress(u32 address) {
	u8 seg = address >> 24;
	static U16 s_dummy;

	if ((seg < 0x10) && s_cacheArray[seg]) {
		address = (address & g_memMask[seg]) >> 1;
		return s_cacheArray[seg][address];
	}
	return s_dummy;
}

/////////////////////////////////////////////////////////////////////////////
//  Build an ARM processor.
ARM::ARM(MemoryDispatcher & systemMemory) : m_sysMem(systemMemory) {
	Reset();
}

/////////////////////////////////////////////////////////////////////////////
//  This will mark the indicated cache element as being recently used.
inline void ARM::Touch(ARM::CacheRecord * c) {
	MType u;

	for (u=c->index+1; u<LINES; ++u) {
		s_LRU[u - 1] = s_LRU[u];
		s_LRU[u - 1]->index = u - 1;
	}
	s_LRU[LINES - 1] = c;
	c->index = LINES - 1;
}

/////////////////////////////////////////////////////////////////////////////
//  For the benefit of profiling.
ARM::CacheRecord * ARM::FindInCache(u32 address,
									u8 state,
									MType & m,
									Bool & sub) {
	MType low, high;
	CacheRecord * r;

	low = 0;
	high = m_maxCache;
	if (low != high) {
		do {
			m = (low + high) >> 1;
			r = m_cache[m];
	
			if ((address - r->addressLow) < (r->addressHigh - r->addressLow)) {
				sub = (r->flags != state);
				return r;
			}
			//  Our breakout condition.
			if (m == low) {
				if (address >= r->addressHigh) {
					++m;
				}
				break;
			}
			if (address < r->addressLow)
				high = m;
			else
				low = m;
		} while (1);
	}
	sub = FALSE;
	return NULL;
}

/////////////////////////////////////////////////////////////////////////////
//  This fills in the cache line.
void ARM::FillLine(CacheRecord * r,
				   DecodedInstruction & d,
				   List < ARM::Fixup > & fixups,
				   List < U8 * > & jtables,
				   u8 state,
				   u32 address,
				   u32 v) {
	U16 & s = GetCacheFromAddress(address);

	s = (r - m_cachePool) |
		(state << 11) |
		0x8000;

	// Let's fill in the indicated cache line with code.
	ClearForcedExceptions();
	r->flags = state;
	r->addressLow = address;
	d.dest = s_scratchDecode;
	d.base = 0;
	d.fixups = &fixups;
	d.jtables = &jtables;
	for (d.ip = 0; d.ip < MAX_ADDRESSES; ++d.ip) {
		r->ops[d.ip] = d.base;
		if (d.ip == MAX_ADDRESSES - 1) {
			ForceExceptions(EXC_CACHE_TERMINAL);
		}
		Decode(d, address);
		ClearForcedExceptions();
		if (d.base > LINE_SIZE) {
			d.base = r->ops[d.ip - 1];
			ForceExceptions(EXC_CACHE_TERMINAL);
			Decode(d, address - v);
			break;
		}
		address += v;
	}
	r->addressHigh = address;
	// Okay, we filled up a scratch buffer with the code we just generated.
	// Now, let's copy that into our codeBlock.
	memcpy(&r->codeBlock[0], s_scratchDecode, d.base);
}

/////////////////////////////////////////////////////////////////////////////
//  Eliminate any overlapping cache lines, or insert me in the cache.
void ARM::EliminateOverlap(u32 address,
						   ARM::CacheRecord * r,
						   u8 state,
						   MType m,			// Index of me,
						   MType n) {		// Index of my new successor.
	CacheRecord * s;

	if (n < m_maxCache) {
		s = m_cache[n];
		if (r->addressHigh >= s->addressLow) {
			s->flags = CACHE_AVAILABLE;
			r = m_cache[n];
			--m_maxCache;
			while (n < m_maxCache) {
				m_cache[n] = m_cache[n + 1];
				++n;
			}
			m_cache[n] = r;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Take care of all the fixups and jumps.
void ARM::DoFixupsAndJumps(List < ARM::Fixup > & fixups,
						   List < U8 * > & jtables,
						   ARM::CacheRecord * r,
						   ARM::DecodedInstruction & d,
						   u32 v) {
	MType m;
	u32 address;

	// Now, we do any "fixups".  This a nice optimization where we check to
	// see if a branch instruction does not exit a cache line; if that's
	// the case, we hardcode a jump instruction.  Sweet!
	for (m=0; m<fixups.Count(); ++m) {
		address = fixups[m].address;
		if (fixups[m].dispSpot >= r->addressHigh - r->addressLow) {
			continue;
		}
		if (((address - r->addressLow) < (r->addressHigh - r->addressLow))) {
			address -= r->addressLow;
			if (v == 4) {
				address >>= 2;
			} else {
				address >>= 1;
			}
			r->codeBlock[fixups[m].dispSpot] = 0xe9;
			*((U32 *)&(r->codeBlock[fixups[m].dispSpot + 1])) =
				r->ops[address] - fixups[m].dispSpot - 5;
			*(fixups[m].exceptionSpot1) &= ~EXC_EXTERNAL_JUMP;
			*(fixups[m].exceptionSpot2) &= ~EXC_EXTERNAL_JUMP;
		}
	}

	// Then, we assign proper numbers to all jump tables.
	for (m=0; m<jtables.Count(); ++m) {
		MType n;
		U32 * dd;

		if (jtables[m] - s_scratchDecode >= d.base) {
			continue;
		}
		dd = (U32 *)(jtables[m] -
					 s_scratchDecode + 
					 (U8 *)r->codeBlock.CastToVoidPointer());
		dd[-1] = (u32)dd;
		for (n=0; n<16; ++n) {
			dd[n] = (dd[n] -
					 (U32)s_scratchDecode +
					 (U32)r->codeBlock.CastToVoidPointer());
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  After we alter the cache, the cache lines have to be sorted.
void ARM::SortCacheLines(void) {
}

static u32 s_debug;

/////////////////////////////////////////////////////////////////////////////
//  This uses the new model to fill a cache line.
ARM::CacheRecord * ARM::DecodeInstruction(u32 address) {
	u8 state = GetRegister(CPSR) & 0x3f;
	u8 v = GetInstructionSize();
	DecodedInstruction d;
	List < Fixup > fixups;
	List < U8 * > jtables;
	CacheRecord * retval;
	MType m, n;					// Where we insert it.
	Bool sub;

	retval = FindInCache(address, state, m, sub);
	n = m + 1;
	if (retval) {
		if (!sub) {
			return retval;
		}
	} else {
		if (m_maxCache != LINES) {
			retval = m_cache[m_maxCache];
			if (m_maxCache) {
				for (n=m_maxCache; n!=m; --n) {
					m_cache[n] = m_cache[n - 1];
				}
				m_cache[m] = retval;
				n = m + 1;
			}
			sub = TRUE;
			++m_maxCache;
		} else {
			retval = s_LRU[0];
			for (m=0; m<LINES; ++m) {
				if (m_cache[m] == retval)
					break;
			}
			if ((m != LINES - 1) &&
				(address > m_cache[m + 1]->addressHigh)) {
				for (n=m+1;
					 (n<LINES) && (address>m_cache[n]->addressHigh);
					 ++n) {
					m_cache[n - 1] = m_cache[n];
				}
				m = n - 1;
				m_cache[m] = retval;
			} else if (m &&
					   (address < m_cache[m - 1]->addressLow)) {
				for (n=m-1; address < m_cache[n]->addressLow; --n) {
					m_cache[n + 1] = m_cache[n];
					if (!n) {
						--n;
						break;
					}
				}
				m = n + 1;
				m_cache[m] = retval;
			}
			n = m + 1;
			sub = TRUE;
		}
	}
	FillLine(retval, d, fixups, jtables, state, address, v);

	// If our cache line overlaps another one, we free up the other.
	address -= v;
	EliminateOverlap(address, retval, state, m, n);
	DoFixupsAndJumps(fixups, jtables, retval, d, v);
	SortCacheLines();
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Advance the ARM processor by one atomic operation at a time (until an
// exception is encountered).
void ARM::Tick(void) {
	u32 pc = GetCurrentIP();
	u32 state = GetRegister(CPSR);
	CacheRecord * cr;
	Bool v;
	u32 retiredCheck;

	if (g_ticksTilNextVideo < 0x7ff) {
		v = TRUE;
		g_clocksTilNext = g_ticksTilNextVideo;
	} else {
		v = FALSE;
		g_clocksTilNext = 0x7ff;
	}

	g_gbaTicks += g_clocksTilNext;
GetCode:
	if (m_pending) {
		u32 pending = m_pending;
		static U32 s_vecs[] = {
			0x00000000,
			0x00000010,
			0x0000001c,
			0x00000018,
			0x0000000c,
			0x00000008,
			0x00000004,
		};
		static u8 s_modes[] = {
			0xdf,			// I & F set.
			0x17,			// HEREHEREHERE don't know about I & F.
			0x51,			// F set.
			0x92,			// I set.
			0x17,			// HEREHEREHERE don't know about I & F.
			0x13,			// I & F not set (SWI).
			0x1b,			// HEREHEREHERE don't know about I & F.
		};
		u8 intr, p;
		u32 cpsr = GetRegister(CPSR);

		if (cpsr & 0x40) {
			pending &= ~INT_FIQ;
		}
		if (cpsr & 0x80) {
			pending &= ~INT_IRQ;
		}
		if (pending) {
			for (intr=1, p=0; intr<=INT_UDEF; intr += intr, ++p) {
				static u8 s_offsets[] = {
					0,
					8,
					4,
					4,
					0,
					0,
					0
				};
				if (pending & intr) {
					SetRegisterHistory(CPSR,	// Change mode and state.
									   (cpsr & ~0x3f) | s_modes[p]);
					SetRegisterHistory(0xe,		// Saves next instr.
									   pc + s_offsets[p]);
					SetRegisterHistory(SPSR, cpsr);
					SetRegisterHistory(0xf, s_vecs[p]);
					m_pending &= ~intr;
					pc = GetCurrentIP();
					state = GetRegister(CPSR);
					break;
				}
			}
		}
	}
	retiredCheck = pc;
	cr = DecodeInstruction(pc);
	CHECK(cr);
	Touch(cr);
	if (GetInstructionSize() == 2) {
		pc = (pc - cr->addressLow) >> 1;
	} else {
		pc = (pc - cr->addressLow) >> 2;
	}
	CHECK(pc < MAX_ADDRESSES);
 	g_exceptions = ExecuteOp(&cr->codeBlock[cr->ops[pc] + g_executeOffset]);
	g_executeOffset = 0;
	if (!(g_exceptions & ~(EXC_CACHE_TERMINAL |
						   EXC_EXTERNAL_JUMP |
						   EXC_STATE_CHANGE |
						   EXC_MODE_CHANGE))) {
		pc = GetCurrentIP();
		if (g_exceptions & EXC_STATE_CHANGE) {
			if (pc & 1) {
				SetRegister(CPSR, GetRegister(CPSR) | 0x20);
				pc &= -2;
			} else {
				SetRegister(CPSR, GetRegister(CPSR) & ~0x20);
				pc &= -4;
			}
		}
		state = GetRegister(CPSR);
		goto GetCode;
	}
	g_gbaTicks -= g_clocksTilNext;
	BreakpointCheck(GetCurrentIP(),
					g_badAddress,
					g_exceptions);
	if (g_exceptions & EXC_STATE_CHANGE) {
		pc = GetCurrentIP();
		if (pc & 1) {
			SetRegister(CPSR, GetRegister(CPSR) | 0x20);
			pc &= -2;
		} else {
			SetRegister(CPSR, GetRegister(CPSR) & ~0x20);
			pc &= -4;
		}
		SetRegister(0xf, pc);
	}
	if (g_exceptions & EXC_USER_MESSAGE) {
		if (RegistrationGetLevel() > 0) {
			ProcessUserMessage(g_exceptions);
		}
	}
	if (g_exceptions & EXC_SWI) {
		Interrupt(INT_SWI);
	}
	if (g_exceptions & EXC_TIMER) {
		IRQ(g_pendingTimerInterrupts);
		g_pendingTimerInterrupts = 0;
	}

	//  HEREHEREHERE we will have special processing if the video is off.
	if (v) {
		g_ticksTilNextVideo = g_clocksTilNext;
		if (g_exceptions & EXC_CYCLE_BREAK) {
			g_gba.GetVideo().Tick();
		}
	}
	if (g_exceptions & EXC_RETIRED) {
		if (retiredCheck != g_retiredCheck) {
			g_exceptions &= ~EXC_RETIRED;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This is where we tick the processor backwards.  This will have different
// meaning for level 0, 1, 2, or 3.
void ARM::TickBackward(void) {
	u32 clocks;
	MType m, n;
	U32 avps[32][2];
	Bool v, updateR15;
	u32 oldR15, oldR14;

	if (g_ticksTilNextVideo < 0x7ff) {
		g_clocksTilNext = g_ticksTilNextVideo;
		v = TRUE;
	} else {
		g_clocksTilNext = 0x7ff;
		v = FALSE;
	}
	oldR15 = GetCurrentIP();
	oldR14 = GetRegister(0xe);
NextOp:
	if (HistoryExists()) {
		//  This does one atomic iteration.
		m = 1;
		g_exceptions = EXC_ATOMIC;
		updateR15 = TRUE;
		HistoryGetPrior(avps[0][0], avps[0][1]);
		if (avps[0][0] == HISTORY_TERMINAL) {
			HistoryGetNext(avps[0][0], avps[0][1]);
			g_exceptions |= EXC_HISTORY_END;
			return;
		}
		CHECK(avps[0][0] == HISTORY_TIMESTAMP);
		do {
			if (!HistoryGetPrior(avps[m][0], avps[m][1])) {
				g_exceptions |= EXC_HISTORY_END;
				return;
			}
			if ((avps[m][0] == HISTORY_TIMESTAMP) ||
				(avps[m][0] == HISTORY_TERMINAL)) {
				HistoryGetNext(avps[m][0], avps[m][1]);
				break;
			}
			++m;
		} while (1);
		clocks = avps[0][1];
		for (n=1; n<m; ++n) {
			switch(avps[n][0]) {
				//  HEREHEREHERE I'm assuming that no opcode will read from
				// or write to more than one memory location at a time.
				case HISTORY_READ:
					g_exceptions |= EXC_READ_BREAKPOINT;
					BreakpointCheck(0, avps[n][1], g_exceptions);
					break;
				case HISTORY_WRITE:
					g_exceptions |= EXC_WRITE_BREAKPOINT;
					BreakpointCheck(0, avps[n][1], g_exceptions);
					break;
				default:
					if (GetRegAddress(0xf) == avps[n][0]) {
						updateR15 = FALSE;	// No banking issues.
					}
					*((U32 *)avps[n][0]) = avps[n][1];
					break;
			}
		}
		if (updateR15) {
			SetRegister(0xf, GetRegister(0xf) - GetInstructionSize());
		}

		// Check for code breakpoints.
		g_exceptions |= EXC_EXEC_BREAKPOINT;
		BreakpointCheck(GetCurrentIP(), 0, g_exceptions);
		g_clocksTilNext -= clocks >> 20;
		if (g_clocksTilNext & 0x80000000) {
			g_exceptions |= EXC_CYCLE_BREAK;
		}
		if (!(g_exceptions &= (g_exceptionMask | EXC_CYCLE_BREAK))) {
			goto NextOp;
		}
		if ((g_exceptions & EXC_CYCLE_BREAK) && v) {
			g_gba.GetVideo().TickBackwards();
		}
	} else {
		g_exceptions |= EXC_HISTORY_END;
		return;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Reset the ARM processor to its reboot state.
void ARM::Reset(void) {
	U16 * u;

	// This lets us switch between the real BIOS and Marat's.
	m_registers.Reset();
	if (g_gba.GetMemory().GetD(0) && !(*g_skipStartup)) {
		m_registers.Set(15, 0, 0x0);
		m_registers.Set(CPSR, 0, 0xd3);
	} else {
		m_registers.Set(13, 2, 0x3007fa0);
		m_registers.Set(13, 3, 0x3007fe0);
		m_registers.Set(13, 0, 0x3007fe0);
		m_registers.Set(15, 0, 0x8000000);
	}

	// Set up the cache table.
	u = s_cacheTable;
	s_cacheArray[0] = u;
	u += MEM_BIOS_SIZE / 2;
	s_cacheArray[2] = u;
	u += MEM_EXTERNAL_RAM_SIZE / 2;
	s_cacheArray[3] = u;
	u += MEM_INTERNAL_RAM_SIZE / 2;
	s_cacheArray[4] = u;
	u += MEM_MMAP_RAM_SIZE / 2;
	s_cacheArray[5] = u;
	u += MEM_PALETTE_RAM_SIZE / 2;
	s_cacheArray[6] = u;
	u += MEM_VRAM_RAM_SIZE / 2;
	s_cacheArray[7] = u;
	u += MEM_OAM_SIZE / 2;
	s_cacheArray[8] = s_cacheArray[0xa] = s_cacheArray[0xc] = u;
	u += MEM_CODE_ROM_SIZE / 2;
	s_cacheArray[9] = s_cacheArray[0xb] = s_cacheArray[0xd] = u;
	u += MEM_EEPROM_SIZE / 2;
	s_cacheArray[0xe] = u;

	g_instructionStep = 4;

	// Reset the timers.  Only call this if memory has been constructed
	// yet, as it relies on the existence of g_readMap[4] (and nothing
	// else).
	if (g_readMap[4]) {
		MakeTimerUpdateCode();
	}

	InvalidateCodeCache();
}

/////////////////////////////////////////////////////////////////////////////
//  This will invalidate the code cache.
void ARM::InvalidateCodeCache(void) {
	MType m;

	for (m=0; m<LINES; ++m) {
		CacheRecord * c;

		c = m_cache[m] = s_LRU[m] = m_cachePool + m;
		c->index = m;
		c->flags = CACHE_AVAILABLE;
	}
	m_maxCache = 0;
}

/////////////////////////////////////////////////////////////////////////////
//  Respond to an interrupt.
void ARM::Interrupt(u8 intr) {
	m_pending |= intr;
}

/////////////////////////////////////////////////////////////////////////////
//  Generate an IRQ, if our IE bits are properly set.
void ARM::IRQ(u16 irq) {
	u16 ie = m_sysMem.GetW(0x4000200);
	Bool ime = m_sysMem.GetB(0x4000208);

	if (ime) {
		irq &= ie;
		if (irq) {
			m_irqs |= irq;
			m_sysMem.SetW(0x4000202, m_irqs);
			Interrupt(INT_IRQ);
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  The user is causing some kind of direct interaction with the debugger.
// This may be a message, a user breakpoint, or some other service the
// debugger may provide.
void ARM::ProcessUserMessage(u32 & exceptions) {
	u32 mType = GetRegister(0);
	u32 mParam = GetRegister(1);
	FString f;
	MType m;

	switch(mType) {
		case USER_BREAKPOINT:
			exceptions |= EXC_USER_BREAKPOINT;
			break;
		case USER_ASCII_TEXT:
			for (m=0; m<256; ++m) {
				wchar_t w = g_gba.GetMemory().GetB(mParam);

				if (!w) {
					break;
				}
				f += w;
				++mParam;
			}
			g_codeWindow->AddUserMessage(f);
			break;
		case USER_WIDE_TEXT:
			for (m=0; m<256; ++m) {
				wchar_t w = g_gba.GetMemory().GetW(mParam);

				if (!w) {
					break;
				}
				f += w;
				mParam += 2;
			}
			g_codeWindow->AddUserMessage(f);
			break;
		case USER_NUMBER:
			ConvertFromHex(f, g_gba.GetCPU().GetRegister(1));
			g_codeWindow->AddUserMessage(f);
			break;
		case USER_FIXED:
			ConvertFromFloat(f, (S32)g_gba.GetCPU().GetRegister(1) / 65536.0, 8);
			g_codeWindow->AddUserMessage(f);
			break;
	}
}

