/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	This will implement the ARM processor found in the GBA.

****************************************************************************/

#ifndef ARM_H
	#define ARM_H

	#define CYCLE_SHIFT			20	// Where we put the CPU cycles in ebp.

	class ARM {
		public:
			// These are the IRQs generated on GBA.
			enum {
				IRQ_VBLANK			= 0x0001,
				IRQ_HBLANK			= 0x0002,
				IRQ_VCOUNT			= 0x0004,
				IRQ_TIMER0			= 0x0008,
				IRQ_TIMER1			= 0x0010,
				IRQ_TIMER2			= 0x0020,
				IRQ_TIMER3			= 0x0040,
				IRQ_DMA0			= 0x0100,
				IRQ_DMA1			= 0x0200,
				IRQ_DMA2			= 0x0400,
				IRQ_DMA3			= 0x0800,
			};

			// These are the kinds of interrupts that can be generated.
			enum {
				INT_RESET			= 0x01,
				INT_ABORT_DATA		= 0x02,
				INT_FIQ				= 0x04,
				INT_IRQ				= 0x08,
				INT_ABORT_PREFETCH	= 0x10,
				INT_SWI				= 0x20,
				INT_UDEF			= 0x40,
			};

			// Okay, we will maintain an ordered list of cache lines, so we
			// can quickly find a cache hit.
			// Here's the rule for filling a cache line:  if we need to execute
			// an address that's not in-cache, we allocate a cache record that
			// starts at that address.  We fill in the cache record until it
			// simply will not receive any more code, then we have the
			// code flag a cache terminal exception.  Next, we insert the
			// cache record in the cache.  If any other cache record exists
			// with an overlapping address, we eliminate the one with the
			// larger starting address.
			// Note:  when we're filling up the cache, we CAN fold jumps to
			// absolute addresses within the cache!!!
			enum {
				LINE_SIZE			= 8192,
				LINES				= 512,
				MAX_ADDRESSES		= 64,
				CACHE_AVAILABLE		= 0x80000000,
			};

			// This structure enables us to hard-code jump instructions that
			// occur within the same cache-line.
			class Fixup {
				public:
					u16 dispSpot;
					u32 address;
					U32 * exceptionSpot1;
					U32 * exceptionSpot2;

					Bool operator ==(const Fixup & a) const;
			};

			struct CacheRecord {
				MType index;					// For LRU.
				U32 flags;
				U32 addressLow;
				U32 addressHigh;
				U16 ops[MAX_ADDRESSES];
				Vec < U8, LINE_SIZE > codeBlock;
			};
			struct DecodedInstruction {
				u16 base;
				U8 * dest;
				u32 ip;
				List < Fixup > * fixups;
				List < U8 * > * jtables;
			};
			typedef u32 (*AtomicOp)(void);

			enum {
				STATE_ARM			= 0,
				STATE_THUMB,
				STATE_Z80
			};
			ARM(MemoryDispatcher & systemMemory);
			void Tick(void);
			void TickBackward(void);
			void Reset(void);

			// Effect an interrupt.
			void Interrupt(u8 intr);
			void IRQ(u16 irq);

			void Decode(FString & f, u32 address);
			void Decode(DecodedInstruction & d, u32 address);

			void Disassemble(FString & disassembly, u32 address);
			u8 GetInstructionSize(void);
			u32 GetMode(void) const;

			U16 & GetCacheFromAddress(u32 address);
			CacheRecord * FindInCache(u32 address, u8 state, MType & m, Bool & sub);
			void FillLine(CacheRecord * r,
						  DecodedInstruction & d,
						  List < Fixup > & fixups,
						  List < U8 * > & jtables,
						  u8 state,
						  u32 address,
						  u32 v);
			void EliminateOverlap(u32 address, CacheRecord * r, u8 state, MType m, MType n);
			void DoFixupsAndJumps(List < Fixup > & fixups,
								  List < U8 * > & jtables,
								  CacheRecord * r,
								  DecodedInstruction & d,
								  u32 v);
			void SortCacheLines(void);

			CacheRecord * DecodeInstruction(u32 address);
			void Invalidate(u32 address);
			void InvalidateCodeCache(void);

			u8 GetState(void);
			u32 GetCurrentIP(void) const;
			u32 GetRegister(u8 index) const;
			void SetRegister(u8 index, u32 value);

			// Set the indicated value, while saving the old value in
			// the history buffer.
			void SetRegisterHistory(u8 index, u32 value);

			// Deal with messages that the user is sending to the app.
			void ProcessUserMessage(u32 & exceptions);

			// Has the most recent instruction retired.
			Bool Retired(void) const;

			// Jump to the indicated address.
			void Goto(u32 address);

			U32 GetRegAddress(u8 index);
			U32 GetRegAddress(u8 index, u8 mode);

			u16 & GetIRQs(void);
			static void MakeTimerUpdateCode(void);
		protected:
		private:
			// These helpers are for messing with the assembly code.
			u32 GetCycles(u32 address, u32 n, u32 i, u32 s);
			static void InsertDispToRMImm(U8 * dest, u32 value);
			static void InsertDispToRegRM(U8 * dest, u32 value);
			static void InsertShiftToReg(U8 * dest, u8 shift);
			static void InsertImmediateToRM(U8 * dest, u32 value);
			static void InsertImmediateToRMNoDisp(U8 * dest, u32 value);
			static void InsertImmediateToReg(U8 * dest, u32 value);
			static void InsertImmediateToAcc(U8 * dest, u32 value);
			static void InsertDispToAcc(U8 * dest, u32 value);
			static void InsertImmediateToShift(U8 * dest, u32 value);
			static void InsertDispToSIBRegRM(U8 * dest, u32 value);
			static void InsertTimerUpdateCode(DecodedInstruction & dest);
			static void Insert(DecodedInstruction & dest,
							   U8 * start,
							   U8 * end);
			U16 * InsertConditionCodeCheck(u32 address,
										   u8 flags,
										   DecodedInstruction & dest,
										   u32 & leftOff);
			void InsertConditionCodePost(DecodedInstruction & dest,
										 U16 * d,
										 u32 leftOff,
										 u32 exceptions,
										 u32 address,
										 u32 cycles,
										 Bool isMultiply=FALSE);
			void InsertExceptionCheck(DecodedInstruction & dest,
									  u32 exceptions,
									  u32 address,
									  u32 cycles);
			U8 * InsertJumpTable(DecodedInstruction & dest, u32 reg);
			static void SaveHistory(DecodedInstruction & dest, u32 address);
			static void MemoryTransfer(DecodedInstruction & dest,
									   u32 d,
									   u32 s);
			static void MemoryTransferNoHistory(ARM::DecodedInstruction & dest,
												u32 d,
												u32 s);
			static void MemoryStore(ARM::DecodedInstruction & dest,
									u32 d,
									u32 s);
			void EffectDataTransfer(u32 address,
									u32 offset,
									DecodedInstruction & dest,
									u32 & exceptions,
									Bool imm,
									Bool pre,
									Bool up,
									Bool wb,
									Bool ld,
									Bool user,
									u8 rn,
									u8 rd,
									u8 rm,
									u8 tType,
									u8 sType,
									u8 s,
									Bool sequential=FALSE);
			static u32 GetImmediateFromOp2(u32 u);
			static Bool Contains(DecodedInstruction & dest,
								 u32 address);
			static U16 * GetJumpDisplacementAddress(U8 * u);
			static U32 * GetDisplacementToRMImm(U8 * u);
			static U32 * GetImmediateToAcc(U8 * u);

			void Touch(CacheRecord * c);

			//  These govern exceptions that we're making explicitly necessary.
			// This will facilitate code breakpoints and cache terminals.
			static void ClearForcedExceptions(void);
			static void ForceExceptions(u32 u);
			static u32 GetForcedExceptions(void);

			//  I will not deface my code with the kind of profanity that
			// Microsoft deserves, but man, do they deserve it.
			friend class HackClass < FString >;
			friend class HackClass < DecodedInstruction >;

			void DecDataProcessing(u32 address, u32 code, FString & dest);
			void DecMultiply(u32 address, u32 code, FString & dest);
			void DecMultiplyLong(u32 address, u32 code, FString & dest);
			void DecSingleDataSwap(u32 address, u32 code, FString & dest);
			void DecBX(u32 address, u32 code, FString & dest);
			void DecHalfwordReg(u32 address, u32 code, FString & dest);
			void DecHalfwordImm(u32 address, u32 code, FString & dest);
			void DecSingleDataTransfer(u32 address, u32 code, FString & dest);
			void DecUndef(u32 address, u32 code, FString & dest);
			void DecBlockDataTransfer(u32 address, u32 code, FString & dest);
			void DecBranch(u32 address, u32 code, FString & dest);
			void DecCoprocDataTransfer(u32 address, u32 code, FString & dest);
			void DecCoprocDataOp(u32 address, u32 code, FString & dest);
			void DecCoprocRegTransfer(u32 address, u32 code, FString & dest);
			void DecSWI(u32 address, u32 code, FString & dest);
			void DecUndefined(u32 address, u32 code, FString & dest);
			void DecTHMSR(u32 address, u16 code, FString & dest);
			void DecTHAddSub(u32 address, u16 code, FString & dest);
			void DecTHOpImm(u32 address, u16 code, FString & dest);
			void DecTHALUOps(u32 address, u16 code, FString & dest);
			void DecTHHiRegOps(u32 address, u16 code, FString & dest);
			void DecTHPCRelLd(u32 address, u16 code, FString & dest);
			void DecTHLdStRegOff(u32 address, u16 code, FString & dest);
			void DecTHLdStSexBHW(u32 address, u16 code, FString & dest);
			void DecTHLdStImm(u32 address, u16 code, FString & dest);
			void DecTHLdStHw(u32 address, u16 code, FString & dest);
			void DecTHSPRelLdSt(u32 address, u16 code, FString & dest);
			void DecTHLEA(u32 address, u16 code, FString & dest);
			void DecTHAddOffSP(u32 address, u16 code, FString & dest);
			void DecTHPushPop(u32 address, u16 code, FString & dest);
			void DecTHMLS(u32 address, u16 code, FString & dest);
			void DecTHCondBr(u32 address, u16 code, FString & dest);
			void DecTHSWI(u32 address, u16 code, FString & dest);
			void DecTHUncondBranch(u32 address, u16 code, FString & dest);
			void DecTHLongBrLink(u32 address, u16 code, FString & dest);

			void DecDataProcessing(u32 address, u32 code, DecodedInstruction & dest);
			void DecMultiply(u32 address, u32 code, DecodedInstruction & dest);
			void DecMultiplyLong(u32 address, u32 code, DecodedInstruction & dest);
			void DecSingleDataSwap(u32 address, u32 code, DecodedInstruction & dest);
			void DecBX(u32 address, u32 code, DecodedInstruction & dest);
			void DecHalfwordReg(u32 address, u32 code, DecodedInstruction & dest);
			void DecHalfwordImm(u32 address, u32 code, DecodedInstruction & dest);
			void DecSingleDataTransfer(u32 address, u32 code, DecodedInstruction & dest);
			void DecUndef(u32 address, u32 code, DecodedInstruction & dest);
			void DecBlockDataTransfer(u32 address, u32 code, DecodedInstruction & dest);
			void DecBranch(u32 address, u32 code, DecodedInstruction & dest);
			void DecCoprocDataTransfer(u32 address, u32 code, DecodedInstruction & dest);
			void DecCoprocDataOp(u32 address, u32 code, DecodedInstruction & dest);
			void DecCoprocRegTransfer(u32 address, u32 code, DecodedInstruction & dest);
			void DecSWI(u32 address, u32 code, DecodedInstruction & dest);
			void DecUndefined(u32 address, u32 code, DecodedInstruction & dest);
			void DecTHMSR(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHAddSub(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHOpImm(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHALUOps(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHHiRegOps(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHPCRelLd(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHLdStRegOff(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHLdStSexBHW(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHLdStImm(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHLdStHw(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHSPRelLdSt(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHLEA(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHAddOffSP(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHPushPop(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHMLS(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHCondBr(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHSWI(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHUncondBranch(u32 address, u16 code, DecodedInstruction & dest);
			void DecTHLongBrLink(u32 address, u16 code, DecodedInstruction & dest);

			// To hold us through until everything's implemented.
			void DecDefaultStub(u32 address,
								u32 code,
								ARM::DecodedInstruction & dest,
								u32 cycles = 1,
								u32 extraExceptions = 0);
			u8 m_pending;					// Interrupts.
			u16 m_irqs;						// IRQs that have fired.
			RegisterSet m_registers;
			MemoryDispatcher & m_sysMem;
			// This saves us having to sort.
			static U16 s_cacheTable[MEM_TOTAL_SIZE];
			static U16 * s_cacheArray[0x10];
			CacheRecord m_cachePool[LINES];
			CacheRecord * m_cache[LINES];	// Above, sorted.
			MType m_maxCache;				// How many lines are allocated.
	};

	extern "C" {
		extern U8 __cdecl ARMEX_SetCycles[];
		extern U8 __cdecl ARMEX_SetCycles_End[];
		extern U8 __cdecl ARMEX_Nop[];
		extern U8 __cdecl ARMEX_Nop_ExceptionsRaised[];
		extern U8 __cdecl ARMEX_Nop_ExceptionsRaisedCopy[];
		extern U8 __cdecl ARMEX_Nop_StoreR15[];
		extern U8 __cdecl ARMEX_Nop_End[];
		extern U8 __cdecl ARMEX_CheckCondition[];
		extern U8 __cdecl ARMEX_CheckCondition_SpecifyNotTaken[];
		extern U8 __cdecl ARMEX_CheckCondition_CPSR[];
		extern U8 __cdecl ARMEX_CheckCondition_Condition[];
		extern U8 __cdecl ARMEX_CheckCondition_Jump[];
		extern U8 __cdecl ARMEX_CheckCondition_End[];
		extern U8 __cdecl ARMEX_ABranchWithin[];
		extern U8 __cdecl ARMEX_ABranchWithin_End[];
		extern U8 __cdecl ARMEX_ABranchWithout[];
		extern U8 __cdecl ARMEX_ABranchWithout_End[];
		extern U8 __cdecl ARMEX_MemTransfer[];
		extern U8 __cdecl ARMEX_MemTransfer_Source[];
		extern U8 __cdecl ARMEX_MemTransfer_Dest[];
		extern U8 __cdecl ARMEX_MemTransfer_Copy[];
		extern U8 __cdecl ARMEX_MemTransfer_DoTransfer[];
		extern U8 __cdecl ARMEX_MemTransfer_End[];
		extern U8 __cdecl ARMEX_MemStore[];
		extern U8 __cdecl ARMEX_MemStore_Dest[];
		extern U8 __cdecl ARMEX_MemStore_DestCopy[];
		extern U8 __cdecl ARMEX_MemStore_DoTransfer[];
		extern U8 __cdecl ARMEX_MemStore_Source[];
		extern U8 __cdecl ARMEX_MemStore_End[];
		extern U8 __cdecl ARMEX_Operand2Immediate[];
		extern U8 __cdecl ARMEX_Operand2Immediate_End[];
		extern U8 __cdecl ARMEX_Operand2Register[];
		extern U8 __cdecl ARMEX_Operand2Register_End[];
		extern U8 __cdecl ARMEX_Operand2LSL[];
		extern U8 __cdecl ARMEX_Operand2LSL_End[];
		extern U8 __cdecl ARMEX_Operand2LSR[];
		extern U8 __cdecl ARMEX_Operand2LSR_End[];
		extern U8 __cdecl ARMEX_Operand2ASR[];
		extern U8 __cdecl ARMEX_Operand2ASR_End[];
		extern U8 __cdecl ARMEX_Operand2ROR[];
		extern U8 __cdecl ARMEX_Operand2ROR_End[];
		extern U8 __cdecl ARMEX_SaveCarry[];
		extern U8 __cdecl ARMEX_SaveCarry_End[];
		extern U8 __cdecl ARMEX_PreserveCarryFlag[];
		extern U8 __cdecl ARMEX_PreserveCarryFlag_End[];
		extern U8 __cdecl ARMEX_LSR0[];
		extern U8 __cdecl ARMEX_LSR0_End[];
		extern U8 __cdecl ARMEX_ASR0[];
		extern U8 __cdecl ARMEX_ASR0_End[];
		extern U8 __cdecl ARMEX_RCR[];
		extern U8 __cdecl ARMEX_RCR_End[];
		extern U8 __cdecl ARMEX_Operand2LSLRegisterCF[];
		extern U8 __cdecl ARMEX_Operand2LSLRegisterCF_GetCPSR[];
		extern U8 __cdecl ARMEX_Operand2LSLRegisterCF_SaveCarry[];
		extern U8 __cdecl ARMEX_Operand2LSLRegisterCF_BigShift[];
		extern U8 __cdecl ARMEX_Operand2LSLRegisterCF_Shift32[];
		extern U8 __cdecl ARMEX_Operand2LSLRegisterCF_Done[];
		extern U8 __cdecl ARMEX_Operand2LSLRegisterCF_End[];
		extern U8 __cdecl ARMEX_Operand2LSLRegisterNC[];
		extern U8 __cdecl ARMEX_Operand2LSLRegisterNC_SaveCarry[];
		extern U8 __cdecl ARMEX_Operand2LSLRegisterNC_End[];
		extern U8 __cdecl ARMEX_Operand2LSRRegisterCF[];
		extern U8 __cdecl ARMEX_Operand2LSRRegisterCF_GetCPSR[];
		extern U8 __cdecl ARMEX_Operand2LSRRegisterCF_SaveCarry[];
		extern U8 __cdecl ARMEX_Operand2LSRRegisterCF_BigShift[];
		extern U8 __cdecl ARMEX_Operand2LSRRegisterCF_Shift32[];
		extern U8 __cdecl ARMEX_Operand2LSRRegisterCF_Done[];
		extern U8 __cdecl ARMEX_Operand2LSRRegisterCF_End[];
		extern U8 __cdecl ARMEX_Operand2LSRRegisterNC[];
		extern U8 __cdecl ARMEX_Operand2LSRRegisterNC_SaveCarry[];
		extern U8 __cdecl ARMEX_Operand2LSRRegisterNC_End[];
		extern U8 __cdecl ARMEX_Operand2ASRRegisterCF[];
		extern U8 __cdecl ARMEX_Operand2ASRRegisterCF_GetCPSR[];
		extern U8 __cdecl ARMEX_Operand2ASRRegisterCF_SaveCarry[];
		extern U8 __cdecl ARMEX_Operand2ASRRegisterCF_BigShift[];
		extern U8 __cdecl ARMEX_Operand2ASRRegisterCF_Done[];
		extern U8 __cdecl ARMEX_Operand2ASRRegisterCF_End[];
		extern U8 __cdecl ARMEX_Operand2ASRRegisterNC[];
		extern U8 __cdecl ARMEX_Operand2ASRRegisterNC_SaveCarry[];
		extern U8 __cdecl ARMEX_Operand2ASRRegisterNC_End[];
		extern U8 __cdecl ARMEX_Operand2RORRegisterCF[];
		extern U8 __cdecl ARMEX_Operand2RORRegisterCF_GetCPSR[];
		extern U8 __cdecl ARMEX_Operand2RORRegisterCF_SaveCarry[];
		extern U8 __cdecl ARMEX_Operand2RORRegisterCF_Done[];
		extern U8 __cdecl ARMEX_Operand2RORRegisterCF_End[];
		extern U8 __cdecl ARMEX_Operand2RORRegisterNC[];
		extern U8 __cdecl ARMEX_Operand2RORRegisterNC_SaveCarry[];
		extern U8 __cdecl ARMEX_Operand2RORRegisterNC_End[];
		extern U8 __cdecl ARMEX_RegisterLoad[];
		extern U8 __cdecl ARMEX_RegisterLoad_End[];
		extern U8 __cdecl ARMEX_ImmediateLoad[];
		extern U8 __cdecl ARMEX_ImmediateLoad_End[];
		extern U8 __cdecl ARMEX_BDP_AND[];
		extern U8 __cdecl ARMEX_BDP_AND_End[];
		extern U8 __cdecl ARMEX_BDP_EOR[];
		extern U8 __cdecl ARMEX_BDP_EOR_End[];
		extern U8 __cdecl ARMEX_BDP_SUB[];
		extern U8 __cdecl ARMEX_BDP_SUB_End[];
		extern U8 __cdecl ARMEX_BDP_RSB[];
		extern U8 __cdecl ARMEX_BDP_RSB_End[];
		extern U8 __cdecl ARMEX_BDP_ADD[];
		extern U8 __cdecl ARMEX_BDP_ADD_End[];
		extern U8 __cdecl ARMEX_BDP_ADC[];
		extern U8 __cdecl ARMEX_BDP_ADC_End[];
		extern U8 __cdecl ARMEX_BDP_SBC[];
		extern U8 __cdecl ARMEX_BDP_SBC_End[];
		extern U8 __cdecl ARMEX_BDP_RSC[];
		extern U8 __cdecl ARMEX_BDP_RSC_End[];
		extern U8 __cdecl ARMEX_BDP_ORR[];
		extern U8 __cdecl ARMEX_BDP_ORR_End[];
		extern U8 __cdecl ARMEX_BDP_BIC[];
		extern U8 __cdecl ARMEX_BDP_BIC_End[];
		extern U8 __cdecl ARMEX_BDP_MVN[];
		extern U8 __cdecl ARMEX_BDP_MVN_End[];
		extern U8 __cdecl ARMEX_BDP_MSR[];
		extern U8 __cdecl ARMEX_BDP_MSR_End[];
		extern U8 __cdecl ARMEX_BDP_StoreOp2[];
		extern U8 __cdecl ARMEX_BDP_StoreOp2_End[];
		extern U8 __cdecl ARMEX_BDP_StoreOp1[];
		extern U8 __cdecl ARMEX_BDP_StoreOp1_End[];
		extern U8 __cdecl ARMEX_SetConditionCodes[];
		extern U8 __cdecl ARMEX_SetConditionCodes_End[];
		extern U8 __cdecl ARMEX_GetCCOnOp1[];
		extern U8 __cdecl ARMEX_GetCCOnOp1_End[];
		extern U8 __cdecl ARMEX_GetCCOnOp2[];
		extern U8 __cdecl ARMEX_GetCCOnOp2_End[];
		extern U8 __cdecl ARMEX_SetLogicalConditionCodesWithCarry[];
		extern U8 __cdecl ARMEX_SetLogicalConditionCodesWithCarry_End[];
		extern U8 __cdecl ARMEX_SetLogicalConditionCodesWithoutCarry[];
		extern U8 __cdecl ARMEX_SetLogicalConditionCodesWithoutCarry_End[];
		extern U8 __cdecl ARMEX_SaveHistory[];
		extern U8 __cdecl ARMEX_SaveHistory_Copy[];
		extern U8 __cdecl ARMEX_SaveHistory_End[];
		extern U8 __cdecl ARMEX_MemTransferNoHistory[];
		extern U8 __cdecl ARMEX_MemTransferNoHistory_Dest[];
		extern U8 __cdecl ARMEX_MemTransferNoHistory_End[];
		extern U8 __cdecl ARMEX_Depart[];
		extern U8 __cdecl ARMEX_Depart_Exceptions[];
		extern U8 __cdecl ARMEX_Depart_End[];
		extern U8 __cdecl ARMEX_SetupMultiply[];
		extern U8 __cdecl ARMEX_SetupMultiply_Op2[];
		extern U8 __cdecl ARMEX_SetupMultiply_NormalEnd[];
		extern U8 __cdecl ARMEX_SetupMultiply_End[];
		extern U8 __cdecl ARMEX_MultiplyUnsigned[];
		extern U8 __cdecl ARMEX_MultiplyUnsigned_End[];
		extern U8 __cdecl ARMEX_MultiplySigned[];
		extern U8 __cdecl ARMEX_MultiplySigned_End[];
		extern U8 __cdecl ARMEX_MultiplyAccumulate[];
		extern U8 __cdecl ARMEX_MultiplyAccumulate_End[];
		extern U8 __cdecl ARMEX_MultiplyDone[];
		extern U8 __cdecl ARMEX_MultiplyDone_End[];
		extern U8 __cdecl ARMEX_MultiplyUpdateFlags[];
		extern U8 __cdecl ARMEX_MultiplyUpdateFlags_GetCPSR[];
		extern U8 __cdecl ARMEX_MultiplyUpdateFlags_CPSRCopy[];
		extern U8 __cdecl ARMEX_MultiplyUpdateFlags_End[];
		extern U8 __cdecl ARMEX_MaskLowBits[];
		extern U8 __cdecl ARMEX_MaskLowBits_End[];
		extern U8 __cdecl ARMEX_WriteByte[];
		extern U8 __cdecl ARMEX_WriteByte_End[];
		extern U8 __cdecl ARMEX_WriteWord[];
		extern U8 __cdecl ARMEX_WriteWord_End[];
		extern U8 __cdecl ARMEX_SaveEBXNoMask[];
		extern U8 __cdecl ARMEX_SaveEBXNoMask_End[];
		extern U8 __cdecl ARMEX_SaveEBXMask[];
		extern U8 __cdecl ARMEX_SaveEBXMask_End[];
		extern U8 __cdecl ARMEX_ReadMemory[];
		extern U8 __cdecl ARMEX_ReadMemory_GetBounds[];
		extern U8 __cdecl ARMEX_ReadMemory_GetWaitStates[];
		extern U8 __cdecl ARMEX_ReadMemory_GetBitvector[];
		extern U8 __cdecl ARMEX_ReadMemory_GetBaseAddress[];
		extern U8 __cdecl ARMEX_ReadMemory_BPType[];
		extern U8 __cdecl ARMEX_ReadMemory_End[];
		extern U8 __cdecl ARMEX_ReadByte[];
		extern U8 __cdecl ARMEX_ReadByte_End[];
		extern U8 __cdecl ARMEX_ReadWord[];
		extern U8 __cdecl ARMEX_ReadWord_End[];
		extern U8 __cdecl ARMEX_JumpTable[];
		extern U8 __cdecl ARMEX_JumpTable_Base[];
		extern U8 __cdecl ARMEX_JumpTable_End[];
		extern U8 __cdecl ARMEX_LongMultiplyDone[];
		extern U8 __cdecl ARMEX_LongMultiplyDone_RDHI[];
		extern U8 __cdecl ARMEX_LongMultiplyDone_End[];
		extern U8 __cdecl ARMEX_LongMultiplyAccDone[];
		extern U8 __cdecl ARMEX_LongMultiplyAccDone_RDHI[];
		extern U8 __cdecl ARMEX_LongMultiplyAccDone_End[];
		extern U8 __cdecl ARMEX_LongMultiplyUpdateFlags[];
		extern U8 __cdecl ARMEX_LongMultiplyUpdateFlags_GetCPSR[];
		extern U8 __cdecl ARMEX_LongMultiplyUpdateFlags_CPSRCopy[];
		extern U8 __cdecl ARMEX_LongMultiplyUpdateFlags_End[];
		extern U8 __cdecl ARMEX_MultiplyUnsigned_ZeroCTEnd[];
		extern U8 __cdecl ARMEX_MultiplyUnsigned_NormalCTEnd[];
		extern U8 __cdecl ARMEX_ReadSByte[];
		extern U8 __cdecl ARMEX_ReadSByte_End[];
		extern U8 __cdecl ARMEX_ReadHalfword[];
		extern U8 __cdecl ARMEX_ReadHalfword_End[];
		extern U8 __cdecl ARMEX_ReadSHalfword[];
		extern U8 __cdecl ARMEX_ReadSHalfword_End[];
		extern U8 __cdecl ARMEX_WriteHalfword[];
		extern U8 __cdecl ARMEX_WriteHalfword_End[];
		extern U8 __cdecl ARMEX_LongBranch[];
		extern U8 __cdecl ARMEX_LongBranch_Offset[];
		extern U8 __cdecl ARMEX_LongBranch_StoreR15[];
		extern U8 __cdecl ARMEX_LongBranch_StoreR14[];
		extern U8 __cdecl ARMEX_LongBranch_End[];
		extern U8 __cdecl ARMEX_CodeBreak[];
		extern U8 __cdecl ARMEX_CodeBreak_StoreR15[];
		extern U8 __cdecl ARMEX_CodeBreak_End[];
		extern U8 __cdecl ARMEX_R15Op1_Mask[];
		extern U8 __cdecl ARMEX_R15Op1_Mask_End[];
		extern U8 __cdecl ARMEX_R15Op2_Mask[];
		extern U8 __cdecl ARMEX_R15Op2_Mask_End[];
		extern U8 __cdecl ARMEX_R15Mask[];
		extern U8 __cdecl ARMEX_R15Mask_R15[];
		extern U8 __cdecl ARMEX_R15Mask_R15Copy[];
		extern U8 __cdecl ARMEX_R15Mask_End[];
		extern U8 __cdecl ARMEX_TimerHeader[];
		extern U8 __cdecl ARMEX_TimerHeader_End[];
		extern U8 __cdecl ARMEX_TimerFooter[];
		extern U8 __cdecl ARMEX_TimerFooter_End[];
		extern U8 __cdecl ARMEX_TimerCall[];
		extern U8 __cdecl ARMEX_TimerCall_End[];
		extern U8 __cdecl ARMEX_1PerCycleTimerUpdate[];
		extern U8 __cdecl ARMEX_1PerCycleTimerUpdate_Cycles[];
		extern U8 __cdecl ARMEX_1PerCycleTimerUpdate_End[];
		extern U8 __cdecl ARMEX_1PerNCycleTimerUpdate[];
		extern U8 __cdecl ARMEX_1PerNCycleTimerUpdate_LowData[];
		extern U8 __cdecl ARMEX_1PerNCycleTimerUpdate_N[];
		extern U8 __cdecl ARMEX_1PerNCycleTimerUpdate_Shift[];
		extern U8 __cdecl ARMEX_1PerNCycleTimerUpdate_LowDataCopy[];
		extern U8 __cdecl ARMEX_1PerNCycleTimerUpdate_End[];
		extern U8 __cdecl ARMEX_CascadeTimerUpdate[];
		extern U8 __cdecl ARMEX_CascadeTimerUpdate_End[];
		extern U8 __cdecl ARMEX_TimerIRQCheck[];
		extern U8 __cdecl ARMEX_TimerIRQCheck_Mask[];
		extern U8 __cdecl ARMEX_TimerIRQCheck_End[];
		extern U8 __cdecl ARMEX_TimerUpdateMultiply[];
		extern U8 __cdecl ARMEX_TimerUpdateMultiply_End[];
		extern U8 __cdecl ARMEX_JoinCycles[];
		extern U8 __cdecl ARMEX_JoinCycles_End[];
		extern U8 __cdecl ARMEX_GetBranchMemTime[];
		extern U8 __cdecl ARMEX_GetBranchMemTime_WaitStates[];
		extern U8 __cdecl ARMEX_GetBranchMemTime_End[];

		extern u32 __cdecl ExecuteOp(U8 * op);
		extern void __cdecl IncrementTimers(u32 cycles);

		extern U32 g_instructionStep;
		extern u32 g_exceptions;
		extern u32 g_exceptionMask;
		extern U8 g_maskTable[];
		extern U8 * g_readMap[];
		extern U8 * g_writeMap[];
		extern U32 g_readBounds[];
		extern U8 * g_readBPMap[];
		extern U32 g_writeBounds[];
		extern U32 g_memMask[];			// HEREHEREHERE have to put bounds
		extern U8 * g_writeBPMap[];		// check and mask into GetB(), etc.
		extern U8 * g_codeBPMap[];
		extern U8 * g_userRBPMap[];
		extern U8 * g_userWBPMap[];
		extern U8 * g_userCBPMap[];
		extern U32 g_clocksTilNext;
		extern U32 g_badAddress;
		extern U32 g_wordS[];
		extern U32 g_wordN[];
		extern U32 g_dwordS[];
		extern U32 g_dwordN[];

		extern U32 g_history[];
		extern MType g_front;
		extern MType g_cursor;
		extern U32 g_pendingTimerInterrupts;
		extern U8 g_timerUpdateCode[];
	};                               

	// These are the various super-atomic stepping options.
	enum {
		STEP_INTO,
		STEP_OVER,
		STEP_OUT
	};

	extern u32 g_executeOffset;
	extern u32 g_retiredCheck;
	extern u32 g_retiredBP;
	extern u32 g_branchBP;
	extern u32 g_branchDepth;
	extern u32 g_nextR13, g_nextR14, g_nextR15;
	extern u8 g_stepOut;
	extern Sticky < Bool > g_skipStartup;
	extern Sticky < Bool > g_useThumbSyntax;
	extern Sticky < Bool > g_useMnemonic;
	extern U32 g_internalCounters[];

	//  We will do experiments to see how long each of these items takes.
	// We may have to alter our timing if we determine that the cycle times
	// depend on the kind of memory we're fetching from (very likely the
	// case. . .)
	#define CYCLES_N				0
	#define CYCLES_I				1	// THIS IS DEFINITE.
	#define CYCLES_S				0

	#include "arm.ipp"

#endif
