/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	Inlines for the ARM processor.

****************************************************************************/

extern Bool g_notRetired;

/////////////////////////////////////////////////////////////////////////////
//  What mode am I in?
inline u32 ARM::GetMode(void) const {
	return m_registers.Get(CPSR, 0) & 0x1f;
}

/////////////////////////////////////////////////////////////////////////////
//  Get the value of the indicated register.
inline u32 ARM::GetRegister(u8 index) const {
	return m_registers.Get(index,
						   GetMode());
}

/////////////////////////////////////////////////////////////////////////////
//  Set the specified register to the specified value.
inline void ARM::SetRegister(u8 index, u32 value) {
	m_registers.Set(index, GetMode(), value);
}

/////////////////////////////////////////////////////////////////////////////
//  Set the specified register to the specified value, using the history
// buffer.
inline void ARM::SetRegisterHistory(u8 index, u32 value) {
	m_registers.SetHistory(index, GetMode(), value);
}

/////////////////////////////////////////////////////////////////////////////
//  Jump to the indicated address.
inline void ARM::Goto(u32 address) {
	SetRegister(15, address);
}

/////////////////////////////////////////////////////////////////////////////
//  How many bytes per instruction, currently.
inline u8 ARM::GetInstructionSize(void) {
	if (GetState() == STATE_ARM) {
		return 4;
	} else if (GetState() == STATE_THUMB) {
		return 2;
	} else {
		return 1;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Am I in ARM state or Thumb state.
inline u8 ARM::GetState(void) {
	return (GetRegister(CPSR) >> 5) & 1;
}

template < class Target >
HackClass < Target >::HackClass(ARM & arm) : m_arm(arm) {
}

/////////////////////////////////////////////////////////////////////////////
//  This template function unifies all of the decoding logic.
template < class Target >
void HackClass < Target >::Decode(Target & t, u32 address) {
	if (m_arm.GetState() == m_arm.STATE_ARM) {
		u32 d = m_arm.m_sysMem.GetD(address);

		switch((d >> 26) & 3) {
			case 0:
				if (!(((d >> 4) + 0xf) & 0xffffUL)) {
					m_arm.DecBX(address, d, t);
				} else {
					if (((d >> 25) & 1) || ((d & 0x90) != 0x90)) {
						m_arm.DecDataProcessing(address, d, t);
					} else {
						if (d & 0x60) {
							if ((d >> 22) & 1) {
								m_arm.DecHalfwordImm(address, d, t);
							} else {
								m_arm.DecHalfwordReg(address, d, t);
							}
						} else {
							switch((d >> 23) & 3) {
								case 0:
									m_arm.DecMultiply(address, d, t);
									break;
								case 1:
									m_arm.DecMultiplyLong(address, d, t);
									break;
								case 2:
									m_arm.DecSingleDataSwap(address, d, t);
									break;
								case 3:
									m_arm.DecUndefined(address, d, t);
									break;
							}
						}
					}
				}
				break;
			case 1:
				if (((d >> 25) & 1) && ((d >> 4) & 1)) {
					m_arm.DecUndefined(address, d, t);
				} else {
					m_arm.DecSingleDataTransfer(address, d, t);
				}
				break;
			case 2:
				if (d & (1 << 25)) {
					m_arm.DecBranch(address, d, t);
				} else {
					m_arm.DecBlockDataTransfer(address, d, t);
				}
				break;
			case 3:
				if (d & (1 << 25)) {
					if (d & (1 << 24)) {
						m_arm.DecSWI(address, d, t);
					} else {
						if (d & (1 << 4)) {
							m_arm.DecCoprocRegTransfer(address, d, t);
						} else {
							m_arm.DecCoprocDataOp(address, d, t);
						}
					}
				} else {
					m_arm.DecCoprocDataTransfer(address, d, t);
				}
				break;
		}
	} else if (m_arm.GetState() == m_arm.STATE_THUMB) {
		u16 w = m_arm.m_sysMem.GetW(address);

		switch((w >> 13) & 7) {
			case 0:
				if (((w >> 11) & 3) == 3) {
					m_arm.DecTHAddSub(address, w, t);
				} else {
					m_arm.DecTHMSR(address, w, t);
				}
				break;
			case 1:
				m_arm.DecTHOpImm(address, w, t);
				break;
			case 2:
				if ((w >> 12) & 1) {
					if ((w >> 9) & 1) {
						m_arm.DecTHLdStSexBHW(address, w, t);
					} else {
						m_arm.DecTHLdStRegOff(address, w, t);
					}
				} else if ((w >> 11) & 1) {
					m_arm.DecTHPCRelLd(address, w, t);
				} else {
					if ((w >> 10) & 1) {
						m_arm.DecTHHiRegOps(address, w, t);
					} else {
						m_arm.DecTHALUOps(address, w, t);
					}
				}
				break;
			case 3:
				m_arm.DecTHLdStImm(address, w, t);
				break;
			case 4:
				if ((w >> 12) & 1) {
					m_arm.DecTHSPRelLdSt(address, w, t);
				} else {
					m_arm.DecTHLdStHw(address, w, t);
				}
				break;
			case 5:
				if ((w >> 12) & 1) {
					if ((w >> 10) & 1) {
						m_arm.DecTHPushPop(address, w, t);
					} else {
						m_arm.DecTHAddOffSP(address, w, t);
					}
				} else {
					m_arm.DecTHLEA(address, w, t);
				}
				break;
			case 6:
				if ((w >> 12) & 1) {
					if (((w >> 8) & 0xf) == 0xf) {
						m_arm.DecTHSWI(address, w, t);
					} else {
						m_arm.DecTHCondBr(address, w, t);
					}
				} else {
					m_arm.DecTHMLS(address, w, t);
				}
				break;
			case 7:
				if ((w >> 12) & 1) {
					m_arm.DecTHLongBrLink(address, w, t);
				} else {
					m_arm.DecTHUncondBranch(address, w, t);
				}
				break;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This function will dive right into the template to give us a disassembly.
inline void ARM::Decode(FString & f, u32 address) {
	HackClass < FString > m(*this);

	m.Decode(f, address);
}

/////////////////////////////////////////////////////////////////////////////
//  This function will fill up our code cache.
inline void ARM::Decode(ARM::DecodedInstruction & d, u32 address) {
	HackClass < ARM::DecodedInstruction > m(*this);

	m.Decode(d, address);
}

/////////////////////////////////////////////////////////////////////////////
//  Has the most recent instruction retired?
inline Bool ARM::Retired(void) const {
	return !g_notRetired;
}

/////////////////////////////////////////////////////////////////////////////
//  What's the current IP?
inline u32 ARM::GetCurrentIP(void) const {
	return GetRegister(15);
}

/////////////////////////////////////////////////////////////////////////////
//  Another (of many) cheats.
inline U32 ARM::GetRegAddress(u8 index) {
	return (U32)m_registers.GetAddress(index, GetMode());
}

/////////////////////////////////////////////////////////////////////////////
//  Get the register address for the indicated mode.
inline U32 ARM::GetRegAddress(u8 index, u8 mode) {
	return (U32)m_registers.GetAddress(index, mode);
}

/////////////////////////////////////////////////////////////////////////////
//  Enables me to use <<= operator for fixup lists.
inline Bool ARM::Fixup::operator ==(const ARM::Fixup & a) const {
	return dispSpot == a.dispSpot;
}

/////////////////////////////////////////////////////////////////////////////
//  Retrieve the list of active IRQs.
inline u16 & ARM::GetIRQs(void) {
	return m_irqs;
}

