/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	This handles all of the ARM disassemblers.

****************************************************************************/

#include "gba.h"

static char s_scratch[16];

Sticky < Bool > g_useThumbSyntax("Use hi-reg syntax", TRUE);
Sticky < Bool > g_useMnemonic("Use mnemonic", FALSE);

/////////////////////////////////////////////////////////////////////////////
//  Extract the condition code from the opcode.
static const char * GetConditionCode(u32 code) {
	static const char * s_cCodes[] = {
		"EQ",
		"NE",
		"CS",
		"CC",
		"MI",
		"PL",
		"VS",
		"VC",
		"HI",
		"LS",
		"GE",
		"LT",
		"GT",
		"LE",
		""		// Actually ALways, but this is generally omitted.
	};                           
	return s_cCodes[code >> 28]; 
}

/////////////////////////////////////////////////////////////////////////////
//  Convert a value to a register.
static const char * ValToReg(u8 r) {
	s_scratch[0] = 'R';          
	itoa(r, s_scratch + 1, 10);  
	return s_scratch;            
}                                
                                 
/////////////////////////////////////////////////////////////////////////////
//  Convert a value to a thumb register.
static const char * THValToReg(u8 r, Bool hi=FALSE) {
	if (*g_useThumbSyntax) {
		if (hi) {                    
			s_scratch[0] = 'H';      
		} else {                     
			s_scratch[0] = 'R';      
		}
		if (hi && (r >= 5)) {
			r += 8;
			switch(r) {
				case 13:
					s_scratch[0] = 's';
					s_scratch[1] = 'p';
					s_scratch[2] = '\0';
					break;
				case 14:
					s_scratch[0] = 'l';
					s_scratch[1] = 'r';
					s_scratch[2] = '\0';
					break;
				case 15:
					s_scratch[0] = 'p';
					s_scratch[1] = 'c';
					s_scratch[2] = '\0';
					break;
			}
		} else {
			itoa(r, s_scratch + 1, 10);  
		}
	} else {
		s_scratch[0] = 'R';
		if (hi) {
			r += 8;
		}
		itoa(r, s_scratch + 1, 10);  
	}
	return s_scratch;            
}                                
                                 
/////////////////////////////////////////////////////////////////////////////
//  Convert a value to an offset.
static const char * THValToOffset(u32 offset, Bool negative=FALSE) {
	MType i=0;                   
                                 
	s_scratch[i++] = '#';        
	if (negative) {              
		s_scratch[i++] = '-';    
	}
	s_scratch[i++] = '#';
	s_scratch[i++] = '0';        
	s_scratch[i++] = 'x';        
	itoa(offset, s_scratch + i, 16);
	return s_scratch;            
}                                

/////////////////////////////////////////////////////////////////////////////
//  Convert a bitvector to a list of registers.
static const FString & ValToRegList(u16 code) {
	s8 first, current;
	Bool needComma;
	static FString retval;

	if (!(code)) {
		retval = ' ';
	} else {
		retval = ' ';

		for (current=0, first = -1, needComma = FALSE;
			 current<17;
			 ++current) {
			if ((code >> current) & 1) {
				if (first == -1) {
					first = current;
				}
			} else if (first != -1) {
				if (needComma) {
					retval += ", ";
				}
				if (first != current - 1) {
					retval += ValToReg(first);
					retval += '-';
					retval += ValToReg(current - 1);
				} else {
					retval += ValToReg(first);
				}
				first = -1;
				needComma = TRUE;
			}
		}
	}
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Get the register list for THUMB set.
static const FString & THValToRegList(u8 code) {
	return ValToRegList(code);
}

/////////////////////////////////////////////////////////////////////////////
//  This function will interpret the operand2 value.
static const char * Operand2(u32 code, Bool justShift) {
	static char s_retval[32];
	u8 index=0, save;
	u8 r;
	Bool rrx=FALSE;

	if (!((code >> 25) & 1) || (justShift)) {
		s_retval[index++] = 'R';
		r = code & 0xf;
		if (r > 9) {
			s_retval[index++] = '1';
			s_retval[index++] = '0' + r - 10;
		} else {
			s_retval[index++] = '0' + r;
		}
		save = index;
		s_retval[index++] = ',';
		s_retval[index++] = ' ';
		s_retval[index + 2] = 'R';
		switch((code >> 5) & 3) {
			case 0:
				s_retval[index + 2] = 'L';
			case 1:
				s_retval[index++] = 'L';
				s_retval[index++] = 'S';
				break;
			case 2:
				s_retval[index++] = 'A';
				s_retval[index++] = 'S';
				break;
			case 3:
				if (((code & 0x10) && !justShift) || ((code >> 7) & 0x1f)) {
					s_retval[index++] = 'R';
					s_retval[index++] = 'O';
				} else {
					s_retval[index+2] = 'X';
					s_retval[index++] = 'R';
					s_retval[index++] = 'R';
					rrx = TRUE;
				}
				break;
		}
		++index;
		if (rrx) {
			save = index;
		}
		s_retval[index++] = ' ';
		if ((code & 0x10) && !justShift) {
			s_retval[index++] = 'R';
			r = ((code >> 8) & 0xf);
			if (r > 9) {
				s_retval[index++] = '1';
				s_retval[index++] = '0' + r - 10;
			} else {
				s_retval[index++] = '0' + r;
			}
		} else {
			s_retval[index++] = '#';
			s_retval[index++] = '0';
			s_retval[index++] = 'x';
			if (r = (code >> 7) & 0x1f) {
				if (r > 0xf) {
					s_retval[index++] = '1';
					r -= 0x10;
				} else {
					s_retval[index++] = '0';
				}
				if (r > 9) {
					s_retval[index++] = 'A' + r - 10;
				} else {
					s_retval[index++] = '0' + r;
				}
			} else {
				s_retval[save] = '\0';
			}
		}
		s_retval[index] = '\0';
	} else {
		U32 val;

		s_retval[index++] = '#';
		s_retval[index++] = '0';
		s_retval[index++] = 'x';
		val = code & 0xff;
		r = (code >> 8) & 0xf;
		r += r;

		val = Ror(val, r);

		// 8 nibbles in a dword.
		for (r=0; r<8; ++r) {
			u8 c = val >> 28;

			if (c > 9) {
				s_retval[index++] = 'A' + c - 10;
			} else {
				s_retval[index++] = '0' + c;
			}
			val <<= 4;
		}
		s_retval[index] = '\0';
	}
	return s_retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a data processing instruction.  This will include MRS/MSR
// instructions.
void ARM::DecDataProcessing(u32 address, u32 code, FString & dest) {
	if ((code & 0x0fbf0fff) == 0x010f0000) {
		dest += "MRS";
		dest += GetConditionCode(code);
		dest += '\t';
		dest += ValToReg((code >> 12) & 0xf);
		dest += ", ";
		if ((code >> 22) & 1) {
			dest += "SPSR";
		} else {
			dest += "CPSR";
		}
	} else if ((code & 0x0fb7fff0) == 0x0121f000) {
		dest += "MSR";
		dest += GetConditionCode(code);
		dest += '\t';
		if ((code >> 22) & 1) {
			dest += "SPSR";
		} else {
			dest += "CPSR";
		}
		dest += ", ";
		dest += ValToReg((code >> 0) & 0xf);
	} else if ((code & 0x0dbff000) == 0x0128f000) {
		dest += "MSR";
		dest += GetConditionCode(code);
		dest += '\t';
		if ((code >> 22) & 1) {
			dest += "SPSR_flg";
		} else {
			dest += "CPSR_flg";
		}
		dest += ", ";
		if ((code >> 25) & 1) {
			u32 val = code & 0xff;
			u8 r = (code >> 8) & 0xf;
			dest += THValToOffset(Ror(val, r + r));
		} else {
			dest += ValToReg((code >> 0) & 0xf);
		}
	} else {
		static const char * s_opcodes[] = {
			"AND",
			"EOR",
			"SUB",
			"RSB",
			"ADD",
			"ADC",
			"SBC",
			"RSC",
			"TST",
			"TEQ",
			"CMP",
			"CMN",
			"ORR",
			"MOV",
			"BIC",
			"MVN",
		};
		u8 opcode = (code >> 21) & 0xf;
	
		dest += s_opcodes[opcode];
		dest += GetConditionCode(code);
		switch(opcode) {
			case 8:
			case 9:
			case 10:
			case 11:
//				if ((code >> 20) & 1) {
//					dest += 'S';
//				}
				dest += '\t';
				dest += ValToReg((code >> 16) & 0xf);
				dest += ", ";
				dest += Operand2(code, FALSE);
				break;
			case 13:
			case 15:
				if ((code >> 20) & 1) {
					dest += 'S';
				}
				dest += '\t';
				dest += ValToReg((code >> 12) & 0xf);
				dest += ", ";
				dest += Operand2(code, FALSE);
				break;
			default:
				if ((code >> 20) & 1) {
					dest += 'S';
				}
				dest += '\t';
				dest += ValToReg((code >> 12) & 0xf);
				dest += ", ";
				dest += ValToReg((code >> 16) & 0xf);
				dest += ", ";
				dest += Operand2(code, FALSE);
				break;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a multiply instruction.
void ARM::DecMultiply(u32 address, u32 code, FString & dest) {
	if ((code >> 21) & 1) {
		dest += "MLA";
	} else {
		dest += "MUL";
	}
	dest += GetConditionCode(code);
	if ((code >> 20) & 1) {
		dest += 'S';
	}
	dest += '\t';
	dest += ValToReg((code >> 16) & 0xf);
	dest += ", ";
	dest += ValToReg(code & 0xf);
	dest += ", ";
	dest += ValToReg((code >> 8) & 0xf);
	if ((code >> 21) & 1) {
		dest += ", ";
		dest += ValToReg((code >> 12) & 0xf);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a multiply long instruction.
void ARM::DecMultiplyLong(u32 address, u32 code, FString & dest) {
	if ((code >> 22) & 1) {
		dest += 'S';
	} else {
		dest += 'U';
	}
	if ((code >> 21) & 1) {
		dest += "MLAL";
	} else {
		dest += "MULL";
	}
	dest += GetConditionCode(code);
	if ((code >> 20) & 1) {
		dest += 'S';
	}
	dest += '\t';
	dest += ValToReg((code >> 16) & 0xf);
	dest += ':';
	dest += ValToReg((code >> 12) & 0xf);
	dest += ", ";
	dest += ValToReg(code & 0xf);
	dest += ", ";
	dest += ValToReg((code >> 8) & 0xf);
}

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a single data swap instruction.
void ARM::DecSingleDataSwap(u32 address, u32 code, FString & dest) {
	dest += "SWP";
	dest += GetConditionCode(code);
	if ((code >> 22) & 1) {
		dest += 'B';
	}
	dest += '\t';
	dest += ValToReg((code >> 12) & 0xf);
	dest += ", ";
	dest += ValToReg(code & 0xf);
	dest += ", [";
	dest += ValToReg((code >> 16) & 0xf);
	dest += "]";
}

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a branch-exchange instruction.
void ARM::DecBX(u32 address, u32 code, FString & dest) {
	dest += "BX";
	dest += GetConditionCode(code);
	dest += '\t';
	dest += ValToReg(code & 0xf);
}

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a half-word register instruction.
void ARM::DecHalfwordReg(u32 address, u32 code, FString & dest) {
	if ((code >> 20) & 1) {
		dest += "LDR";
	} else {
		dest += "STR";
	}
	dest += GetConditionCode(code);
	switch((code >> 5) & 3) {
		case 0:
			Warn("Shouldn't get here, %s, %d.\n", __FILE__, __LINE__);
			break;
		case 1:
			dest += "H";
			break;
		case 2:
			dest += "SB";
			break;
		case 3:
			dest += "SH";
			break;
	}
	dest += '\t';
	dest += ValToReg((code >> 12) & 0xf);
	dest += ", [";
	dest += ValToReg((code >> 16) & 0xf);
	if (!((code >> 24) & 1)) {
		dest += "] ";
	} else {
		dest += ' ';
	}
	if ((code >> 23) & 1) {
		dest += '+';
	} else {
		dest += '-';
	}
	dest += ValToReg(code & 0xf);
	if ((code >> 24) & 1) {
		dest += ']';
		if ((code >> 21) & 1) {
			dest += '!';
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a half-word immediate instruction.
void ARM::DecHalfwordImm(u32 address, u32 code, FString & dest) {
	u8 val;

	if ((code >> 20) & 1) {
		dest += "LDR";
	} else {
		dest += "STR";
	}
	dest += GetConditionCode(code);
	switch((code >> 5) & 3) {
		case 0:
			Warn("Shouldn't get here, %s, %d.\n", __FILE__, __LINE__);
			break;
		case 1:
			dest += "H";
			break;
		case 2:
			dest += "SB";
			break;
		case 3:
			dest += "SH";
			break;
	}
	dest += '\t';
	dest += ValToReg((code >> 12) & 0xf);
	dest += ", [";
	dest += ValToReg((code >> 16) & 0xf);
	if (!((code >> 24) & 1)) {
		dest += "] ";
	} else {
		dest += ' ';
	}
	if ((code >> 23) & 1) {
		dest += '+';
	} else {
		dest += '-';
	}
	val = (((code >> 8) & 0xf) << 4) | (code & 0xf);
	dest += "0x";
	itoa(val, s_scratch, 16);
	dest += s_scratch;
	if ((code >> 24) & 1) {
		dest += ']';
		if ((code >> 21) & 1) {
			dest += '!';
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a single data transfer instruction.
void ARM::DecSingleDataTransfer(u32 address, u32 code, FString & dest) {
	Bool pre = (code >> 24) & 1;

	if ((code >> 20) & 1) {
		dest += "LDR";
	} else {
		dest += "STR";
	}
	dest += GetConditionCode(code);
	if ((code >> 22) & 1) {
		dest += 'B';
	}
	if (!pre && ((code >> 21) & 1)) {
		dest += 'T';
	}
	dest += '\t';
	dest += ValToReg((code >> 12) & 0xf);
	dest += ", ";
	dest += '[';
	dest += ValToReg((code >> 16) & 0xf);
	if (!pre) {
		dest += ']';
	}
	if ((code >> 23) & 1) {
		dest += '+';
	} else {
		dest += '-';
	}
	if ((code >> 25) & 1) {
		dest += Operand2(code, TRUE);
	} else {
		dest += "0x";
		itoa(code & 4095, s_scratch, 16);
		dest += s_scratch;
	}
	if (pre) {
		dest += ']';
		if ((code >> 21) & 1) {
			dest += '!';
		}
	}
}

static TString s_undefined("[UNDEFINED]");

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a undefined instruction.
void ARM::DecUndef(u32 address, u32 code, FString & dest) {
	dest += s_undefined;
	dest += "\tDW  0x";
	itoa(code, s_scratch, 16);
	dest += s_scratch;
}

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a block data transfer instruction.
void ARM::DecBlockDataTransfer(u32 address, u32 code, FString & dest) {
	static const char * s_names[] = {
		"ED",
		"FD",
		"EA",
		"FA",
		"FA",
		"EA",
		"FD",
		"ED",
		"DA",
		"DB",
		"IA",
		"IB",
		"DA",
		"DB",
		"IA",
		"IB"
	};
	u8 index;
	u8 reg;

	if ((code >> 20) & 1) {
		dest += "LDM";
	} else {
		dest += "STM";
	}
	dest += GetConditionCode(code);
	reg = (code >> 16) & 0xf;
	if (*g_useMnemonic) {
		if (reg == 13) {
			index = 0;
		} else {    
			index = 8;
		}
		if ((code >> 20) & 1) {
			index |= 4;
		}           
		if ((code >> 23) & 1) {
			index |= 2;
		}           
		if ((code >> 24) & 1) {
			index |= 1;
		}
		dest += s_names[index];
		dest += '\t';
		if (reg == 13) {
			dest += "SP";
		} else {
			dest += ValToReg(reg);
		}
	} else {
		dest += "\t[";
		if ((code >> 24) & 1) {
			if ((code >> 23) & 1) {
				dest += '+';
			} else {
				dest += '-';
			}
			dest += ValToReg(reg);
		} else {
			dest += ValToReg(reg);
			if ((code >> 23) & 1) {
				dest += '+';
			} else {
				dest += '-';
			}
		}
		dest += ']';
	}
	if ((code >> 21) & 1) {
		dest += '!';
	}
	dest += ", ";
	// Limits the loop below.
	code &= ~((u32)1 << 16);
	if (!(code & (u32)0xffff)) {
		dest += "{ }";
	} else {
		dest += '{';
		dest += ValToRegList(code);
		dest += " }";
	}
	if ((code >> 22) & 1) {
		dest += '^';
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a branch instruction.
void ARM::DecBranch(u32 address,
					u32 code,
					FString & dest) {
	u32 offset;

	dest += 'B';
	if ((code >> 24) & 1) {
		dest += 'L';
	}
	dest += GetConditionCode(code);
	dest += '\t';
	offset = code & (u32)0xffffff;
	address += 8;
	if (offset & 0x800000) {
		address += (offset | 0xff000000) << 2;
		g_gba.GetSymbol(dest, address);
	} else {
		address += offset << 2;
		g_gba.GetSymbol(dest, address);
	}
}

static TString s_coproc("COPROC");

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a coprocessor data transfer instruction.  Until I found out
// that the GBA has a coprocessor, I will put dummy stuff here for now.
void ARM::DecCoprocDataTransfer(u32 address, u32 code, FString & dest) {
	dest += s_coproc;
}

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a coprocessor data operation instruction.
void ARM::DecCoprocDataOp(u32 address, u32 code, FString & dest) {
	dest += s_coproc;
}

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a coprocessor register transfer instruction.
void ARM::DecCoprocRegTransfer(u32 address, u32 code, FString & dest) {
	dest += s_coproc;
}

/////////////////////////////////////////////////////////////////////////////
//  Disassemble a software interrupt instruction.
void ARM::DecSWI(u32 address, u32 code, FString & dest) {
	dest += "SWI";
	dest += GetConditionCode(code);
	dest += "\t0x";
	itoa((code & (u32)0xffffff), s_scratch, 16);
	dest += s_scratch;
}

/////////////////////////////////////////////////////////////////////////////
//  An undefined instruction.
void ARM::DecUndefined(u32 address, u32 code, FString & dest) {
	dest += s_undefined;
	dest += "\tDW  0x";
	itoa(code, s_scratch, 16);
	dest += s_scratch;
}

/////////////////////////////////////////////////////////////////////////////
//  Thumb ADD/Subtract.
void ARM::DecTHMSR(u32 address, u16 code, FString & dest) {
	static const char * s_codes[] = {
		"LSL",
		"LSR",
		"ASR"
	};
	dest += s_codes[(code >> 11) & 3];
	dest += '\t';
	dest += THValToReg((code >> 0) & 7);
	dest += ", ";
	dest += THValToReg((code >> 3) & 7);
	dest += ", ";
	dest += THValToOffset((code >> 6) & 0x1f);
}

/////////////////////////////////////////////////////////////////////////////
//  Perorm a simple add/sub.
void ARM::DecTHAddSub(u32 address, u16 code, FString & dest) {
	if ((code >> 9) & 1) {
		dest += "SUB";
	} else {
		dest += "ADD";
	}
	dest += '\t';
	dest += THValToReg((code >> 0) & 7);
	dest += ", ";
	dest += THValToReg((code >> 3) & 7);
	dest += ", ";
	if ((code >> 10) & 1) {
		dest += THValToOffset((code >> 6) & 7);
	} else {
		dest += THValToReg((code >> 6) & 7);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Various ops with immediate.
void ARM::DecTHOpImm(u32 address, u16 code, FString & dest) {
	static const char * s_codes[] = {
		"MOV",
		"CMP",
		"ADD",
		"SUB"
	};
	dest += s_codes[(code >> 11) & 3];
	dest += '\t';
	dest += THValToReg((code >> 8) & 7);
	dest += ", ";
	dest += THValToOffset((code >> 0) & 0xff);
}

/////////////////////////////////////////////////////////////////////////////
//  Various ALU ops.
void ARM::DecTHALUOps(u32 address, u16 code, FString & dest) {
	static const char * s_codes[] = {
		"AND",
		"EOR",
		"LSL",
		"LSR",
		"ASR",
		"ADC",
		"SBC",
		"ROR",
		"TST",
		"NEG",
		"CMP",
		"CMN",
		"ORR",
		"MUL",
		"BIC",
		"MVN"
	};
	dest += s_codes[(code >> 6) & 0xf];
	dest += '\t';
	dest += THValToReg((code >> 0) & 7);
	dest += ", ";
	dest += THValToReg((code >> 3) & 7);
}

/////////////////////////////////////////////////////////////////////////////
//  Hi-register operations.
void ARM::DecTHHiRegOps(u32 address, u16 code, FString & dest) {
	static const char * s_codes[] = {
		"ADD",
		"CMP",
		"MOV",
		"BX"
	};
	u8 u = (code >> 8) & 3;

	dest += s_codes[u];
	dest += '\t';
	if (u < 3) {
		dest += THValToReg((code >> 0) & 7, (code >> 7) & 1);
		dest += ", ";
		dest += THValToReg((code >> 3) & 7, (code >> 6) & 1);
	} else {
		dest += THValToReg((code >> 3) & 7, (code >> 6) & 1);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  PC-relative load.
void ARM::DecTHPCRelLd(u32 address, u16 code, FString & dest) {
	dest += "LDR\t";
	dest += THValToReg((code >> 8) & 7);
	dest += ", [";
	dest += THValToOffset(((address + 4) & ~3) +
						  ((code >> 0) & 0xff) * 4);
	dest += ']';
}

/////////////////////////////////////////////////////////////////////////////
//  Load/store with register offset.
void ARM::DecTHLdStRegOff(u32 address, u16 code, FString & dest) {
	if ((code >> 11) & 1) {
		dest += "LDR";
	} else {
		dest += "STR";
	}
	if ((code >> 10) & 1) {
		dest += 'B';
	}
	dest += '\t';
	dest += THValToReg((code >> 0) & 7);
	dest += ", [";
	dest += THValToReg((code >> 3) & 7);
	dest += ", ";
	dest += THValToReg((code >> 6) & 7);
	dest += ']';
}

/////////////////////////////////////////////////////////////////////////////
//  Load/store with/sign extension.
void ARM::DecTHLdStSexBHW(u32 address, u16 code, FString & dest) {
	static const char * s_codes[] = {
		"STRH",
		"LDSB",
		"LDRH",
		"LDSH"
	};
	dest += s_codes[(code >> 10) & 3];
	dest += '\t';
	dest += THValToReg((code >> 0) & 7);
	dest += ", [";
	dest += THValToReg((code >> 3) & 7);
	dest += ", ";
	dest += THValToReg((code >> 6) & 7);
	dest += ']';
}

/////////////////////////////////////////////////////////////////////////////
//  Load/store with immediate offset.
void ARM::DecTHLdStImm(u32 address, u16 code, FString & dest) {
	if ((code >> 11) & 1) {
		dest += "LDR";
	} else {
		dest += "STR";
	}
	if ((code >> 12) & 1) {
		dest += 'B';
	}
	dest += '\t';
	dest += THValToReg((code >> 0) & 7);
	dest += ", [";
	dest += THValToReg((code >> 3) & 7);
	dest += ", ";
	dest += THValToOffset((code >> 6) & 0x1f);
	dest += ']';
}

/////////////////////////////////////////////////////////////////////////////
//  Load/store halfword.
void ARM::DecTHLdStHw(u32 address, u16 code, FString & dest) {
	if ((code >> 11) & 1) {
		dest += "LDRH";
	} else {
		dest += "STRH";
	}
	dest += '\t';
	dest += THValToReg((code >> 0) & 7);
	dest += ", [";
	dest += THValToReg((code >> 3) & 7);
	dest += ", ";
	dest += THValToOffset((code >> 6) & 0x1f);
	dest += ']';
}

/////////////////////////////////////////////////////////////////////////////
//  Stack-relative load-store.
void ARM::DecTHSPRelLdSt(u32 address, u16 code, FString & dest) {
	if ((code >> 11) & 1) {
		dest += "LDR";
	} else {
		dest += "STR";
	}
	dest += '\t';
	dest += THValToReg((code >> 8) & 7);
	if (*g_useThumbSyntax) {
		dest += ", [SP, ";
	} else {
		dest += ", [R13, ";
	}
	dest += THValToOffset((code >> 0) & 0xff);
	dest += ']';
}

/////////////////////////////////////////////////////////////////////////////
//  Load effective address.
void ARM::DecTHLEA(u32 address, u16 code, FString & dest) {
	if ((code >> 11) & 1) {
		dest += "ADD\t";
		dest += THValToReg((code >> 8) & 7);
		if (*g_useThumbSyntax) {
			dest += ", SP, ";
		} else {
			dest += ", R13, ";
		}
		dest += THValToOffset(((code >> 0) & 0xff) * 4);
	} else {
		dest += "LDR\t";
		dest += THValToReg((code >> 8) & 7);
		dest += ", ";
		dest += THValToOffset(((address + 4) & ~3) +
							  ((code >> 0) & 0xff) * 4);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Add an offset to SP.
void ARM::DecTHAddOffSP(u32 address, u16 code, FString & dest) {
	if (*g_useThumbSyntax) {
		dest += "ADD\tSP, ";
	} else {
		dest += "ADD\tR13, ";
	}
	dest += THValToOffset(((code >> 0) & 0x7f) * 4,
						  ((code >> 7) & 1));
}

/////////////////////////////////////////////////////////////////////////////
//  Push or pop registers.
void ARM::DecTHPushPop(u32 address, u16 code, FString & dest) {
	u8 u = (code >> 11) & 1;

	if (u) {
		dest += "POP\t{";
	} else {
		dest += "PUSH\t{";
	}
	dest += THValToRegList((code >> 0) & 0xff);
	if ((code >> 8) & 1) {
		if ((code >> 0) & 0xff) {
			dest += ",";
		}
		if (u) {
			if (*g_useThumbSyntax) {
				dest += " PC }";
			} else {
				dest += " R15 }";
			}
		} else {
			if (*g_useThumbSyntax) {
				dest += " LR }";
			} else {
				dest += " R14 }";
			}
		}
	} else {
		dest += '}';
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Multiple load/store.
void ARM::DecTHMLS(u32 address, u16 code, FString & dest) {
	if ((code >> 11) & 1) {
		dest += "LDMIA\t";
	} else {
		dest += "STMIA\t";
	}
	dest += '[';
	dest += THValToReg((code >> 8) & 7);
	dest += ']';
	dest += "!, {";
	dest += THValToRegList((code >> 0) & 0xff);
	dest += '}';
}

/////////////////////////////////////////////////////////////////////////////
//  Conditional branch.
void ARM::DecTHCondBr(u32 address, u16 code, FString & dest) {
	dest += 'B';
	dest += GetConditionCode((code >> 8) << 28);
	dest += '\t';
	dest += THValToOffset(address + 4 + (s32)((S8)(code & 0xff)) * 2);
}

/////////////////////////////////////////////////////////////////////////////
//  Generate a software interrupt.
void ARM::DecTHSWI(u32 address, u16 code, FString & dest) {
	dest += "SWI\t";
	dest += THValToOffset(code & 0xff);
}

/////////////////////////////////////////////////////////////////////////////
//  Perform an unconditional branch.
void ARM::DecTHUncondBranch(u32 address, u16 code, FString & dest) {
	u32 offset;

	dest += "B\t";
	offset = code & 0x7ff;
	if (offset & 0x400) {
		offset |= 0xfffffc00;
	}
	dest += THValToOffset(address + 4 + offset * 2);
}

/////////////////////////////////////////////////////////////////////////////
//  Long branch and link.  HERE I'm making assumptions about sequential
// insructions here, in the future we may want to make this a bit more
// "intelligent". . .
void ARM::DecTHLongBrLink(u32 address, u16 code, FString & dest) {
	if (!((code >> 11) & 1)) {
		u32 target = address + 4;
		u32 imm = ((code >> 0) & 0x7ff);

		if (imm & 0x400) {
			imm |= 0xfffffc00;
		}
		target += imm << 12;
		code = m_sysMem.GetW(address + 2);
		target += ((code >> 0) & 0x7ff) * 2;
		dest += "BL\t";
		dest += THValToOffset(target);
	}
}

