/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	All of the ARM op decoders reside here.  HEREHEREHERE we have to find out
the difference in timing when a branch is taken vs. when it's not.

****************************************************************************/

#include "gba.h"

static u32 s_forceInclude;			// Exceptions we're forced to include.

//  These are the transfer types.
enum {
	TFR_WORD		= 0,
	TFR_UHWORD,
	TFR_SBYTE,
	TFR_SHWORD,
	TFR_UBYTE
};

U32 g_internalCounters[4];

/////////////////////////////////////////////////////////////////////////////
//  It's about time we made a function for this!  We give this the address,
// N, S, and I cycles; and it returns the total clocks.
u32 ARM::GetCycles(u32 address, u32 n, u32 i, u32 s) {
	u32 retval;

	address = (address >> 24) & 0xf;
	if (GetInstructionSize() == 2) {
		retval = s * (g_wordS[address] >> CYCLE_SHIFT) +
				 n * (g_wordN[address] >> CYCLE_SHIFT);
	} else {
		retval = s * (g_dwordS[address] >> CYCLE_SHIFT) +
				 n * (g_dwordN[address] >> CYCLE_SHIFT);
	}
	retval += i * CYCLES_I;
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Insert the stream of bytes into the decoded instruction stream.  I am
// inlining this for performance.
void ARM::Insert(ARM::DecodedInstruction & dest,
				 U8 * start,
				 U8 * end) {
	memcpy(&dest.dest[dest.base],
		   start,
		   end - start);
	dest.base += end - start;
}

/////////////////////////////////////////////////////////////////////////////
//  Clear all forced exceptions.
void ARM::ClearForcedExceptions(void) {
	s_forceInclude = 0;
}

/////////////////////////////////////////////////////////////////////////////
//  Force the following exceptions on the currently-decoded instruction.
void ARM::ForceExceptions(u32 u) {
	s_forceInclude |= u;
}

/////////////////////////////////////////////////////////////////////////////
//  Retrieve the exceptions that are currently being forced.
u32 ARM::GetForcedExceptions(void) {
	return s_forceInclude;
}

/////////////////////////////////////////////////////////////////////////////
//  Insert the indicated displacement into an x86 instruction that uses RM
// (but no SIB)
void ARM::InsertDispToRMImm(U8 * dest, u32 value) {
	*((U32 *)(dest + 2)) = value;
}

/////////////////////////////////////////////////////////////////////////////
//  Insert the indicated immediate value into a shift instruction.
void ARM::InsertShiftToReg(U8 * dest, u8 shift) {
	*((U8 *)(dest + 2)) = shift;
}

/////////////////////////////////////////////////////////////////////////////
//  Insert the indicated displacement into an x86 instruction that uses RM
// but no immediate.
void ARM::InsertDispToRegRM(U8 * dest, u32 value) {
	*((U32 *)(dest + 2)) = value;
}

/////////////////////////////////////////////////////////////////////////////
//  Insert the indicated immediate value into a x86 instruction that uses
// RM (but no SIB).
void ARM::InsertImmediateToRM(U8 * dest, u32 value) {
	*((U32 *)(dest + 6)) = value;
}

/////////////////////////////////////////////////////////////////////////////
//  Insert the indicated immediate value into a x86 instruction that uses
// RM (but no SIB or displacement).
void ARM::InsertImmediateToRMNoDisp(U8 * dest, u32 value) {
	*((U32 *)(dest + 2)) = value;
}

/////////////////////////////////////////////////////////////////////////////
//  Insert an immediate to a register-based instruction.
void ARM::InsertImmediateToReg(U8 * dest, u32 value) {
	*((U32 *)(dest + 1)) = value;
}

/////////////////////////////////////////////////////////////////////////////
//  Insert an immediate to an accumulator-based instruction.
void ARM::InsertImmediateToAcc(U8 * dest, u32 value) {
	*((U32 *)(dest + 1)) = value;
}

/////////////////////////////////////////////////////////////////////////////
//  Insert a displacement to an accumulator-based instruction.
void ARM::InsertDispToAcc(U8 * dest, u32 value) {
	*((U32 *)(dest + 1)) = value;
}

/////////////////////////////////////////////////////////////////////////////
//  Insert a displacement to a RegRM that uses SIB.
void ARM::InsertDispToSIBRegRM(U8 * dest, u32 value) {
	*((U32 *)(dest + 3)) = value;
}

/////////////////////////////////////////////////////////////////////////////
//  Insert an immediate to a shift instruction.
void ARM::InsertImmediateToShift(U8 * dest, u32 value) {
	*((U8 *)(dest + 2)) = value;
}

/////////////////////////////////////////////////////////////////////////////
//  When we detect that a change has been made to the timers, we change the
// timer update code.  However, we don't have to invalidate the cache!
// Sweet.
void ARM::MakeTimerUpdateCode(void) {
	u8 u;
	U16 * td = (U16 *)(g_readMap[4] + 0x100);
	u32 irq;
	DecodedInstruction dest;

	dest.dest = g_timerUpdateCode;
	dest.base = 0;
	dest.fixups = NULL;
	dest.jtables = NULL;
	dest.ip = 0;

	Insert(dest, ARMEX_TimerHeader, ARMEX_TimerHeader_End);
	for (u=0, irq=8; u<4; ++u, irq+=irq, td += 2) {
		if (td[1] & 0x80) {
			if (u && (td[1] & 4)) {
				// Cascade timer.
				InsertImmediateToReg(ARMEX_CascadeTimerUpdate,
									 (u32)td);
				Insert(dest,
					   ARMEX_CascadeTimerUpdate, 
					   ARMEX_CascadeTimerUpdate_End);
			} else {
				u8 f = td[1] & 3;

				switch(f) {
					case 0:
						InsertImmediateToReg(ARMEX_1PerCycleTimerUpdate,
											 (u32)td);
						Insert(dest,
							   ARMEX_1PerCycleTimerUpdate,
							   ARMEX_1PerCycleTimerUpdate_End);
						break;
					case 1:
					case 2:
					case 3:
						InsertShiftToReg(ARMEX_1PerNCycleTimerUpdate_Shift,
										 26 - 2 * f);
						InsertImmediateToReg(ARMEX_1PerNCycleTimerUpdate,
											 (u32)td);
						InsertDispToRegRM(ARMEX_1PerNCycleTimerUpdate_LowData,
												  (u32)(g_internalCounters + u));
						InsertDispToRegRM(ARMEX_1PerNCycleTimerUpdate_LowDataCopy,
												  (u32)(g_internalCounters + u));
						Insert(dest,
							   ARMEX_1PerNCycleTimerUpdate,
							   ARMEX_1PerNCycleTimerUpdate_End);
				}
			}
			if (td[1] & 0x40) {
				InsertImmediateToRMNoDisp(ARMEX_TimerIRQCheck_Mask,
										  irq);
				Insert(dest,
					   ARMEX_TimerIRQCheck,
					   ARMEX_TimerIRQCheck_End);
			}
		}
	}
	Insert(dest, ARMEX_TimerFooter, ARMEX_TimerFooter_End);
}

/////////////////////////////////////////////////////////////////////////////
//  This will insert code that updates our timers, and if need be, generates
// interrupts.
void ARM::InsertTimerUpdateCode(ARM::DecodedInstruction & dest) {
	Insert(dest,
		   ARMEX_TimerCall,
		   ARMEX_TimerCall_End);
}

/////////////////////////////////////////////////////////////////////////////
//  This will insert a condition-code check and return the offset to the
// jump displacement.  Since this operation is used everywhere, we will also
// check for code breakpoints here.
U16 * ARM::InsertConditionCodeCheck(u32 address,
									u8 flags,
									ARM::DecodedInstruction & dest,
									u32 & leftOff) {
	if (BreakpointCode(address)) {
		InsertImmediateToAcc(ARMEX_CodeBreak, EXC_EXEC_BREAKPOINT);
		InsertImmediateToRM(ARMEX_CodeBreak_StoreR15, address);
		InsertDispToRMImm(ARMEX_CodeBreak_StoreR15, GetRegAddress(0xf));
		Insert(dest, ARMEX_CodeBreak, ARMEX_CodeBreak_End);
	}
	if (flags != 0xe) {
		U16 * retval;

		InsertImmediateToReg(ARMEX_CheckCondition_SpecifyNotTaken,
							 GetCycles(address, 0, 0, 1) << CYCLE_SHIFT);
		InsertDispToAcc(ARMEX_CheckCondition_CPSR,
								GetRegAddress(CPSR) + 3);
		InsertDispToRegRM(ARMEX_CheckCondition_Condition,
								  (U32)(flags + g_maskTable));
		retval = ((U16 *)(dest.dest +
						  dest.base + 
						  (ARMEX_CheckCondition_Jump - ARMEX_CheckCondition) +
						  2));
		Insert(dest, ARMEX_CheckCondition, ARMEX_CheckCondition_End);
		leftOff = dest.base;
		return retval;
	} else {
		return NULL;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This does the wrap-up after a condition code check fails.
void ARM::InsertConditionCodePost(ARM::DecodedInstruction & dest,
								  U16 * d,
								  u32 leftOff,
								  u32 exceptions,
								  u32 address,
								  u32 cycles,
								  Bool isMultiply) {
	InsertImmediateToReg(ARMEX_SetCycles, cycles << CYCLE_SHIFT);
	Insert(dest, ARMEX_SetCycles, ARMEX_SetCycles_End);
	if (isMultiply) {
		Insert(dest, ARMEX_TimerUpdateMultiply, ARMEX_TimerUpdateMultiply_End);
	}

	//  If we have the external jump flag set, it means that this instruction
	// has altered R15 with an unknown value (known values will use the
	// fixup system).
	if (exceptions & EXC_EXTERNAL_JUMP) {
		InsertTimerUpdateCode(dest);
		InsertImmediateToAcc(ARMEX_Depart_Exceptions, exceptions);
		Insert(dest, ARMEX_Depart, ARMEX_Depart_End);
		if (!d)
			return;
		exceptions &= ~EXC_EXTERNAL_JUMP;
	}
	if (d) {
		*d = (U16)(dest.base - leftOff);
	}
	InsertTimerUpdateCode(dest);
	InsertImmediateToRM(ARMEX_Nop_StoreR15,
						address + GetInstructionSize());
	InsertDispToRMImm(ARMEX_Nop_StoreR15, GetRegAddress(0xf));
	InsertImmediateToRMNoDisp(ARMEX_Nop_ExceptionsRaised, exceptions);
	InsertImmediateToAcc(ARMEX_Nop_ExceptionsRaisedCopy, exceptions);
	Insert(dest, ARMEX_Nop, ARMEX_Nop_End);
}

/////////////////////////////////////////////////////////////////////////////
//  This does an exception check, but not an R15 update (for BDT).
void ARM::InsertExceptionCheck(ARM::DecodedInstruction & dest,
							   u32 exceptions,
							   u32 address,
							   u32 cycles) {
	InsertImmediateToReg(ARMEX_SetCycles, cycles << CYCLE_SHIFT);
	Insert(dest, ARMEX_SetCycles, ARMEX_SetCycles_End);
	Insert(dest, ARMEX_TimerUpdateMultiply, ARMEX_TimerUpdateMultiply_End);
	InsertTimerUpdateCode(dest);

	//  Do a check to see if any of the exceptions have been thrown.  If so,
	// we depart - but R15 contains its unchanged value.
	InsertImmediateToRM(ARMEX_Nop_StoreR15,
						address);
	InsertDispToRMImm(ARMEX_Nop_StoreR15, GetRegAddress(0xf));
	InsertImmediateToRMNoDisp(ARMEX_Nop_ExceptionsRaised, exceptions);
	InsertImmediateToAcc(ARMEX_Nop_ExceptionsRaisedCopy, exceptions);
	Insert(dest, ARMEX_Nop, ARMEX_Nop_End);
}

/////////////////////////////////////////////////////////////////////////////
//  This will insert a jump table using the specified register as an index.
// We return a pointer to the jump table itself.
U8 * ARM::InsertJumpTable(ARM::DecodedInstruction & dest,
						  u32 reg) {
	U8 * retval = (U8 *)(dest.dest + dest.base + 11);

	reg = GetRegAddress(reg);

	InsertDispToAcc(ARMEX_JumpTable, reg);
	InsertDispToRegRM(ARMEX_JumpTable_Base,
							  (u32)(retval));
	Insert(dest,
		   ARMEX_JumpTable,
		   ARMEX_JumpTable_End);
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Effect a SaveHistory on the indicated operand.
void ARM::SaveHistory(ARM::DecodedInstruction & dest, u32 address) {
	InsertImmediateToReg(ARMEX_SaveHistory, address);
	InsertDispToRegRM(ARMEX_SaveHistory_Copy, address);
	Insert(dest, ARMEX_SaveHistory, ARMEX_SaveHistory_End);
}

/////////////////////////////////////////////////////////////////////////////
//  Effect a memory transfer between the two locations (registers; we do no
// physical address conversion here).
void ARM::MemoryTransfer(ARM::DecodedInstruction & dest,
						 u32 d,
						 u32 s) {
	InsertDispToAcc(ARMEX_MemTransfer_Source, s);
	InsertDispToRegRM(ARMEX_MemTransfer_Copy, d);
	InsertImmediateToReg(ARMEX_MemTransfer_Dest, d);
	Insert(dest, ARMEX_MemTransfer, ARMEX_MemTransfer_End);
}

/////////////////////////////////////////////////////////////////////////////
//  Effect a memory transfer with no history.
void ARM::MemoryTransferNoHistory(ARM::DecodedInstruction & dest,
								  u32 d,
								  u32 s) {
	InsertDispToAcc(ARMEX_MemTransferNoHistory, s);
	InsertDispToAcc(ARMEX_MemTransferNoHistory_Dest, d);
	Insert(dest, ARMEX_MemTransferNoHistory, ARMEX_MemTransferNoHistory_End);
}

/////////////////////////////////////////////////////////////////////////////
//  Write the indicated value to the indicated register.
void ARM::MemoryStore(ARM::DecodedInstruction & dest,
					  u32 d,
					  u32 s) {
	InsertImmediateToReg(ARMEX_MemStore_Dest, d);
	InsertDispToRegRM(ARMEX_MemStore_DestCopy, d);
	InsertImmediateToRMNoDisp(ARMEX_MemStore_Source, s);
	Insert(dest, ARMEX_MemStore, ARMEX_MemStore_End);
}

/////////////////////////////////////////////////////////////////////////////
//  Extract the immediate operand from op2.
u32 ARM::GetImmediateFromOp2(u32 u) {
	if ((u >> 26) & 1) {
		u32 retval = u & 0xfff;

		if (retval & 0x800) {
			retval |= 0xfffff800;
		}
		return retval;
	} else {
		U8 r = ((u >> 8) & 0xf) * 2;
		U32 val = u & 0xff;
	
		return Ror(val, r);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Get the address of the jump displacement.
U16 * ARM::GetJumpDisplacementAddress(U8 * u) {
	return (U16 *)(u + 2);
}

/////////////////////////////////////////////////////////////////////////////
//  These return the addresses of the desired fields.
U32 * ARM::GetDisplacementToRMImm(U8 * u) {
	return (U32 *)(u + 2);
}

/////////////////////////////////////////////////////////////////////////////
//  Get the address of the immediate value.
U32 * ARM::GetImmediateToAcc(U8 * u) {
	return (U32 *)(u + 1);
}

/////////////////////////////////////////////////////////////////////////////
//  A stub for instructions that have not yet been implemented.
void ARM::DecDefaultStub(u32 address,
						 u32 code,
						 ARM::DecodedInstruction & dest,
						 u32 extraExceptions,
						 u32 cycles) {
	u32 exceptions = EXC_ATOMIC |
					 EXC_RETIRED |
					 EXC_CYCLE_BREAK |
					 s_forceInclude |
					 extraExceptions;
	U16 * d;
	u32 leftOff;

	d = InsertConditionCodeCheck(address, code >> 28, dest, leftOff);
	InsertConditionCodePost(dest,
							d,
							leftOff,
							exceptions,
							address,
							cycles);
}

/////////////////////////////////////////////////////////////////////////////
//  Our basic data processing operations.  There's quite a lot to do here!
void ARM::DecDataProcessing(u32 address,
							u32 code,
							ARM::DecodedInstruction & dest) {
	u32 exceptions = EXC_ATOMIC |
					 EXC_RETIRED |
					 EXC_CYCLE_BREAK |
					 s_forceInclude;
	u8 flags = code >> 28;
	U16 * d;
	u32 leftOff;
	u32 oldAddress = address;
	u8 sType;
	Bool overwriteCarry = TRUE;
	u32 cycles;

	// First, let's figure out if we're doing one of our MSR instructions.
	d = InsertConditionCodeCheck(address, flags, dest, leftOff);
	if ((code & 0x0fbf0fff) == 0x010f0000) {			// MRS.
		SaveHistory(dest, GetRegAddress((code >> 12) & 0xf));
		if ((code >> 22) & 1) {
			if (GetMode()) {
				MemoryTransfer(dest,
							   GetRegAddress((code >> 12) & 0xf),
							   GetRegAddress(SPSR));
			} else {
				MemoryTransfer(dest,
							   GetRegAddress((code >> 12) & 0xf),
							   GetRegAddress(CPSR));
			}
		} else {
			MemoryTransfer(dest,
						   GetRegAddress((code >> 12) & 0xf),
						   GetRegAddress(CPSR));
		}
		cycles = GetCycles(address, 0, 0, 1);
	} else if ((code & 0x0fb7fff0) == 0x0121f000) {		// MSR(priviliged)
		if ((code >> 22) & 1) {
			SaveHistory(dest, GetRegAddress(SPSR));
			MemoryTransfer(dest,
						   GetRegAddress(SPSR),
						   GetRegAddress((code >> 0) & 0xf));
		} else {
			SaveHistory(dest, GetRegAddress(CPSR));
			MemoryTransfer(dest,
						   GetRegAddress(CPSR),
						   GetRegAddress((code >> 0) & 0xf));
			exceptions |= EXC_MODE_CHANGE;
		}
		cycles = GetCycles(address, 0, 0, 1);
	} else if ((code & 0x0dbff000) == 0x0128f000) {		// MSR(user)
		u8 operand;

		if ((code >> 22) & 1) {
			operand = SPSR;
		} else {
			operand = CPSR;
		}
		SaveHistory(dest, GetRegAddress(operand));
		//  Are we doing a memory transfer, or a memory store?
		if ((code >> 25) & 1) {
			InsertImmediateToReg(ARMEX_Operand2Immediate,
								 GetImmediateFromOp2(code));
			Insert(dest, ARMEX_Operand2Immediate, ARMEX_Operand2Immediate_End);
		} else {
			InsertDispToRegRM(ARMEX_Operand2Register,
							  GetRegAddress((code >> 0) & 0xf));
			Insert(dest, ARMEX_Operand2Register, ARMEX_Operand2Register_End);
		}
		//  Okay, that loads op2.  Now, let's apply the operation.
		Insert(dest, ARMEX_BDP_MSR, ARMEX_BDP_MSR_End);

		InsertDispToRegRM(ARMEX_BDP_StoreOp1, GetRegAddress(operand));
		Insert(dest, ARMEX_BDP_StoreOp1, ARMEX_BDP_StoreOp1_End);
		cycles = GetCycles(address, 0, 0, 1);
	} else {
		//  This is where we actually do the data-ops.
		Bool immOp = ((code >> 25) & 1);
		Bool sBitSet = ((code >> 20) & 1);
		Bool setConds;
		u8 opcode = ((code >> 21) & 0xf);
		Bool isThumb = GetState() == STATE_THUMB;
		Bool storeDest;
		Bool isLogical;
		Bool usesOp1;
		Bool storeOp2;
		u8 rd = ((code >> 12) & 0xf);
		u8 rn = ((code >> 16) & 0xf);
		u8 rm = ((code >> 0) & 0xf);
		Bool isUserBreak = ((rd == 0) && (rm == 0) && (!sBitSet));

		setConds = sBitSet && (rd != 0xf);
		storeDest = (opcode < 8) || (opcode > 0xb);
		isLogical = (opcode < 2) || (opcode == 8) || (opcode == 9) ||
					(opcode > 0xb);
		usesOp1 = (opcode != 0xd) && (opcode != 0xf);
		storeOp2 = (opcode == 3) || (opcode == 7) ||
				   (opcode == 0xd) || (opcode == 0xf);

		if (sBitSet) {
			SaveHistory(dest, GetRegAddress(CPSR));
		}
		if (storeDest) {
			SaveHistory(dest, GetRegAddress(rd & 0xf));
		}
		if (immOp) {
			InsertImmediateToReg(ARMEX_Operand2Immediate,
								 GetImmediateFromOp2(code));
			Insert(dest, ARMEX_Operand2Immediate, ARMEX_Operand2Immediate_End);
			isUserBreak = FALSE;
		} else {
			if (rm == 0xf) {
				u32 toLoad;

				if ((code >> 4) & 1) {
					if (isThumb) {
						toLoad = address + 8;		// HEREHEREHERE confirm.
					} else {
						toLoad = address + 12;
					}
				} else {
					u8 imm;

					if (isThumb) {
						toLoad = address + 4;		// HEREHEREHERE confirm.
					} else {
						toLoad = address + 8;
					}
					imm = (code >> 7) & 0x1f;
					sType = (code >> 5) & 3;
					switch(sType) {
						case 0:
							if (imm) {
								toLoad <<= imm;
							}
							break;
						case 1:
							if (imm) {
								toLoad >>= imm;
							}
							break;
						case 2:
							if (imm) {
								toLoad = ((S32)toLoad) >> imm;
							}
							break;
						case 3:
							if (imm) {
								toLoad = Ror(toLoad, imm);
							}
							break;
					}
				}
				InsertImmediateToReg(ARMEX_Operand2Immediate,
									 toLoad);
				Insert(dest,
					   ARMEX_Operand2Immediate,
					   ARMEX_Operand2Immediate_End);
			} else {
				InsertDispToRegRM(ARMEX_Operand2Register,
										  GetRegAddress(rm));
				Insert(dest, ARMEX_Operand2Register, ARMEX_Operand2Register_End);
			}

			// Are we shifting by an immediate, or by a register?
			if ((code >> 4) & 1) {		// Shift by register.
				u8 rs = ((code >> 8) & 0xf);

				isUserBreak = FALSE;
				if (rs == 0xf) {
					exceptions |= EXC_INVALID_REG;
				}
				sType = ((code >> 5) & 3);
				switch(sType) {
					case 0:
						if (isLogical && setConds) {
							InsertDispToRegRM(ARMEX_Operand2LSLRegisterCF,
													  GetRegAddress(rs));
							InsertDispToRegRM(ARMEX_Operand2LSLRegisterCF_GetCPSR,
													  GetRegAddress(CPSR));
							Insert(dest,
								   ARMEX_Operand2LSLRegisterCF,
								   ARMEX_Operand2LSLRegisterCF_End);
						} else {
							InsertDispToRegRM(ARMEX_Operand2LSLRegisterNC,
													  GetRegAddress(rs));
							Insert(dest,
								   ARMEX_Operand2LSLRegisterNC,
								   ARMEX_Operand2LSLRegisterNC_End);
						}
						break;
					case 1:
						if (isLogical && setConds) {
							InsertDispToRegRM(ARMEX_Operand2LSRRegisterCF,
													  GetRegAddress(rs));
							InsertDispToRegRM(ARMEX_Operand2LSRRegisterCF_GetCPSR,
													  GetRegAddress(CPSR));
							Insert(dest,
								   ARMEX_Operand2LSRRegisterCF,
								   ARMEX_Operand2LSRRegisterCF_End);
						} else {
							InsertDispToRegRM(ARMEX_Operand2LSRRegisterNC,
													  GetRegAddress(rs));
							Insert(dest,
								   ARMEX_Operand2LSRRegisterNC,
								   ARMEX_Operand2LSRRegisterNC_End);
						}
						break;
					case 2:
						if (isLogical && setConds) {
							InsertDispToRegRM(ARMEX_Operand2ASRRegisterCF,
													  GetRegAddress(rs));
							InsertDispToRegRM(ARMEX_Operand2ASRRegisterCF_GetCPSR,
													  GetRegAddress(CPSR));
							Insert(dest,
								   ARMEX_Operand2ASRRegisterCF,
								   ARMEX_Operand2ASRRegisterCF_End);
						} else {
							InsertDispToRegRM(ARMEX_Operand2ASRRegisterNC,
													  GetRegAddress(rs));
							Insert(dest,
								   ARMEX_Operand2ASRRegisterNC,
								   ARMEX_Operand2ASRRegisterNC_End);
						}
						break;
					case 3:
						if (isLogical && setConds) {
							InsertDispToRegRM(ARMEX_Operand2RORRegisterCF,
													  GetRegAddress(rs));
							InsertDispToRegRM(ARMEX_Operand2RORRegisterCF_GetCPSR,
													  GetRegAddress(CPSR));
							Insert(dest,
								   ARMEX_Operand2RORRegisterCF,
								   ARMEX_Operand2RORRegisterCF_End);
						} else {
							InsertDispToRegRM(ARMEX_Operand2RORRegisterNC,
													  GetRegAddress(rs));
							Insert(dest,
								   ARMEX_Operand2RORRegisterNC,
								   ARMEX_Operand2RORRegisterNC_End);
						}
						break;
				}
			} else if (rm != 0xf) {
				// Shift by immediate.
				u8 imm = ((code >> 7) & 0x1f);

				sType = ((code >> 5) & 3);

				if (imm || sType) {
					isUserBreak = FALSE;
				}
				switch(sType) {
					case 0:
						if (imm) {
							InsertImmediateToShift(ARMEX_Operand2LSL,
												   imm);
							Insert(dest,
								   ARMEX_Operand2LSL,
								   ARMEX_Operand2LSL_End);
						} else {
							overwriteCarry = FALSE;
						}
						break;
					case 1:
						if (imm) {
							InsertImmediateToShift(ARMEX_Operand2LSR,
												   imm);
							Insert(dest,
								   ARMEX_Operand2LSR,
								   ARMEX_Operand2LSR_End);
						}
						break;
					case 2:
						if (imm) {
							InsertImmediateToShift(ARMEX_Operand2ASR,
												   imm);
							Insert(dest,
								   ARMEX_Operand2ASR,
								   ARMEX_Operand2ASR_End);
						}
						break;
					case 3:
						if (imm) {
							InsertImmediateToShift(ARMEX_Operand2ROR,
												   imm);
							Insert(dest,
								   ARMEX_Operand2ROR,
								   ARMEX_Operand2ROR_End);
						} else {
							InsertDispToRegRM(ARMEX_RCR,
													  GetRegAddress(CPSR));
							Insert(dest,
								   ARMEX_RCR,
								   ARMEX_RCR_End);
						}
						break;
				}
				if (isLogical && setConds) {
					if (imm) {
						Insert(dest, ARMEX_SaveCarry, ARMEX_SaveCarry_End);
					} else {
						switch(sType) {
							case 0:
								InsertDispToRegRM(ARMEX_PreserveCarryFlag,
														  GetRegAddress(CPSR));
								Insert(dest,
									   ARMEX_PreserveCarryFlag,
									   ARMEX_PreserveCarryFlag_End);
								break;
							case 1:
								Insert(dest,
									   ARMEX_LSR0,
									   ARMEX_LSR0_End);
								break;
							case 2:
								Insert(dest,
									   ARMEX_ASR0,
									   ARMEX_ASR0_End);
								break;
							case 3:
								Insert(dest,
									   ARMEX_SaveCarry,
									   ARMEX_SaveCarry_End);
								break;
						}
					}
				}
			}
		}

		// Okay!  We've done our operand2 decoding.  Now, let's get operand
		// 1 into a register (if we need it).
		if (usesOp1) {
			if (rn != 0xf) {
				InsertDispToRegRM(ARMEX_RegisterLoad,
										  GetRegAddress(rn));
				Insert(dest,
					   ARMEX_RegisterLoad,
					   ARMEX_RegisterLoad_End);
			} else {
				if (GetState() == STATE_ARM) {
					if (immOp || !((code >> 4) & 1)) {
						InsertImmediateToReg(ARMEX_ImmediateLoad,
											 (address & -4) + 8);
					} else {
						InsertImmediateToReg(ARMEX_ImmediateLoad,
											 (address & -4) + 12);
					}
				} else {
					if (immOp || !((code >> 4) & 1)) {
						InsertImmediateToReg(ARMEX_ImmediateLoad,
											 (address & -4) + 4);	// HEREHEREHERE confirm.
					} else {
						InsertImmediateToReg(ARMEX_ImmediateLoad,
											 (address & -4) + 8);	// HEREHEREHERE confirm.
					}
				}
				Insert(dest,
					   ARMEX_ImmediateLoad,
					   ARMEX_ImmediateLoad_End);
			}
		}
		switch(opcode) {
			case 0:
				Insert(dest,
					   ARMEX_BDP_AND,
					   ARMEX_BDP_AND_End);
				break;
			case 1:
				Insert(dest,
					   ARMEX_BDP_EOR,
					   ARMEX_BDP_EOR_End);
				break;
			case 2:
				Insert(dest,
					   ARMEX_BDP_SUB,
					   ARMEX_BDP_SUB_End);
				break;
			case 3:
				Insert(dest,
					   ARMEX_BDP_RSB,
					   ARMEX_BDP_RSB_End);
				break;
			case 4:
				Insert(dest,
					   ARMEX_BDP_ADD,
					   ARMEX_BDP_ADD_End);
				break;
			case 5:
				InsertDispToRegRM(ARMEX_BDP_ADC,
										  GetRegAddress(CPSR));
				Insert(dest,
					   ARMEX_BDP_ADC,
					   ARMEX_BDP_ADC_End);
				break;
			case 6:
				InsertDispToRegRM(ARMEX_BDP_SBC,
										  GetRegAddress(CPSR));
				Insert(dest,
					   ARMEX_BDP_SBC,
					   ARMEX_BDP_SBC_End);
				break;
			case 7:
				InsertDispToRegRM(ARMEX_BDP_RSC,
										  GetRegAddress(CPSR));
				Insert(dest,
					   ARMEX_BDP_RSC,
					   ARMEX_BDP_RSC_End);
				break;
			case 8:
				Insert(dest,
					   ARMEX_BDP_AND,
					   ARMEX_BDP_AND_End);
				break;
			case 9:
				Insert(dest,
					   ARMEX_BDP_EOR,
					   ARMEX_BDP_EOR_End);
				break;
			case 10:
				Insert(dest,
					   ARMEX_BDP_SUB,
					   ARMEX_BDP_SUB_End);
				break;
			case 11:
				Insert(dest,
					   ARMEX_BDP_ADD,
					   ARMEX_BDP_ADD_End);
				break;
			case 12:
				Insert(dest,
					   ARMEX_BDP_ORR,
					   ARMEX_BDP_ORR_End);
				break;
			case 13:
				if (isUserBreak) {
					exceptions |= EXC_USER_MESSAGE;
				}
				if (setConds) {
					Insert(dest,
						   ARMEX_GetCCOnOp2,
						   ARMEX_GetCCOnOp2_End);
				}
				break;
			case 14:
				Insert(dest,
					   ARMEX_BDP_BIC,
					   ARMEX_BDP_BIC_End);
				break;
			case 15:
				Insert(dest,
					   ARMEX_BDP_MVN,
					   ARMEX_BDP_MVN_End);
				break;
		}

		// Okay!  Now we store our result (if that is to be the case) and
		// we store the condition codes (if that is to be the case!)
		if (storeDest) {
			if (storeOp2) {
				InsertDispToRegRM(ARMEX_BDP_StoreOp2,
										  GetRegAddress(rd));
				Insert(dest,
					   ARMEX_BDP_StoreOp2,
					   ARMEX_BDP_StoreOp2_End);
			} else {
				InsertDispToRegRM(ARMEX_BDP_StoreOp1,
										  GetRegAddress(rd));
				Insert(dest,
					   ARMEX_BDP_StoreOp1,
					   ARMEX_BDP_StoreOp1_End);
			}

			// If we just stored R15, then we flag an external jump.
			if (rd == 0xf) {
				exceptions |= EXC_EXTERNAL_JUMP;
			}
		}

		// Now, we modify the CPSR.  If we chose a logical operation, then
		// we use esi for the carry flag, otherwise we use results from the
		// last operation.
		if (setConds) {
			if (isLogical) {
				if (overwriteCarry) {
					InsertImmediateToAcc(ARMEX_SetLogicalConditionCodesWithCarry,
										 GetRegAddress(CPSR));
					Insert(dest,
						   ARMEX_SetLogicalConditionCodesWithCarry,
						   ARMEX_SetLogicalConditionCodesWithCarry_End);
				} else {
					InsertImmediateToAcc(ARMEX_SetLogicalConditionCodesWithoutCarry,
										 GetRegAddress(CPSR));
					Insert(dest,
						   ARMEX_SetLogicalConditionCodesWithoutCarry,
						   ARMEX_SetLogicalConditionCodesWithoutCarry_End);
				}
			} else {
				InsertImmediateToAcc(ARMEX_SetConditionCodes,
									 GetRegAddress(CPSR));
				Insert(dest,
					   ARMEX_SetConditionCodes,
					   ARMEX_SetConditionCodes_End);
			}
		} else if (sBitSet) {
			// Copy SPSR into CPSR.
			MemoryTransferNoHistory(dest,
									GetRegAddress(CPSR),
									GetRegAddress(SPSR));
			exceptions |= EXC_MODE_CHANGE;
		}
		// Now, because of some tomfoolery with the sequence of the
		// store operation, we have to mask the R15 bits AFTER the
		// transfer based on the thumb state.
		if (storeDest && (rd == 0xf)) {
			InsertDispToRegRM(ARMEX_R15Mask, GetRegAddress(CPSR));
			InsertDispToRegRM(ARMEX_R15Mask_R15, GetRegAddress(0xf));
			InsertDispToRegRM(ARMEX_R15Mask_R15Copy, GetRegAddress(0xf));
			Insert(dest,
				   ARMEX_R15Mask,
				   ARMEX_R15Mask_End);
		}
		if (immOp || !((code >> 4) & 1)) {
			if (rd == 0xf) {
				cycles = GetCycles(address, 1, 0, 2);
			} else {
				cycles = GetCycles(address, 0, 0, 1);
			}
		} else {
			if (rd == 0xf) {
				cycles = GetCycles(address, 1, 1, 2);
			} else {
				cycles = GetCycles(address, 0, 1, 1);
			}
		}
	}
	InsertConditionCodePost(dest,
							d,
							leftOff,
							exceptions,
							oldAddress,
							cycles);
}

/////////////////////////////////////////////////////////////////////////////
//  Multiply (with possible accumulate).
void ARM::DecMultiply(u32 address,
					  u32 code,
					  ARM::DecodedInstruction & dest) {
	u32 exceptions = EXC_ATOMIC |
					 EXC_RETIRED |
					 EXC_CYCLE_BREAK |
					 s_forceInclude;
	u8 flags = code >> 28;
	U16 * d;
	u32 leftOff;
	u32 oldAddress = address;
	Bool setConds = ((code >> 20) & 1);
	Bool acc = ((code >> 21) & 1);
	u8 rd = ((code >> 16) & 0xf);
	u8 rn = ((code >> 12) & 0xf);
	u8 rs = ((code >> 8) & 0xf);
	u8 rm = ((code >> 0) & 0xf);
	u32 cycles;

	if ((rd == 0xf) || (rm == 0xf) || (rs == 0xf) ||
		(acc && (rn == 0xf)) ||
		(rd == rm)) {
		exceptions |= EXC_INVALID_REG;
	}
	d = InsertConditionCodeCheck(address, flags, dest, leftOff);
	if (setConds) {
		SaveHistory(dest, GetRegAddress(CPSR));
	}
	SaveHistory(dest, GetRegAddress(rd));

	// Okay, let's do the multiplication.
	InsertDispToAcc(ARMEX_SetupMultiply,
							GetRegAddress(rs));
	InsertDispToRegRM(ARMEX_SetupMultiply_Op2,
							  GetRegAddress(rm));
	if (acc) {
		InsertDispToRegRM(ARMEX_SetupMultiply_NormalEnd,
								  GetRegAddress(rn));
		Insert(dest,
			   ARMEX_SetupMultiply,
			   ARMEX_SetupMultiply_End);
	} else {
		Insert(dest,
			   ARMEX_SetupMultiply,
			   ARMEX_SetupMultiply_NormalEnd);
	}
	Insert(dest,
		   ARMEX_MultiplyUnsigned,
		   ARMEX_MultiplyUnsigned_End);
	if (acc) {
		Insert(dest,
			   ARMEX_MultiplyAccumulate,
			   ARMEX_MultiplyAccumulate_End);
	}
	InsertDispToAcc(ARMEX_MultiplyDone,
							GetRegAddress(rd));
	Insert(dest,
		   ARMEX_MultiplyDone,
		   ARMEX_MultiplyDone_End);
	if (setConds) {
		InsertDispToRegRM(ARMEX_MultiplyUpdateFlags_GetCPSR,
								  GetRegAddress(CPSR));
		InsertDispToRegRM(ARMEX_MultiplyUpdateFlags_CPSRCopy,
								  GetRegAddress(CPSR));
		Insert(dest,
			   ARMEX_MultiplyUpdateFlags,
			   ARMEX_MultiplyUpdateFlags_End);
	}
	if (acc) {
		cycles = GetCycles(address, 0, 1, 1);
	} else {
		cycles = GetCycles(address, 0, 0, 1);
	}
	InsertConditionCodePost(dest,
							d,
							leftOff,
							exceptions,
							oldAddress,
							cycles,
							TRUE);
}

/////////////////////////////////////////////////////////////////////////////
//  Perform long multiplication.
void ARM::DecMultiplyLong(u32 address,
						  u32 code,
						  ARM::DecodedInstruction & dest) {
	u32 exceptions = EXC_ATOMIC |
					 EXC_RETIRED |
					 EXC_CYCLE_BREAK |
					 s_forceInclude;
	u8 flags = code >> 28;
	U16 * d;
	u32 leftOff;
	u32 oldAddress = address;
	Bool setConds = ((code >> 20) & 1);
	Bool acc = ((code >> 21) & 1);
	Bool uns = !((code >> 22) & 1);
	u8 rdhi = ((code >> 16) & 0xf);
	u8 rdlo = ((code >> 12) & 0xf);
	u8 rs = ((code >> 8) & 0xf);
	u8 rm = ((code >> 0) & 0xf);
	u32 cycles;

	if ((rdlo == 0xf) || (rm == 0xf) || (rs == 0xf) || (rdhi == 0xf) ||
		(rdlo == rm) || (rdhi == rm) || (rdlo == rdhi)) {
		exceptions |= EXC_INVALID_REG;
	}
	d = InsertConditionCodeCheck(address, flags, dest, leftOff);
	if (setConds) {
		SaveHistory(dest, GetRegAddress(CPSR));
	}
	SaveHistory(dest, GetRegAddress(rdlo));
	SaveHistory(dest, GetRegAddress(rdhi));

	// Okay, let's do the multiplication.
	InsertDispToAcc(ARMEX_SetupMultiply,
							  GetRegAddress(rs));
	InsertDispToRegRM(ARMEX_SetupMultiply_Op2,
							  GetRegAddress(rm));
	Insert(dest,
		   ARMEX_SetupMultiply,
		   ARMEX_SetupMultiply_NormalEnd);
	if (uns) {
		Insert(dest,
			   ARMEX_MultiplyUnsigned,
			   ARMEX_MultiplyUnsigned_ZeroCTEnd);
		Insert(dest,
			   ARMEX_MultiplyUnsigned_NormalCTEnd,
			   ARMEX_MultiplyUnsigned_End);
	} else {
		Insert(dest,
			   ARMEX_MultiplySigned,
			   ARMEX_MultiplySigned_End);
	}
	if (acc) {
		InsertDispToRegRM(ARMEX_LongMultiplyAccDone,
						  GetRegAddress(rdlo));
		InsertDispToRegRM(ARMEX_LongMultiplyAccDone_RDHI,
						  GetRegAddress(rdhi));
		Insert(dest,
			   ARMEX_LongMultiplyAccDone,
			   ARMEX_LongMultiplyAccDone_End);
	} else {
		InsertDispToAcc(ARMEX_LongMultiplyDone,
								GetRegAddress(rdlo));
		InsertDispToRegRM(ARMEX_LongMultiplyDone_RDHI,
								  GetRegAddress(rdhi));
		Insert(dest,
			   ARMEX_LongMultiplyDone,
			   ARMEX_LongMultiplyDone_End);
	}
	if (setConds) {
		InsertDispToRegRM(ARMEX_LongMultiplyUpdateFlags_GetCPSR,
								  GetRegAddress(CPSR));
		InsertDispToRegRM(ARMEX_LongMultiplyUpdateFlags_CPSRCopy,
								  GetRegAddress(CPSR));
		Insert(dest,
			   ARMEX_LongMultiplyUpdateFlags,
			   ARMEX_LongMultiplyUpdateFlags_End);
	}
	if (acc) {
		cycles = GetCycles(address, 0, 2, 1);
	} else {
		cycles = GetCycles(address, 0, 1, 1);
	}
	InsertConditionCodePost(dest,
							d,
							leftOff,
							exceptions,
							oldAddress,
							cycles,
							TRUE);
}

/////////////////////////////////////////////////////////////////////////////
//  Decode a single-data-swap instruction, it's an atomic exchange op.
void ARM::DecSingleDataSwap(u32 address,
							u32 code,
							ARM::DecodedInstruction & dest) {
	u8 flags = (code >> 28);
	U16 * d;
	u32 leftOff;
	u32 oldAddress = address;
	u32 exceptions = EXC_ATOMIC |
					 EXC_RETIRED |
			 	 	 EXC_READ_BREAKPOINT |
					 EXC_WRITE_BREAKPOINT |
					 EXC_CYCLE_BREAK |
					 s_forceInclude;
	u8 rd = ((code >> 12) & 0xf);
	u8 rn = ((code >> 16) & 0xf);
	u8 rm = ((code >> 0) & 0xf);
	Bool b = ((code >> 22) & 1);

	// I see no better way to do this than with two transfers.  If it becomes
	// an issue, we can optimize it later.
	d = InsertConditionCodeCheck(address, flags, dest, leftOff);
	if ((rd == 0xf) || (rn == 0xf) || (rm == 0xf)) {
		exceptions |= EXC_INVALID_REG;
	}
	EffectDataTransfer(address,
					   0,
					   dest,
					   exceptions,
					   TRUE,
					   TRUE,
					   TRUE,
					   FALSE,
					   TRUE,
					   FALSE,
					   rn,
					   rd,
					   0,
					   b ? TFR_UBYTE : TFR_WORD,
					   0,
					   0);
	EffectDataTransfer(address,
					   0,
					   dest,
					   exceptions,
					   TRUE,
					   TRUE,
					   TRUE,
					   FALSE,
					   FALSE,
					   FALSE,
					   rn,
					   rm,
					   0,
					   b ? TFR_UBYTE : TFR_WORD,
					   0,
					   0);
	Insert(dest, ARMEX_JoinCycles, ARMEX_JoinCycles_End);
	InsertConditionCodePost(dest,
							d,
							leftOff,
							exceptions,
							oldAddress,
							GetCycles(address, 1, 1, 0),
							TRUE);
}

/////////////////////////////////////////////////////////////////////////////
//  Do a branch-exchange.  This instruction permits transition to THUMB code.
void ARM::DecBX(u32 address,
				u32 code,
				ARM::DecodedInstruction & dest) {
	u8 flags = (code >> 28);
	U16 * d;
	u32 leftOff;
	u32 oldAddress = address;
	u32 exceptions = EXC_ATOMIC |
					 EXC_RETIRED |
					 EXC_CYCLE_BREAK |
					 EXC_STATE_CHANGE |
					 EXC_EXTERNAL_JUMP |
					 s_forceInclude;
	u8 rn = ((code >> 0) & 0xf);
	u32 cycles = GetCycles(address, 0, 0, 2);

	d = InsertConditionCodeCheck(address, flags, dest, leftOff);
	if (rn == 0xf) {
		if (address & 3) {
			flags |= EXC_INVALID_OPCODE;
		}
		cycles += GetCycles(address, 1, 0, 0);
		SaveHistory(dest, GetRegAddress(CPSR));
		// HEREHEREHERE confirm this.
		if (GetState() == STATE_THUMB) {
			MemoryStore(dest, GetRegAddress(0xf), (address & -4) + 4);
		} else {
			MemoryStore(dest, GetRegAddress(0xf), (address & -4) + 8);
		}
	} else {
		InsertDispToRegRM(ARMEX_GetBranchMemTime,
								  GetRegAddress(rn));
		if (GetInstructionSize() == 2) {
			InsertDispToSIBRegRM(ARMEX_GetBranchMemTime_WaitStates,
										 (u32)g_wordN);
		} else {
			InsertDispToSIBRegRM(ARMEX_GetBranchMemTime_WaitStates,
										 (u32)g_dwordN);
		}
		Insert(dest,
			   ARMEX_GetBranchMemTime,
			   ARMEX_GetBranchMemTime_End);
		SaveHistory(dest, GetRegAddress(CPSR));
		MemoryTransfer(dest, GetRegAddress(0xf), GetRegAddress(rn));
	}
	InsertConditionCodePost(dest,
							d,
							leftOff,
							exceptions,
							oldAddress,
							GetCycles(address, 0, 0, 2),
							rn != 0xf);
}

/////////////////////////////////////////////////////////////////////////////
//  A halfword load/store, using a register (rather than an immediate) as an
// offset.
void ARM::DecHalfwordReg(u32 address,
						 u32 code,
						 ARM::DecodedInstruction & dest) {
	u8 flags = (code >> 28);
	U16 * d;
	u32 leftOff;
	u32 oldAddress = address;
	u32 exceptions = EXC_ATOMIC |
					 EXC_RETIRED |
					 EXC_CYCLE_BREAK |
					 s_forceInclude;
	Bool pre = ((code >> 24) & 1);
	Bool up = ((code >> 23) & 1);
	Bool wb = ((code >> 21) & 1);
	Bool ld = ((code >> 20) & 1);
	u8 rn = ((code >> 16) & 0xf);
	u8 rd = ((code >> 12) & 0xf);
	u8 rm = ((code >> 0) & 0xf);
	u8 tType = ((code >> 5) & 0x3);
	u32 cycles;


	if (ld) {
		exceptions |= EXC_READ_BREAKPOINT;
	} else {
		exceptions |= EXC_WRITE_BREAKPOINT;
	}
	if (rm == 0xf) {
		exceptions |= EXC_INVALID_REG;
	}
	switch(tType) {
		case 0:
			exceptions |= EXC_INVALID_OPCODE;
			tType = TFR_WORD;
			break;
		case 1:
			tType = TFR_UHWORD;
			break;
		case 2:
			if (!ld) {
				exceptions |= EXC_INVALID_OPCODE;
			}
			tType = TFR_SBYTE;
			break;
		case 3:
			if (!ld) {
				exceptions |= EXC_INVALID_OPCODE;
			}
			tType = TFR_SHWORD;
			break;
	}
	d = InsertConditionCodeCheck(address, flags, dest, leftOff);
	EffectDataTransfer(address,
					   0,
					   dest,
					   exceptions,
					   FALSE,
					   pre,
					   up,
					   wb,
					   ld,
					   FALSE,
					   rn,
					   rd,
					   rm,
					   tType,
					   0,
					   0);
	if (ld) {
		if (rd == 0xf) {
			cycles = GetCycles(address, 2, 1, 1);
		} else {
			cycles = GetCycles(address, 1, 1, 0);
		}
	} else {
		cycles = GetCycles(address, 1, 0, 0);
	}
	InsertConditionCodePost(dest,
							d,
							leftOff,
							exceptions,
							oldAddress,
							cycles,
							TRUE);
}

/////////////////////////////////////////////////////////////////////////////
//  A halfword load/store involving an immediate offset.
void ARM::DecHalfwordImm(u32 address,
						 u32 code,
						 ARM::DecodedInstruction & dest) {
	u8 flags = (code >> 28);
	U16 * d;
	u32 leftOff;
	u32 oldAddress = address;
	u32 exceptions = EXC_ATOMIC |
					 EXC_RETIRED |
					 EXC_CYCLE_BREAK |
					 s_forceInclude;
	Bool pre = ((code >> 24) & 1);
	Bool up = ((code >> 23) & 1);
	Bool wb = ((code >> 21) & 1);
	Bool ld = ((code >> 20) & 1);
	u8 rn = ((code >> 16) & 0xf);
	u8 rd = ((code >> 12) & 0xf);
	u32 offset = ((code >> 4) & 0xf0) | ((code >> 0) & 0xf);
	u8 tType = ((code >> 5) & 0x3);
	u32 cycles;

	if (ld) {
		exceptions |= EXC_READ_BREAKPOINT;
	} else {
		exceptions |= EXC_WRITE_BREAKPOINT;
	}
	switch(tType) {
		case 0:
			exceptions |= EXC_INVALID_OPCODE;
			tType = TFR_WORD;
			break;
		case 1:
			tType = TFR_UHWORD;
			break;
		case 2:
			if (!ld) {
				exceptions |= EXC_INVALID_OPCODE;
			}
			tType = TFR_SBYTE;
			break;
		case 3:
			if (!ld) {
				exceptions |= EXC_INVALID_OPCODE;
			}
			tType = TFR_SHWORD;
			break;
	}
	d = InsertConditionCodeCheck(address, flags, dest, leftOff);
	EffectDataTransfer(address,
					   offset,
					   dest,
					   exceptions,
					   TRUE,
					   pre,
					   up,
					   wb,
					   ld,
					   FALSE,
					   rn,
					   rd,
					   0,
					   tType,
					   0,
					   0);
	if (ld) {
		if (rd == 0xf) {
			cycles = GetCycles(address, 2, 1, 1);
		} else {
			cycles = GetCycles(address, 1, 1, 0);
		}
	} else {
		cycles = GetCycles(address, 1, 0, 0);
	}
	InsertConditionCodePost(dest,
							d,
							leftOff,
							exceptions,
							oldAddress,
							cycles,
							TRUE);
}

/////////////////////////////////////////////////////////////////////////////
//  I am making this helper to enable me to implement all of the data-
// transfer operations with the same code.
// ConditionCode pre and post are handled by the caller.
void ARM::EffectDataTransfer(u32 address,
							 u32 offset,
							 ARM::DecodedInstruction & dest,
							 u32 & exceptions,
							 Bool imm,
							 Bool pre,
							 Bool up,
							 Bool wb,
							 Bool ld,
							 Bool user,
							 u8 rn,
							 u8 rd,
							 u8 rm,
							 u8 tType,
							 u8 sType,
							 u8 s,
							 Bool sequential) {
	u8 mode;

	if (user) {
		mode = 0;
	} else {
		mode = GetMode();
	}

	if (!pre && !imm && (rn == rm)) {
		exceptions |= EXC_ABORT_AMBIGUOUS;
	}

	if (ld) {
		SaveHistory(dest, GetRegAddress(rd, mode));
	}
	if (wb || !pre) {
		SaveHistory(dest, GetRegAddress(rn, mode));
	}

	// Let's start the address formation.
	if (rn == 0xf) {
		u32 insert;

		if (GetState() == STATE_ARM) {
			insert = address + 8;
		} else {
			insert = (address & -4) + 4;
		}

		if (wb) {
			exceptions |= EXC_INVALID_REG;
		}
		if (pre) {
			if (imm) {
				if (up) {
					insert += offset;
				} else {
					insert -= offset;
				}
			}
		}
		InsertImmediateToReg(ARMEX_ImmediateLoad,
							 insert);
		Insert(dest,
			   ARMEX_ImmediateLoad,
			   ARMEX_ImmediateLoad_End);
	} else {
		InsertDispToRegRM(ARMEX_RegisterLoad,
								  GetRegAddress(rn, mode));
		Insert(dest,
			   ARMEX_RegisterLoad,
			   ARMEX_RegisterLoad_End);
	}

	// We re-use our operand2 stuff here.
	if (pre) {
		if (imm) {
			if (rn != 0xf) {
				InsertImmediateToReg(ARMEX_Operand2Immediate,
									 offset);
				Insert(dest, ARMEX_Operand2Immediate, ARMEX_Operand2Immediate_End);
			}
		} else if (rm == 0xf) {
			exceptions |= EXC_INVALID_REG;
		} else {
			InsertDispToRegRM(ARMEX_Operand2Register,
									  GetRegAddress(rm, mode));
			Insert(dest, ARMEX_Operand2Register, ARMEX_Operand2Register_End);
			switch(sType) {
				case 0:
					if (s) {
						InsertImmediateToShift(ARMEX_Operand2LSL,
											   s);
						Insert(dest,
							   ARMEX_Operand2LSL,
							   ARMEX_Operand2LSL_End);
					}
					break;
				case 1:
					if (s) {
						InsertImmediateToShift(ARMEX_Operand2LSR,
											   s);
						Insert(dest,
							   ARMEX_Operand2LSR,
							   ARMEX_Operand2LSR_End);
					}
					break;
				case 2:
					if (s) {
						InsertImmediateToShift(ARMEX_Operand2ASR,
											   s);
						Insert(dest,
							   ARMEX_Operand2ASR,
							   ARMEX_Operand2ASR_End);
					}
					break;
				case 3:
					if (s) {
						InsertImmediateToShift(ARMEX_Operand2ROR,
											   s);
						Insert(dest,
							   ARMEX_Operand2ROR,
							   ARMEX_Operand2ROR_End);
					} else {
						Insert(dest,
							   ARMEX_RCR,
							   ARMEX_RCR_End);
					}
					break;
			}
		}
		if ((rn != 0xf) || !imm) {
			if (up) {
				Insert(dest,
					   ARMEX_BDP_ADD,
					   ARMEX_BDP_ADD_End);
			} else {
				Insert(dest,
					   ARMEX_BDP_SUB,
					   ARMEX_BDP_SUB_End);
			}
			if (wb) {
				InsertDispToRegRM(ARMEX_BDP_StoreOp1,
										  GetRegAddress(rn, mode));
				Insert(dest,
					   ARMEX_BDP_StoreOp1,
					   ARMEX_BDP_StoreOp1_End);
			}
		}
	}

	// Now we actually fetch the physical address.
	if (ld) {
		InsertImmediateToAcc(ARMEX_ReadMemory,
							 HISTORY_READ);
		InsertDispToSIBRegRM(ARMEX_ReadMemory_GetBounds,
									 (u32)g_readBounds);
		if (tType == TFR_WORD) {
			if (sequential) {
				InsertDispToSIBRegRM(ARMEX_ReadMemory_GetWaitStates,
											 (u32)g_dwordS);
			} else {
				InsertDispToSIBRegRM(ARMEX_ReadMemory_GetWaitStates,
											 (u32)g_dwordN);
			}
		} else {
			if (sequential) {
				InsertDispToSIBRegRM(ARMEX_ReadMemory_GetWaitStates,
											 (u32)g_wordS);
			} else {
				InsertDispToSIBRegRM(ARMEX_ReadMemory_GetWaitStates,
											 (u32)g_wordN);
			}
		}
		InsertDispToSIBRegRM(ARMEX_ReadMemory_GetBitvector,
									 (u32)g_readBPMap);
		InsertDispToSIBRegRM(ARMEX_ReadMemory_GetBaseAddress,
									 (u32)g_readMap);
		InsertImmediateToRMNoDisp(ARMEX_ReadMemory_BPType,
								  EXC_READ_BREAKPOINT);
		Insert(dest,
			   ARMEX_ReadMemory,
			   ARMEX_ReadMemory_End);
		if (tType == TFR_WORD) {
			Insert(dest,
				   ARMEX_MaskLowBits,
				   ARMEX_MaskLowBits_End);
		}

		// Read from memory.
		switch(tType) {
			case TFR_WORD:
				Insert(dest,
					   ARMEX_ReadWord,
					   ARMEX_ReadWord_End);
				break;
			case TFR_UHWORD:
				Insert(dest,
					   ARMEX_ReadHalfword,
					   ARMEX_ReadHalfword_End);
				break;
			case TFR_SHWORD:
				Insert(dest,
					   ARMEX_ReadSHalfword,
					   ARMEX_ReadSHalfword_End);
				break;
			case TFR_SBYTE:
				Insert(dest,
					   ARMEX_ReadSByte,
					   ARMEX_ReadSByte_End);
				break;
			case TFR_UBYTE:
				Insert(dest,
					   ARMEX_ReadByte,
					   ARMEX_ReadByte_End);
				break;
		}

		// Store the result.
		if (rd == 0xf) {
			exceptions |= EXC_EXTERNAL_JUMP;
			InsertImmediateToRMNoDisp(ARMEX_R15Op2_Mask, -GetInstructionSize());
			Insert(dest, ARMEX_R15Op2_Mask, ARMEX_R15Op2_Mask_End);
		}
		InsertDispToRegRM(ARMEX_BDP_StoreOp2,
								  GetRegAddress(rd, mode));
		Insert(dest,
			   ARMEX_BDP_StoreOp2,
			   ARMEX_BDP_StoreOp2_End);
	} else {
		InsertImmediateToAcc(ARMEX_ReadMemory,
							 HISTORY_WRITE);
		InsertDispToSIBRegRM(ARMEX_ReadMemory_GetBounds,
									 (u32)g_writeBounds);
		if (tType == TFR_WORD) {
			if (sequential) {
				InsertDispToSIBRegRM(ARMEX_ReadMemory_GetWaitStates,
											 (u32)g_dwordS);
			} else {
				InsertDispToSIBRegRM(ARMEX_ReadMemory_GetWaitStates,
											 (u32)g_dwordN);
			}
		} else {
			if (sequential) {
				InsertDispToSIBRegRM(ARMEX_ReadMemory_GetWaitStates,
											 (u32)g_wordS);
			} else {
				InsertDispToSIBRegRM(ARMEX_ReadMemory_GetWaitStates,
											 (u32)g_wordN);
			}
		}
		InsertDispToSIBRegRM(ARMEX_ReadMemory_GetBitvector,
									 (u32)g_writeBPMap);
		InsertDispToSIBRegRM(ARMEX_ReadMemory_GetBaseAddress,
									 (u32)g_writeMap);
		InsertImmediateToRMNoDisp(ARMEX_ReadMemory_BPType,
								  EXC_WRITE_BREAKPOINT);
		Insert(dest,
			   ARMEX_ReadMemory,
			   ARMEX_ReadMemory_End);
		if (tType == TFR_WORD) {
			Insert(dest,
				   ARMEX_MaskLowBits,
				   ARMEX_MaskLowBits_End);
		}

		// Now, we have to fetch the operand that we're putting in memory.
		if (rd == 0xf) {
			InsertImmediateToReg(ARMEX_Operand2Immediate,
								 address + 12);
			Insert(dest,
				   ARMEX_Operand2Immediate,
				   ARMEX_Operand2Immediate_End);
		} else {
			InsertDispToRegRM(ARMEX_Operand2Register,
									  GetRegAddress(rd, mode));
			Insert(dest,
				   ARMEX_Operand2Register,
				   ARMEX_Operand2Register_End);
		}
		switch(tType) {
			case TFR_SHWORD:		// These will throw exceptions anyway.
			case TFR_SBYTE:			// These will throw exceptions anyway.
			case TFR_WORD:
				Insert(dest,
					   ARMEX_SaveEBXNoMask,
					   ARMEX_SaveEBXNoMask_End);
				Insert(dest,
					   ARMEX_WriteWord,
					   ARMEX_WriteWord_End);
				break;
			case TFR_UHWORD:
				Insert(dest,
					   ARMEX_SaveEBXMask,
					   ARMEX_SaveEBXMask_End);
				Insert(dest,
					   ARMEX_WriteHalfword,
					   ARMEX_WriteHalfword_End);
				break;
			case TFR_UBYTE:
				Insert(dest,
					   ARMEX_SaveEBXMask,
					   ARMEX_SaveEBXMask_End);
				Insert(dest,
					   ARMEX_WriteByte,
					   ARMEX_WriteByte_End);
				break;
		}
	}
	if (!pre) {
		if (imm) {
			InsertImmediateToReg(ARMEX_Operand2Immediate,
								 offset);
			Insert(dest, ARMEX_Operand2Immediate, ARMEX_Operand2Immediate_End);
		} else {
			InsertDispToRegRM(ARMEX_Operand2Register,
									  GetRegAddress(rm, mode));
			Insert(dest, ARMEX_Operand2Register, ARMEX_Operand2Register_End);
			switch(sType) {
				case 0:
					if (s) {
						InsertImmediateToShift(ARMEX_Operand2LSL,
											   s);
						Insert(dest,
							   ARMEX_Operand2LSL,
							   ARMEX_Operand2LSL_End);
					}
					break;
				case 1:
					if (s) {
						InsertImmediateToShift(ARMEX_Operand2LSR,
											   s);
						Insert(dest,
							   ARMEX_Operand2LSR,
							   ARMEX_Operand2LSR_End);
					}
					break;
				case 2:
					if (s) {
						InsertImmediateToShift(ARMEX_Operand2ASR,
											   s);
						Insert(dest,
							   ARMEX_Operand2ASR,
							   ARMEX_Operand2ASR_End);
					}
					break;
				case 3:
					if (s) {
						InsertImmediateToShift(ARMEX_Operand2ROR,
											   s);
						Insert(dest,
							   ARMEX_Operand2ROR,
							   ARMEX_Operand2ROR_End);
					} else {
						Insert(dest,
							   ARMEX_RCR,
							   ARMEX_RCR_End);
					}
					break;
			}
		}
		if (up) {
			Insert(dest,
				   ARMEX_BDP_ADD,
				   ARMEX_BDP_ADD_End);
		} else {
			Insert(dest,
				   ARMEX_BDP_SUB,
				   ARMEX_BDP_SUB_End);
		}
		InsertDispToRegRM(ARMEX_BDP_StoreOp1,
								  GetRegAddress(rn, mode));
		Insert(dest,
			   ARMEX_BDP_StoreOp1,
			   ARMEX_BDP_StoreOp1_End);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This handles the LDR, STR instructions.
void ARM::DecSingleDataTransfer(u32 address,
								u32 code,
								ARM::DecodedInstruction & dest) {
	u32 exceptions = EXC_ATOMIC |
					 EXC_RETIRED |
					 EXC_CYCLE_BREAK |
					 s_forceInclude;
	u8 flags = code >> 28;
	Bool imm = !((code >> 25) & 1);
	Bool pre = ((code >> 24) & 1);
	Bool up = ((code >> 23) & 1);
	Bool b = ((code >> 22) & 1);
	Bool wb = ((code >> 21) & 1);
	Bool ld = ((code >> 20) & 1);
	Bool user = wb && !pre;
	u8 rn = ((code >> 16) & 0xf);
	u8 rd = ((code >> 12) & 0xf);
	u8 rm = ((code >> 0) & 0xf);
	U16 * d;
	u32 leftOff;
	u32 oldAddress = address;
	u32 offset = code & 0xfff;
	u8 sType = ((code >> 5) & 3);
	u8 s = ((code >> 7) & 0x1f);
	u32 cycles;

	if (ld) {
		exceptions |= EXC_READ_BREAKPOINT;
	} else {
		exceptions |= EXC_WRITE_BREAKPOINT;
	}
	// First, let's place the condition check in there.
	d = InsertConditionCodeCheck(address, flags, dest, leftOff);
	EffectDataTransfer(address,
					   offset,
					   dest,
					   exceptions,
					   imm,
					   pre,
					   up,
					   wb,
					   ld,
					   user,
					   rn,
					   rd,
					   rm,
					   b ? TFR_UBYTE : TFR_WORD,
					   sType,
					   s);
	if (ld) {
		if (rd == 0xf) {
			cycles = GetCycles(address, 2, 1, 1);
		} else {
			cycles = GetCycles(address, 1, 1, 0);
		}
	} else {
		cycles = GetCycles(address, 1, 0, 0);
	}
	InsertConditionCodePost(dest,
							d,
							leftOff,
							exceptions,
							oldAddress,
							cycles,
							TRUE);
}

/////////////////////////////////////////////////////////////////////////////
//  Undefined instruction of some sort.  The user will be given the choice
// of ignoring these, or signalling an exception.
void ARM::DecUndef(u32 address, u32 code, ARM::DecodedInstruction & dest) {
	DecDefaultStub(address,
				   code,
				   dest,
				   EXC_UNDEFINED,
				   GetCycles(address, 1, 1, 2));
}

/////////////////////////////////////////////////////////////////////////////
//  Decode a block-data-transfer.  These instructions are (apparently) non-
// atomic.
void ARM::DecBlockDataTransfer(u32 address,
							   u32 code,
							   ARM::DecodedInstruction & dest) {
	u8 regs[0x10];
	u32 exceptions = EXC_ATOMIC |
					 EXC_RETIRED |
					 EXC_CYCLE_BREAK |
					 s_forceInclude;
	u8 flags = code >> 28;
	Bool imm = TRUE;
	Bool pre = ((code >> 24) & 1);
	Bool up = ((code >> 23) & 1);
	Bool um = ((code >> 22) & 1);
	Bool wb = ((code >> 21) & 1);
	Bool ld = ((code >> 20) & 1);
	Bool user = !ld && um;
	u8 rn = ((code >> 16) & 0xf);
	U16 * d;
	u32 leftOff;
	u32 oldAddress = address;
	u32 offset = 4;
	U8 * jTable;
	MType m, n, o;
	u32 cycles;

	if (ld) {
		exceptions |= EXC_READ_BREAKPOINT;
	} else {
		exceptions |= EXC_WRITE_BREAKPOINT;
	}
	if (user && wb) {
		exceptions |= EXC_INVALID_OPCODE;
	}
	if (rn == 0xf) {
		exceptions |= EXC_INVALID_REG;
	}
	for (m=o=0, n=1; m<16; ++m, n+=n) {
		if (code & n) {
			regs[o++] = m;
		}
	}
	if (ld && ((code & (1 << rn)))) {
		wb = FALSE;
	}
	if (!o) {
		exceptions |= EXC_INVALID_OPCODE;
	}
	if (ld && (code & (1 << 0xf))) {
		exceptions |= EXC_EXTERNAL_JUMP;
	}

	//  The first thing we have to do is make a jump table that looks at
	// one of our hidden registers to see how far along we are.  Then, we
	// issue a sequence of data transfers, with special conditions on the
	// first, second, and last.
	//	- First:  We do the actual condition code test.  If the test passes,
	//			  we load the hidden register 2 with the base register plus
	//			  an appropriate offset that we will determine.
	//	- Second:  We do writeback for the base register, if necessary.
	//			   This will be implemented as hidden register 2 plus an
	//			   offset that we will determine.
	//	- All:  We do an unconditional transfer with hidden register 2
	//			and the target selected by this instruction.  HR2 gets
	//			updated.  HR1 gets incremented.  All transfers are done as
	//			post-increment operations.
	//	- Last:  Hidden register 1 is set to zero.
	if (o) {
		u32 initialJump;
		u32 final;

		// The theory:  because the CPSR cannot be modified within this
		// instruction until the very end, we can do the condition code
		// check before every iteration.  In practice, we'll find out!
		d = InsertConditionCodeCheck(address, flags, dest, leftOff);
		*(dest.jtables) <<= (jTable = InsertJumpTable(dest, HR1));
		((U32 *)jTable)[0] = (U32)(dest.base + dest.dest);
		for (m=0; m<o; ++m) {
			Bool lwb = TRUE;

			if (m) {
				((U32 *)jTable)[m] = (u32)(dest.dest + dest.base);
			}

			// First iteration.
			if (m == 0) {
				if (up) {
					if (pre) {
						initialJump = 4;
					} else {
						initialJump = 0;
					}
					final = o * 4;
				} else {
					if (pre) {
						initialJump = (u32)(o * -4);
					} else {
						initialJump = (u32)((o - 1) * -4);
					}
					final = 0;
				}
				if (o != 1)
					final -= 4;
				SaveHistory(dest, GetRegAddress(HR2));
				InsertImmediateToReg(ARMEX_Operand2Immediate,
									 initialJump);
				Insert(dest,
					   ARMEX_Operand2Immediate,
					   ARMEX_Operand2Immediate_End);
				InsertDispToRegRM(ARMEX_RegisterLoad,
										  GetRegAddress(rn));
				Insert(dest,
					   ARMEX_RegisterLoad,
					   ARMEX_RegisterLoad_End);
				Insert(dest,
					   ARMEX_BDP_ADD,
					   ARMEX_BDP_ADD_End);
				InsertDispToRegRM(ARMEX_BDP_StoreOp1,
										  GetRegAddress(HR2));
				Insert(dest,
					   ARMEX_BDP_StoreOp1,
					   ARMEX_BDP_StoreOp1_End);
			}

			// Second iteration.  If we only have one item to push/pop, we
			// do this on the first iteration (since the documentation does
			// not suggest what happens in this case).  Perhaps we should
			// artificially force a second iteration that only does write-
			// back, HEREHEREHERE this needs to be investigated.
			if ((m == 1) || (o == 1)) {
				// If we are not loading, or if we are loading but not
				// loading a new base register, we overwrite the base
				// register on the second iteration.
				if (wb) {
					SaveHistory(dest, GetRegAddress(rn));
					InsertImmediateToReg(ARMEX_Operand2Immediate,
										 final);
					Insert(dest,
						   ARMEX_Operand2Immediate,
						   ARMEX_Operand2Immediate_End);
					InsertDispToRegRM(ARMEX_RegisterLoad,
											  GetRegAddress(HR2));
					Insert(dest,
						   ARMEX_RegisterLoad,
						   ARMEX_RegisterLoad_End);
					Insert(dest,
						   ARMEX_BDP_ADD,
						   ARMEX_BDP_ADD_End);
					InsertDispToRegRM(ARMEX_BDP_StoreOp1,
											  GetRegAddress(rn));
					Insert(dest,
						   ARMEX_BDP_StoreOp1,
						   ARMEX_BDP_StoreOp1_End);
				}
			}
			// Final iteration.
			if (m == o - 1) {
				lwb = FALSE;

				// Update HR1.
				SaveHistory(dest, GetRegAddress(HR1));
				InsertImmediateToReg(ARMEX_Operand2Immediate,
									 0);
				Insert(dest,
					   ARMEX_Operand2Immediate,
					   ARMEX_Operand2Immediate_End);
				InsertDispToRegRM(ARMEX_BDP_StoreOp2,
										  GetRegAddress(HR1));
				Insert(dest,
					   ARMEX_BDP_StoreOp2,
					   ARMEX_BDP_StoreOp2_End);
			} else {
				// Update HR1.
				SaveHistory(dest, GetRegAddress(HR1));
				InsertImmediateToReg(ARMEX_Operand2Immediate,
									 4);
				Insert(dest,
					   ARMEX_Operand2Immediate,
					   ARMEX_Operand2Immediate_End);
				InsertDispToRegRM(ARMEX_RegisterLoad,
										  GetRegAddress(HR1));
				Insert(dest,
					   ARMEX_RegisterLoad,
					   ARMEX_RegisterLoad_End);
				Insert(dest,
					   ARMEX_BDP_ADD,
					   ARMEX_BDP_ADD_End);
				InsertDispToRegRM(ARMEX_BDP_StoreOp1,
										  GetRegAddress(HR1));
				Insert(dest,
					   ARMEX_BDP_StoreOp1,
					   ARMEX_BDP_StoreOp1_End);
			}
			EffectDataTransfer(address,
							   offset,
							   dest,
							   exceptions,
							   TRUE,
							   FALSE,
							   TRUE,
							   lwb,
							   ld,
							   user,
							   HR2,
							   regs[m],
							   0,
							   TFR_WORD,
							   0,
							   0,
							   m!=0);
			if (m != o - 1) {
				InsertExceptionCheck(dest,
									 exceptions & ~EXC_RETIRED,
									 oldAddress,
									 1);
			}
		}
		// HEREHEREHERE we have to find out how much time each iteration
		// takes, under the different circumstances.
		if (ld) {
			if (code & (1 << 0xf)) {
				cycles = GetCycles(address, 2, 1, 1);
			} else {
				cycles = GetCycles(address, 1, 1, 0);
			}
		} else {
			cycles = GetCycles(address, 1, 0, 0);
		}
		InsertConditionCodePost(dest,
								d,
								leftOff,
								exceptions,
								oldAddress,
								cycles,
								TRUE);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Decode a branch instruction.  First, we check condition (if necessary).
// If condition is met, we alter R14 and R15 accordingly.  Then we check for
// exceptions.
void ARM::DecBranch(u32 address,
					u32 code,
					ARM::DecodedInstruction & dest) {
	u32 exceptions;
	u8 flags = code >> 28;
	U16 * d;
	u32 leftOff;
	u32 disp;
	u32 oldAddress = address;
	Fixup fixup;
	Bool link = (code >> 24) & 1;
	u32 cycles;
	u8 s1 = (address >> 24) & 0xf;

	d = InsertConditionCodeCheck(address, flags, dest, leftOff);

	// If we need to do a link, we're doing a memory-transfer operation.
	if (link) {
		InsertImmediateToRMNoDisp(ARMEX_MemStore_Source,
								  address + 4);
		disp = GetRegAddress(14);
		InsertImmediateToReg(ARMEX_MemStore_Dest,
							 disp);
		InsertDispToRegRM(ARMEX_MemStore_DestCopy,
							   disp);
		Insert(dest, ARMEX_MemStore, ARMEX_MemStore_End);
	}

	// Here's where we execute the branch.
	if (GetState() == STATE_ARM) {
		address += 8;
		disp = (code & 0xffffff) * 4;
	} else {
		address += 4;
		disp = (code & 0xffffff) * 2;
	}
	if (disp & 0x800000) {
		disp |= 0xff000000;
	}
	address += disp;

	// Copy the new number into R15, while saving R15's history.
	InsertImmediateToRMNoDisp(ARMEX_MemStore_Source,
							  address);
	disp = GetRegAddress(15);
	InsertImmediateToReg(ARMEX_MemStore_Dest,
						 disp);
	InsertDispToRegRM(ARMEX_MemStore_DestCopy,
							  disp);
	Insert(dest, ARMEX_MemStore, ARMEX_MemStore_DoTransfer);

	cycles = GetCycles(oldAddress, 1, 0, 0);
	cycles += GetCycles(address, 0, 0, 2);
	InsertImmediateToReg(ARMEX_SetCycles,
						 cycles << CYCLE_SHIFT);
	Insert(dest, ARMEX_SetCycles, ARMEX_SetCycles_End);
	InsertTimerUpdateCode(dest);

	// Now, we do our exception checks.
	fixup.exceptionSpot1 = GetDisplacementToRMImm(ARMEX_Nop_ExceptionsRaised -
												  ARMEX_Nop +
												  dest.base +
												  dest.dest);
	fixup.exceptionSpot2 = GetImmediateToAcc(ARMEX_Nop_ExceptionsRaisedCopy -
											 ARMEX_Nop +
											 dest.base +
											 dest.dest);

	InsertImmediateToRM(ARMEX_Nop_StoreR15,
						address);
	exceptions = EXC_ATOMIC |
			 	 EXC_RETIRED |
			 	 EXC_CYCLE_BREAK |
			 	 EXC_EXTERNAL_JUMP |
			 	 s_forceInclude;
	
	InsertDispToRMImm(ARMEX_Nop_StoreR15, GetRegAddress(0xf));
	InsertImmediateToRMNoDisp(ARMEX_Nop_ExceptionsRaised, exceptions);
	InsertImmediateToAcc(ARMEX_Nop_ExceptionsRaisedCopy, exceptions);
	Insert(dest, ARMEX_Nop, ARMEX_Nop_End);

	// Okay, one last wrinkle:  we want to jump within the cache line if
	// we later determine that the cache line contains the target address.
	// How will we do this fix-up?  Easy!  We will keep a list of fix-ups
	// in the cache-line structure.
	fixup.dispSpot = dest.base;
	fixup.address = address;
	*(dest.fixups) <<= fixup;
	InsertImmediateToAcc(ARMEX_ABranchWithout, exceptions);
	Insert(dest, ARMEX_ABranchWithout, ARMEX_ABranchWithout_End);

	// If we did a conditional branch, we have to deal with the case where
	// the branch is not taken.
	exceptions &= ~EXC_EXTERNAL_JUMP;

	// HEREHEREHERE I'm adding 2 I-cycles to branch instructions to refill
	// the prefetch.
	InsertConditionCodePost(dest,
							d,
							leftOff,
							exceptions,
							oldAddress,
							cycles);
}

/////////////////////////////////////////////////////////////////////////////
//  Coprocessor ops.  As far as I know, no such device exists on GBA.
void ARM::DecCoprocDataTransfer(u32 address,
								u32 code,
								ARM::DecodedInstruction & dest) {
	DecDefaultStub(address,
				   code,
				   dest,
				   EXC_COPROCESSOR,
				   GetCycles(address, 2, 0, 0));		// HEREHEREHERE test behavior.
}

/////////////////////////////////////////////////////////////////////////////
//  As far as I know, there's no coprocessor on GBA.
void ARM::DecCoprocDataOp(u32 address,
						  u32 code,
						  ARM::DecodedInstruction & dest) {
	DecDefaultStub(address,
				   code,
				   dest,
				   EXC_COPROCESSOR,
				   GetCycles(address, 0, 0, 1));		// HEREHEREHERE test behavior.
}

/////////////////////////////////////////////////////////////////////////////
//  As far as I know, there's no coprocessor on GBA.
void ARM::DecCoprocRegTransfer(u32 address,
							   u32 code,
							   ARM::DecodedInstruction & dest) {
	DecDefaultStub(address,
				   code,
				   dest,
				   EXC_COPROCESSOR,
				   GetCycles(address, 0, 0, 1));
}

/////////////////////////////////////////////////////////////////////////////
//  Software interrupts.
void ARM::DecSWI(u32 address,
				 u32 code,
				 ARM::DecodedInstruction & dest) {
	DecDefaultStub(address,
				   code,
				   dest,
				   EXC_SWI,
				   GetCycles(address, 1, 0, 2));
}

/////////////////////////////////////////////////////////////////////////////
//  Another undefined. . . .
void ARM::DecUndefined(u32 address,
					   u32 code,
					   ARM::DecodedInstruction & dest) {
	DecDefaultStub(address,
				   code,
				   dest,
				   EXC_UNDEFINED,
				   GetCycles(address, 1, 1, 2));
}

