/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	The implementation of our breakpoint support.

****************************************************************************/

#include "gba.h"

List < Breakpoint > g_userBreakpoints;
static U16 s_eeprom[0x200];
static bool s_hereherehere;

/////////////////////////////////////////////////////////////////////////////
//  This will set all of the breakpoints that are used internally.
void BreakpointSetAllInternal(void) {
if (!s_hereherehere) {
	memset(s_eeprom, -1, sizeof(s_eeprom));
	s_hereherehere = TRUE;
}
	// DMA.
	BreakpointSet(0x40000b8, 2, 1, BP_WRITE, BP_INTERNAL);
	BreakpointSet(0x40000c4, 2, 1, BP_WRITE, BP_INTERNAL);
	BreakpointSet(0x40000d0, 2, 1, BP_WRITE, BP_INTERNAL);
	BreakpointSet(0x40000dc, 2, 1, BP_WRITE, BP_INTERNAL);

	// IRQ flags.
	BreakpointSet(0x4000202, 2, 1, BP_WRITE, BP_INTERNAL);

	// Timers.
	BreakpointSet(0x4000102, 2, 1, BP_WRITE, BP_INTERNAL);
	BreakpointSet(0x4000106, 2, 1, BP_WRITE, BP_INTERNAL);
	BreakpointSet(0x400010a, 2, 1, BP_WRITE, BP_INTERNAL);
	BreakpointSet(0x400010e, 2, 1, BP_WRITE, BP_INTERNAL);

	// EEPROM.
	BreakpointSet(0xd000010, 2, 1, BP_WRITE, BP_INTERNAL);
	BreakpointSet(0xd000090, 2, 1, BP_WRITE, BP_INTERNAL);
}

/////////////////////////////////////////////////////////////////////////////
//  Alter the indicated user breakpoint.
void BreakpointReset(MType index,
					 u32 base,
					 u32 range,
					 u32 countdown,
					 u8 t) {
	Breakpoint & b = g_userBreakpoints[index];

	// Clear the old one.
	BreakpointSet(b.base, b.range, b.countdown, 0, BP_USER);

	// Update the values.
	b.base = base;
	b.range = range;
	b.countdown = b.curCount = countdown;
	b.options = t;

	// Set the new one.
	BreakpointSet(base, range, countdown, t, BP_USER);
}

static TString s_boundsError("Breakpoint bounds error.");
static TString s_overlapError("This breakpoint overlaps another one.");

/////////////////////////////////////////////////////////////////////////////
//  Set a write breakpoint at the indicated address.
void BreakpointSet(u32 base, u32 range, u32 countdown, u8 t, u8 op) {
	u32 u;
	U8 * bpm;
	U8 * ubp;
	u32 seg = base >> 24;
	u32 offs;
	u32 bounds;
	u8 type;
	u32 of2;

	if (seg > 0xf) {
		WarnUser(s_boundsError);
		return;
	}
	offs = base & g_memMask[seg];
	of2 = range + offs;
	for (type=BP_READ; type<=BP_EXEC; type += type) {
		switch(type) {
			case BP_READ:
				bpm = g_readBPMap[seg];
				ubp = g_userRBPMap[seg];
				bounds = g_readBounds[seg];
				break;
			case BP_WRITE:
				bpm = g_writeBPMap[seg];
				ubp = g_userWBPMap[seg];
				bounds = g_writeBounds[seg];
				break;
			case BP_EXEC:
				bpm = g_codeBPMap[seg];
				ubp = g_userCBPMap[seg];
				bounds = g_readBounds[seg];
				break;
		}

		// Clip the interval to the permitted range.
		if (offs >= bounds) {
			if (bounds) {
				WarnUser(s_boundsError);
				return;
			} else {
				continue;
			}
		}
		if (of2 >= bounds) {
			of2 = bounds - 1;
		}

		for (u=offs; u<of2; u+=4) {
			u32 s = u >> 5;
			u32 o = 1 << ((u >> 2) & 0x7);
	
			if (t & type) {
				bpm[s] |= o;
				if (op == BP_USER) {
					ubp[s] |= o;
				}
			} else {
				bpm[s] &= ~o;
				if (op == BP_USER) {
					ubp[s] &= ~o;
				}
			}
		}
	}
	if (op == BP_USER) {
		if (t) {
			Breakpoint b;
	
			b.base = base;
			b.range = range;
			b.options = t;
			b.countdown = b.curCount = countdown;
			g_userBreakpoints |= b;
		}
		BreakpointSetAllInternal();
	}
	g_gba.GetCPU().InvalidateCodeCache();
}

/////////////////////////////////////////////////////////////////////////////
//  This removes from the exception mask all breakpoints that were triggered
// that were not user-specified, or that haven't counted down yet.
void BreakpointCheck(u32 ip, u32 address, u32 & exceptions) {
	if (exceptions & (EXC_READ_BREAKPOINT |
					  EXC_WRITE_BREAKPOINT |
					  EXC_EXEC_BREAKPOINT)) {
		u8 u;
		u32 exc;
		U8 * ubp;
		static U8 ** s_bps[] = {
			g_userRBPMap,
			g_userWBPMap,
			g_userCBPMap,
		};
		static U32 * s_bounds[] = {
			g_readBounds,
			g_writeBounds,
			g_readBounds
		};
		u8 seg;
		Bool cd = FALSE;
		Bool cdt = FALSE;
	
		for (u=0, exc=EXC_READ_BREAKPOINT;
			 exc<=EXC_EXEC_BREAKPOINT;
			 exc += exc, ++u) {
			u32 s;
			u32 o;
			u32 a;
	
			if (u == 2) {
				a = ip;
			} else {
				a = address;
			}
			seg = a >> 24;
			if (seg > 0xf)
				return;
			if ((a & g_memMask[seg]) >= s_bounds[u][seg]) {
				continue;
			}
			o = 1 << ((a >> 2) & 0x7);
			s = (a & g_memMask[seg]) >> 5;
			if (exceptions & exc) {
				ubp = s_bps[u][seg];
				if (exc == EXC_WRITE_BREAKPOINT) {
					if (a == 0x4000202) {
						u16 & irqs = g_gba.GetCPU().GetIRQs();
						MemoryDispatcher & m = g_gba.GetMemory();
	
						irqs &= ~m.GetW(0x4000202);
						m.SetW(0x4000202, irqs);
					} else if ((a == 0x4000102) ||
							   (a == 0x4000106) ||
							   (a == 0x400010a) ||
							   (a == 0x400010e)) {
						u16 tct;

						// We used to hardcode our timer behavior, but it
						// makes much more sense to generate it dynamically,
						// so we don't have to invalidate cache all the time
						// (and don't burn so much extra memory!).
						g_gba.GetCPU().MakeTimerUpdateCode();

						// The instruction that brought us here was a write
						// to the timer control register.  This will equal
						// the cost of the code address plus the cost of
						// I/O ram.  Let's factor that into the initial
						// value.  HEREHEREHERE this will cause a problem
						// if the enabled bit is being set more than once,
						// we have to cache a copy of the old value for 100%
						// accuracy.
						tct = g_gba.GetMemory().GetW(a);
						if ((tct & 0x80) && ((a == 0x4000102) || !(tct & 0x4))) {
							u32 cycles = g_wordS[(ip >> 24) & 0xf] +
										 g_wordS[(address >> 24) & 0xf];
							u8 shift;

							cycles = (cycles >> CYCLE_SHIFT) + 3;
							switch(tct & 3) {
								case 0:
									tct = g_gba.GetMemory().GetW(a - 2);
									tct += cycles;
									g_gba.GetMemory().SetW(a - 2, tct);
									break;
								case 1:
								case 2:
								case 3:
									shift = 26 - 2 * (tct & 3);
									tct = (a - 0x4000102) / 2;
									g_internalCounters[tct] += (cycles << shift);
									break;
							}
						}
					} else if (a == 0xd000010) {
						u32 cmd = g_gba.GetMemory().GetD(0xd000000);
						u32 value;

						cmd = ((cmd & 0x10000) >> 15) + (cmd & 1);
						if (cmd == 3) {
							u32 daddr = 0xd000004;

							// Get the address.
							for (value=0, cmd=0; cmd<6; ++cmd) {
								value += value;
								value += g_gba.GetMemory().GetW(0xd000004 + cmd * 2) & 1;
							}

							// Copy the requested data into the buffer.
							value *= 4;

							for (cmd=0; cmd<4; ++cmd) {
								u16 v = s_eeprom[value];
								u8 w;

								for (w=0; w<16; ++w) {
									if (v & 0x8000) {
										g_gba.GetMemory().SetW(daddr, 1);
									} else {
										g_gba.GetMemory().SetW(daddr, 0);
									}
									v <<= 1;
									daddr += 2;
								}
								++value;
							}
						}
					} else if (a == 0xd000090) {
						u32 cmd = g_gba.GetMemory().GetD(0xd000000);
						u32 value;

						cmd = ((cmd & 0x10000) >> 15) + (cmd & 1);
						if (cmd == 1) {
							u32 daddr = 0xd000004;

							// Get the address.
							for (value=0, cmd=0; cmd<6; ++cmd) {
								value += value;
								value += g_gba.GetMemory().GetW(daddr) & 1;
								daddr += 2;
							}

							// Copy.
							value *= 4;
							for (cmd=0; cmd<4; ++cmd) {
								u8 w;
								u16 v = 0;

								for (w=0; w<16; ++w) {
									v += v;
									v += g_gba.GetMemory().GetW(daddr) & 1;
									daddr += 2;
								}
								s_eeprom[value] = v;
								++value;
							}
//							g_gba.GetMemory().SetD(0xd000000, 0);
						}
					}
				}
				if (!(ubp[s] & o)) {
					exceptions &= ~exc;
				} else if (!cd) {
					MType m;
	
					for (m=0; m<g_userBreakpoints.Count(); ++m) {
						if (a - g_userBreakpoints[m].base < 
							g_userBreakpoints[m].range) {
							if (--g_userBreakpoints[m].curCount) {
								exceptions &= ~exc;
								cdt = TRUE;
							} else {
								g_userBreakpoints[m].curCount = 1;
							}
							break;
						}
					}
					if (m == g_userBreakpoints.Count()) {
						exceptions &= ~exc;
						cdt = TRUE;
					}
					cd = TRUE;
				} else if (cdt) {
					exceptions &= ~exc;
				}
				if (exc == EXC_EXEC_BREAKPOINT) {
					if (a == g_retiredBP) {
						if (!g_branchDepth) {
							exceptions |= exc;
						} else {
							--g_branchDepth;
						}
					} else if (a == g_branchBP) {
						++g_branchDepth;
					}
					g_executeOffset = ARMEX_CodeBreak_End - ARMEX_CodeBreak;
				}
			}
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Check if there's a code breakpoint at the specified address.
Bool BreakpointCode(u32 address) {
	u32 s;
	u32 o = 1 << ((address >> 2) & 0x7);
	u8 seg = address >> 24;

	if (seg > 0xf)
		return FALSE;
	if ((address & g_memMask[seg]) >= g_readBounds[seg])
		return FALSE;
	s = (address & g_memMask[seg]) >> 5;
	return g_codeBPMap[seg][s] & o;
}

/////////////////////////////////////////////////////////////////////////////
//  A (slow) routine to check if the user has set a breakpoint on the
// indicated address.
MType BreakpointIsActive(u32 address, u8 flags) {
	MType m;

	for (m=0; m<g_userBreakpoints.Count(); ++m) {
		Breakpoint & b = g_userBreakpoints[m];

		if (address - b.base < b.range) {
			if (b.options & flags) {
				return m + 1;
			}
		}
	}
	return 0;
}

/////////////////////////////////////////////////////////////////////////////
//  Clear the indicated user breakpoint.
void BreakpointClearUser(MType m) {
	CHECK(m < g_userBreakpoints.Count());
	Breakpoint & b = g_userBreakpoints[m];

	BreakpointSet(b.base, b.range, 0, 0, BP_USER);
	g_userBreakpoints -= b;
}

/////////////////////////////////////////////////////////////////////////////
//  Toggle the breakpoint at the cursor.  We will want to make this smarter
// downstream.
void BreakpointToggle(u32 address) {
	u32 bp;

	if (address != 0xffffffff) {
		if (bp = BreakpointIsActive(address, BP_EXEC)) {
			BreakpointClearUser(--bp);
		} else {
			u8 range = (g_gba.GetCPU().GetState() == ARM::STATE_ARM) ?
					   4 :
					   2;

			BreakpointSet(address,
						  range,
						  1,
						  BP_EXEC,
						  BP_USER);
		}
	}
	g_codeWindow->GenerateContent(g_codeWindow->GetTab());
	g_dataWindow->GenerateContent(DataWindow::TAB_BREAKPOINTS);
}

