/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	The implementation of the DMA chip.

****************************************************************************/

#include "gba.h"

/////////////////////////////////////////////////////////////////////////////
//  Make a DMA chip.
DMA::DMA(void) {
	u8 u;

	for (u=0; u<4; ++u) {
		m_count[u] = 0;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Update the DMA processor.  HEREHEREHERE the following items need to be
// added:
//	- Accurate cycle times
//	- History saving.  We may want to add something to the history logic
//	  to keep track of who's effecting a change.
//	Return TRUE if I do NOT own the bus, FALSE otherwise.
Bool DMA::Tick(u8 condition) {
	MemoryDispatcher & m = g_gba.GetMemory();
	static u32 s_masks[] = {
		0x7fff,
		0x7fff,
		0x7fff,
		0x1ffff,
	};
	u8 v;
	Bool vid;
	Bool retval = TRUE;

	if (g_ticksTilNextVideo < 0x7ff) {
		vid = TRUE;
		g_clocksTilNext = g_ticksTilNextVideo;
	} else {
		vid = FALSE;
		g_clocksTilNext = 0x7ff;
	}
	g_gbaTicks += g_clocksTilNext;

	for (v=0; v<4; ++v) {
		if (m_count[v]) {
			if (!condition) {
				// Here's where we actually do the transfers.  We have to update
				// the whole system, as we do with ARM code.
				retval = FALSE;
				while (m_count[v]) {
					u32 ss = m_src[v];
					u32 dd = m_dest[v];
					u8 sa = (dd >> 24) & 0xf;
					u8 sb = (ss >> 24) & 0xf;
					u8 mask;
					u32 cycleCost;
					Bool breakout = FALSE;

					g_history[g_front++] = HISTORY_DMA;
					g_front &= HISTORY_BUFFER_SIZE / 4 - 1;
					g_history[g_front++] = v;
					g_front &= HISTORY_BUFFER_SIZE / 4 - 1;
					g_history[g_front++] = dd & -4;
					g_front &= HISTORY_BUFFER_SIZE / 4 - 1;
					g_history[g_front++] = m.GetD(dd & -4);
					g_front &= HISTORY_BUFFER_SIZE / 4 - 1;
					if (m_size[v] == 2) {
						cycleCost = ((g_wordS[sb] + g_wordS[sa]) >> CYCLE_SHIFT);
						m.SetW(dd, m.GetW(ss));
					} else {
						cycleCost = ((g_dwordS[sb] + g_dwordS[sa]) >> CYCLE_SHIFT);
						m.SetD(dd, m.GetD(ss));
					}
					g_clocksTilNext -= cycleCost;
					IncrementTimers(cycleCost);

					m_dest[v] += m_incd[v];
					m_src[v] += m_incs[v];
					if (!--m_count[v]) {
						u32 cnt = m.GetD(v * 12 + DMA0_CONTROL);
		
						// HEREHEREHERE is this correct?
						if (v) {
							m.SetD(v * 12 + DMA0_DEST, m_dest[v]);
						}
						m.SetD(v * 12 + DMA0_SOURCE, m_src[v]);
						if (!m_repeat[v]) {
							m.SetD(v * 12 + DMA0_CONTROL,
								   (m.GetD(v * 12 + DMA0_CONTROL) & ~DMA_ENABLE));
						}
						if (cnt & DMA_IE) {
							g_gba.GetCPU().IRQ(ARM::IRQ_DMA0 << v);
						}
					}

					// Fine.  Now, we have to check if we've hit any data
					// breakpoints, or if we have run out of time.  In either
					// case, we set the exception mask accordingly.
					g_exceptions = EXC_ATOMIC;
					if (vid) {
						g_ticksTilNextVideo = g_clocksTilNext;
						if (g_clocksTilNext & 0x80000000) {
							g_exceptions |= EXC_CYCLE_BREAK;
							g_gba.GetVideo().Tick();
						}
					} else if (g_clocksTilNext & 0x80000000) {
						g_exceptions |= EXC_CYCLE_BREAK;
					}
					mask = 1 << ((ss >> 2) & 0x7);
					g_exceptions &= g_exceptionMask;
					if (g_readBPMap[sb][(ss & g_memMask[sb]) >> 5] & mask) {
						breakout = TRUE;
						g_exceptions |= EXC_READ_BREAKPOINT;
						BreakpointCheck(g_gba.GetCPU().GetCurrentIP(),
										ss,
										g_exceptions);
					}
					mask = 1 << ((dd >> 2) & 0x7);
					if (g_writeBPMap[sa][(dd & g_memMask[sa]) >> 5] & mask) {
						breakout = TRUE;
						g_exceptions |= EXC_WRITE_BREAKPOINT;
						BreakpointCheck(g_gba.GetCPU().GetCurrentIP(),
										dd,
										g_exceptions);
					}
					if (g_exceptions || breakout) {
						goto allDone;
					}
				}
			}
			retval = TRUE;
		} else {
			u32 u;
	
			u = m.GetD(v * 12 + DMA0_CONTROL);
			m_mode[v] = (u >> 28) & 3;
			if ((u & DMA_ENABLE) &&
				(u & s_masks[v]) &&
				(!m_mode[v] || (m_mode[v] == condition))) {
				u8 a = (u >> 21) & 3;
				u8 b = (u >> 23) & 3;
				Bool s = (u & DMA_XFER32) ? TRUE : FALSE;
	
				m_repeat[v] = (u >> 25) & 1;
				m_src[v] = m.GetD(v * 12 + DMA0_SOURCE);
				m_dest[v] = m.GetD(v * 12 + DMA0_DEST);
				m_count[v] = u & s_masks[v];		// How many bits?
				m_size[v] = s ? 4 : 2;
				m_src[v] &= -m_size[v];
				m_dest[v] &= -m_size[v];
				switch(a) {
					case 0:
					case 3:
						m_incd[v] = m_size[v];
						break;
					case 1:
						m_incd[v] = (u32) -((s32)m_size[v]);
						break;
					case 2:
						m_incd[v] = 0;
						break;
				}
				switch(b) {
					case 0:
						m_incs[v] = m_size[v];
						break;
					case 1:
						m_incs[v] = (u32) -((s32)m_size[v]);
						break;
					case 2:
					case 3:
						m_incs[v] = 0;
						break;
				}
				if (!condition)
					--v;
			}
		}
	}
allDone:
	g_gbaTicks -= g_clocksTilNext;
	return retval;
}
