/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	g_exceptionMask is the list of exceptions that we want to always inter-
rupt the emulator with.
	g_exceptions is those exceptions that caused the emulator to stop.
	EXC_NORMAL is a list of exceptions that should always be enabled.
	EXC_USER is a list of exceptions that should throw us back into the
debugger.

****************************************************************************/

#include "gba.h"

extern "C" {
	U8 g_maskTable[256];
	u32 g_exceptions;
	u32 g_exceptionMask;
	u32 g_instructionStep;
	U32 g_badAddress;		// This address caused an exception.
	U32 g_clocksTilNext;	// How many clocks before we take a break.
};

Bool g_execute;

//  How many instructions to execute in a burst in app mode.
#define INSTRUCTIONS_PER_CYCLE			32

char * g_appName = "AGD";
char * g_appTitle = "AGD Emulator / Debugger Version "VERSION;

FontHandle g_font;
Cursor g_mouse;

color g_colors[NUMBER_OF_TEXT_COLORS];

static TString s_fontFile("eng.bmp");
static TString s_getCodeAddress("Specify the symbol or address to jump to.");
static TString s_enterBreakpoint("Specify the breakpoint you'd like to set.");
static TString s_getDisassemblyRange("Specify the range of addresses you wish to disassemble.");
static TString s_getDumpRange("Specify the range of addresses you wish to dump.");
static TString s_unregistered("Unregistered copy.");
static TString s_copyright("Copyright 2001 by John Panettiere, all rights reserved.");
static TString s_registrationLevel("Read/Write breakpoints not available in unregistered version.");
static TString s_editMemory("Specify address and value of memory you wish to edit.");

static u8 s_mode = 0xff;

CodeWindow * g_codeWindow;
DataWindow * g_dataWindow;
AppWindow * g_appWindow;

Sticky < U16 > g_userVideoWidth("Video Width", 640);
Sticky < U16 > g_userVideoHeight("Video Height", 480);

/////////////////////////////////////////////////////////////////////////////
//  When we are going to execute the emulator, with any number of
// instructions, we go to here.  This will handle some of the odd circum-
// stances surrounding code breakpoints.
static void SignalExecute(u8 orientation) {
	g_execute = TRUE;
	HistorySetOrientation(orientation);

	// Actually, if no history exists at this orientation.
	if (orientation == HISTORY_FORWARD) {
		if (BreakpointCode(g_gba.GetCPU().GetCurrentIP())) {
			g_executeOffset = ARMEX_CodeBreak_End - ARMEX_CodeBreak;
		} else {
			g_executeOffset = 0;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Do all shutdown that follows the main thread.
void ShutdownApp(void) {
	g_gba.ShutDown();
	SafeDelete(g_codeWindow);
	SafeDelete(g_dataWindow);
	SafeDelete(g_appWindow);
}

static TString s_videoMode("Unable to set a suitable video mode.");

/////////////////////////////////////////////////////////////////////////////
//  Do all initializations that precede the main thread.
void InitializeApp(void) {
	IString i;
	RenderOptions r;

	// Let's retrieve the user configuration options.
	g_userWidth = *g_userVideoWidth;
	g_userHeight = *g_userVideoHeight;
	g_userDepth = 2;
	if (!View::SetMode(g_userWidth, g_userHeight, g_userDepth, !*g_runAppWindow)) {
		g_userWidth = 640;
		g_userHeight = 480;
		if (!View::SetMode(g_userWidth, g_userHeight, g_userDepth, !*g_runAppWindow)) {
			Shutdown(s_videoMode);
		}
	}

	g_gba.GetVideo().Initialize();
	g_monitor.SetMode(g_userWidth,
					  g_userHeight,
					  g_userDepth,
					  !*g_runAppWindow);
	i = s_fontFile;
	g_font = TextLoadFixedFont(i,
							   8,
							   12,
							   256);
	g_windowFont = g_font;
	g_mouse.Import("mouse.bmp");
	r.options = Bitmap::CUTOUT;
	r.key = 0;
	g_mouse.SetOptions(r);

	g_colors[UNSELECTED] = g_monitor.MapColor(0.75, 0.75, 0.75);
	g_colors[SELECTED] = g_monitor.MapColor(1.0, 1.0, 0);
	g_colors[BORDER] = g_monitor.MapColor(0, 0, 1.0);
	g_colors[BACKGROUND] = g_monitor.MapColor(0, 0, 0);
	g_colors[TEXT] = g_monitor.MapColor(1.0, 0, 0);

	// This will generate all of the condition codes.
	MType a, b, c;

	for (c=a=0; c<0x10; ++c) {
		for (b=0; b<0x10; ++b, ++a) {
			switch(b) {
				case 0:
					if (c & (1 << 2)) {
						g_maskTable[a] = 0xff;
					} else {
						g_maskTable[a] = 0;
					}
					break;
				case 1:
					if (c & (1 << 2)) {
						g_maskTable[a] = 0;
					} else {
						g_maskTable[a] = 0xff;
					}
					break;
				case 2:
					if (c & (1 << 1)) {
						g_maskTable[a] = 0xff;
					} else {
						g_maskTable[a] = 0;
					}
					break;
				case 3:
					if (c & (1 << 1)) {
						g_maskTable[a] = 0;
					} else {
						g_maskTable[a] = 0xff;
					}
					break;
				case 4:
					if (c & (1 << 3)) {
						g_maskTable[a] = 0xff;
					} else {
						g_maskTable[a] = 0;
					}
					break;
				case 5:
					if (c & (1 << 3)) {
						g_maskTable[a] = 0;
					} else {
						g_maskTable[a] = 0xff;
					}
					break;
				case 6:
					if (c & (1 << 0)) {
						g_maskTable[a] = 0xff;
					} else {
						g_maskTable[a] = 0;
					}
					break;
				case 7:
					if (c & (1 << 0)) {
						g_maskTable[a] = 0;
					} else {
						g_maskTable[a] = 0xff;
					}
					break;
				case 8:
					if ((c & (1 << 1)) &&
						!(c & (1 << 2))) {
						g_maskTable[a] = 0xff;
					} else {
						g_maskTable[a] = 0;
					}
					break;
				case 9:
					if ((c & (1 << 1)) &&
						!(c & (1 << 2))) {
						g_maskTable[a] = 0;
					} else {
						g_maskTable[a] = 0xff;
					}
					break;
				case 10:
					if (((c >> 3) & 1) == ((c >> 0) & 1)) {
						g_maskTable[a] = 0xff;
					} else {
						g_maskTable[a] = 0;
					}
					break;
				case 11:
					if (((c >> 3) & 1) == ((c >> 0) & 1)) {
						g_maskTable[a] = 0;
					} else {
						g_maskTable[a] = 0xff;
					}
					break;
				case 12:
					if ((((c >> 3) & 1) == ((c >> 0) & 1)) &&
						!(c & (1 << 2))) {
						g_maskTable[a] = 0xff;
					} else {
						g_maskTable[a] = 0;
					}
					break;
				case 13:
					if ((((c >> 3) & 1) == ((c >> 0) & 1)) &&
						!(c & (1 << 2))) {
						g_maskTable[a] = 0;
					} else {
						g_maskTable[a] = 0xff;
					}
					break;
				case 14:
					g_maskTable[a] = 0xff;
					break;
				case 15:
					g_maskTable[a] = 0;
					break;
			}
		}
	}
	HistoryInitialize();
}

/////////////////////////////////////////////////////////////////////////////
//  This function handles a code "goto" requestor.
static void CodeGoto(void) {
	if (g_userString.Length() > 1) {
		u32 u = g_gba.GetAddress(g_userString);

		if (u != 0xffffffff) {
			g_codeWindow->Goto(u);
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This function handles a data "goto" requestor.
static void DataGoto(void) {
	if (g_userString.Length() > 1) {
		u32 u = g_gba.GetAddress(g_userString);

		if (u != 0xffffffff) {
			g_dataWindow->Goto(u);
		}
	}
}

static TString s_editError("Specify the address you'd like to modify, followed by the value, followed by b(byte), h(halfword), or w(word).");

/////////////////////////////////////////////////////////////////////////////
//  The user wishes to edit the memory at the indicated address.
static void EditMemory(void) {
	MType m;
	FString low, high;
	FString b;
	u8 option = 0;
	u32 value;

	for (m=0; m<g_userString.Length(); ++m) {
		if (isspace(g_userString[m])) {
			break;
		}
		low += g_userString[m];
	}
	if (m == g_userString.Length()) {
		WarnUser(s_editError);
	}
	while((m < g_userString.Length()) && isspace(g_userString[m]))
		++m;
	for (; m<g_userString.Length(); ++m) {
		high += g_userString[m];
	}
	switch(high[high.Length() - 1]) {
		case 'b':
		case 'B':
			option = 2;
			high.Chop();
			break;
		case 'h':
		case 'H':
			option = 1;
			high.Chop();
			break;
		case 'w':
		case 'W':
			option = 0;
			high.Chop();
			break;
	}
	value = ConvertToHex(high);
	switch(option) {
		case 0:
			g_gba.GetMemory().ForceD(g_gba.GetAddress(low),
									 value);
			break;
		case 1:
			g_gba.GetMemory().ForceW(g_gba.GetAddress(low),
									 value);
			break;
		case 2:
			g_gba.GetMemory().ForceB(g_gba.GetAddress(low),
									 value);
			break;
	}
	g_codeWindow->GenerateContent(CodeWindow::TAB_DISASSEMBLY);
	g_dataWindow->GenerateContent(DataWindow::TAB_MEMORY);
	g_appWindow->GenerateContent(AppWindow::TAB_REGISTERS);
}

static TString s_parseError("Parse Error:  breakpoints have the format [rwx] <address>.");

/////////////////////////////////////////////////////////////////////////////
//  The user has specified a breakpoint.
static void SetBreakpoint(void) {
	MType m;
	u8 options=0;
	FString f;
	u8 range = 4;

	for (m=0; m<g_userString.Length(); ++m) {
		if (g_userString[m] == ' ') {
			break;
		}
		if (g_userString[m] < TEXT_NUM_CONTROL_CODES) {
			continue;
		}
		switch(g_userString[m]) {
			case 'R':
			case 'r':
				options |= BP_READ;
				break;
			case 'W':
			case 'w':
				options |= BP_WRITE;
				break;
			case 'X':
			case 'x':
				options |= BP_EXEC;
				range = 1;
				break;
			default:
				WarnUser(s_parseError);
				return;
				break;
		}
	}
	if (options & (BP_READ | BP_WRITE)) {
		if (RegistrationGetLevel() < 1) {
			WarnUser(s_registrationLevel);
			return;
		}
	}
	if (!options) {
		WarnUser(s_parseError);
		return;
	}
	++m;
	if (m < g_userString.Length()) {
		for (; m<g_userString.Length(); ++m) {
			f += g_userString[m];
		}
	} // else error.
	g_dataWindow->AddBreakpoint(g_gba.GetAddress(f),
								range,
								options);
	g_dataWindow->SetTab(DataWindow::TAB_BREAKPOINTS);
}

static TString s_dumpRangeError("Dump commands expect two addresses [low] [high] (omit brackets, include space).");
static TString s_disassembleFileOpenFailure("Unable to open dis.txt.");
static TString s_disassemblyHeader("====== Disassembly ");

/////////////////////////////////////////////////////////////////////////////
//  The user wishes to dump a disassembly.
static void DumpDisassembly(void) {
	MType m;
	FString low, high;
	u32 lval, hval;
	u32 step;
	FILE * f;
	char buf[1024];
	FString b;

	for (m=0; m<g_userString.Length(); ++m) {
		if (isspace(g_userString[m])) {
			break;
		}
		low += g_userString[m];
	}
	while((m < g_userString.Length()) && isspace(g_userString[m]))
		++m;
	for (; m<g_userString.Length(); ++m) {
		high += g_userString[m];
	}
	lval = ConvertToHex(low);
	hval = ConvertToHex(high);
	if (lval > hval) {
		WarnUser(s_dumpRangeError);
		return;
	}
	if (g_gba.GetCPU().GetState() == ARM::STATE_ARM) {
		step = 4;
	} else {
		step = 2;
	}
	lval &= -step;
	hval &= -step;
	if (!(f = fopen("dis.txt", "a"))) {
		WarnUser(s_disassembleFileOpenFailure);
		return;
	}
	CopyToChar(s_disassemblyHeader, buf, 1024);
	fprintf(f, "%s %x - %x:\n", buf, lval, hval);
	while (lval < hval) {
		b.Destroy();
		g_gba.GetSymbol(b, lval);
		b += ":  ";
		if (step == 2) {
			sprintf(buf, "%04lx  ", (unsigned long)g_gba.GetMemory().GetW(lval));
		} else {
			sprintf(buf, "%08lx  ", (unsigned long)g_gba.GetMemory().GetD(lval));
		}
		b += buf;
		g_gba.GetCPU().Decode(b, lval);
		CopyToChar(b, buf, 1024);
		fprintf(f, "%s\n", buf);
		lval += step;
	}
	fclose(f);
}

static TString s_dumpFileOpenFailure("Unable to open dump.txt.");
static TString s_dumpHeader("====== Memory Dump ");

/////////////////////////////////////////////////////////////////////////////
//  The user wishes to dump the contents of memory.
static void DumpMemory(void) {
	MType m;
	FString low, high;
	u32 lval, hval;
	FILE * f;
	char buf[1024];
	FString b;

	for (m=0; m<g_userString.Length(); ++m) {
		if (isspace(g_userString[m])) {
			break;
		}
		low += g_userString[m];
	}
	while((m < g_userString.Length()) && isspace(g_userString[m]))
		++m;
	for (; m<g_userString.Length(); ++m) {
		high += g_userString[m];
	}
	lval = ConvertToHex(low);
	hval = ConvertToHex(high);
	if (lval > hval) {
		WarnUser(s_dumpRangeError);
		return;
	}

	if (!(f = fopen("dump.txt", "a"))) {
		WarnUser(s_dumpFileOpenFailure);
		return;
	}
	CopyToChar(s_dumpHeader, buf, 1024);
	fprintf(f, "%s %x - %x:\n", buf, lval, hval);
	while (lval < hval) {
		sprintf(buf, "%08lx:  ", (unsigned long)lval);
		fprintf(f, "%s", buf);
		for (m=0; m<16; ++m) {
			u8 size;
			u32 value;

			b.Destroy();
			switch(*g_memoryFormat) {
				case MTEXT:
					size = 1;
					break;
				case MWTEXT:
					size = 2;
					break;
				case MFLOAT:
					size = 4;
					break;
				default:
					size = 1 << (*g_memoryFormat & 3);
					break;
			}
			switch(size) {
				case 1:
					value = g_gba.GetMemory().GetB(lval);
					lval += 1;
					break;
				case 2:
					value = g_gba.GetMemory().GetW(lval);
					lval += 2;
					break;
				case 4:
					value = g_gba.GetMemory().GetD(lval);
					lval += 4;
					break;
			}
			MemoryTranslate(b, value, *g_memoryFormat);
			CopyToChar(b, buf, 1024);
			fprintf(f, "%s", buf);
		}
		fprintf(f, "\n");
	}
	fclose(f);
}

/////////////////////////////////////////////////////////////////////////////
//	A routine to update the mouse graphic.
static void UpdateMouse(void) {
	if (MouseIsVisible()) {
		u16 x = MouseGetX();
		u16 y = MouseGetY();
	
		if (x >= g_monitor.GetWidth() - g_mouse.GetWidth()) {
			x = g_monitor.GetWidth() - g_mouse.GetWidth() - 1;
		}
		if (y >= g_monitor.GetHeight() - g_mouse.GetHeight()) {
			y = g_monitor.GetHeight() - g_mouse.GetHeight() - 1;
		}
		g_mouse.SetPixelPosition(x, y);
		g_mouse.Render(g_monitor);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Update all of the windows.
static void UpdateWindows(void) {
	Window::UpdateAll();
	StatusUpdate();
}

/////////////////////////////////////////////////////////////////////////////
//  Set the mode to the mode indicated.
void SetMode(u8 mode) {
	if (s_mode != mode) {
		switch(mode) {
			case MODE_DEBUGGER:
				if (*g_runAppModeSwitch) {
					g_monitor.SetMode(g_userWidth,
									  g_userHeight,
									  g_userDepth,
									  !*g_runAppWindow);
					g_monitor.Clear(0);
					g_monitor.Refresh();
					g_monitor.Clear(0);
				}
				g_gba.GetVideo().SetTarget(VideoProcessor::TARGET_WINDOW);
				g_codeWindow->ForceRedraw();
				g_dataWindow->ForceRedraw();
				g_appWindow->ForceRedraw();
				MouseDisplay(TRUE);
				break;
			case MODE_APP:
				if (*g_runAppModeSwitch) {
					g_monitor.SetMode(EMU_WIDTH,
									  EMU_HEIGHT,
									  EMU_DEPTH);
					g_monitor.Clear(0);
					g_monitor.Refresh();
					g_monitor.Clear(0);
					g_gba.GetVideo().SetTarget(VideoProcessor::TARGET_FULLSCREEN);
				} else {
					// HEREHEREHERE disable mouse, do page copy.
					g_gba.GetVideo().SetTarget(VideoProcessor::TARGET_FAST_WINDOW);
				}
				break;
			case MODE_QUERY_USER:
				break;
		}
		s_mode = mode;
	}
}

static TString s_gbaReset1("Resetting GBA processor.");
static TString s_gbaReset2("All memory cleared except for ROM.");

static Key s_animating;
/////////////////////////////////////////////////////////////////////////////
//  Sample the user's keyboard input.
static void UpdateDebuggerKB(void) {
	Key k;
	Bool s, c;
	Ptr < Window > w;
	static u8 s_animOrientation;
	Key todo = s_animating;
	u8 orientation;
	u8 ano = s_animOrientation;

	if (KeysThere()) {
		k = KeyPop();
		s = k & KB_SHIFT;
		c = k & KB_CTRL;
		if (s) {
			w = g_dataWindow;
			#if !defined(RELEASE_BUILD)
				orientation = HISTORY_BACKWARD;
			#else
				orientation = HISTORY_FORWARD;
			#endif
		} else {
			w = g_codeWindow;
			orientation = HISTORY_FORWARD;
		}
		if (c) {
			w = g_appWindow;
		}
		k &= KB_BASE_MASK;
		switch(k) {
			case KB_P:
				VideoMakeScreenShot();
				Warn("Screenshot detected.\n");
				break;
			case KB_0:
				g_globalEnable ^= ENABLE_BG0;
				break;
			case KB_1:
				g_globalEnable ^= ENABLE_BG1;
				break;
			case KB_2:
				g_globalEnable ^= ENABLE_BG2;
				break;
			case KB_3:
				g_globalEnable ^= ENABLE_BG3;
				break;
			case KB_B:
				DialogGetString(s_enterBreakpoint,
								SetBreakpoint);
				break;
			case KB_D:
				if (s) {
					DialogGetString(s_getDumpRange,
									DumpMemory);
				} else {
					DialogGetString(s_getDisassemblyRange,
									DumpDisassembly);
				}
				break;
			case KB_E:
				if (s) {
					DialogGetString(s_editMemory,
									EditMemory);
				} else {
					// Reserved for "assemble".
				}
				break;
			case KB_G:
				if (s) {
					DialogGetString(s_getCodeAddress,
									DataGoto);
				} else {
					DialogGetString(s_getCodeAddress,
									CodeGoto);
				}
				break;
			case KB_S:
				g_globalEnable ^= ENABLE_OAM;
				break;
			case KB_X:
				++(*g_grid);
				if (*g_grid == 3)
					*g_grid = 0;
				g_appWindow->GenerateContent(AppWindow::TAB_APP);
				break;
			case KB_PGUP:
				w->PageUp();
				break;
			case KB_PGDN:
				w->PageDown();
				break;
			case KB_UP:
				w->CursorUp();
				break;
			case KB_DOWN:
				w->CursorDown();
				break;
			case KB_F2:
				BreakpointToggle(g_codeWindow->GetCurrentCursorIP());
				break;
			case KB_F4:
				g_gba.Reset();
				DialogDeliverMessage(s_gbaReset1,
									 s_gbaReset2);
				break;
			case KB_F5:
				s_animating = todo = 0;
				SignalExecute(orientation);
				g_exceptionMask = EXC_NORMAL;
				SetMode(MODE_APP);
				break;
			#if !defined(RELEASE_BUILD)
				case KB_F6:
			#endif
			case KB_F7:
			case KB_F8:
			case KB_F9:
				todo = k;
				ano = orientation;
				if (c) {
					s_animating = k;
					s_animOrientation = orientation;
				} else {
					if (s_animating) {
						todo = 0;
					}
					s_animating = 0;
				}
				break;
			default:
				break;
		}
	}
	switch(todo) {
		#if !defined(RELEASE_BUILD)
			case KB_F6:
				g_exceptionMask = EXC_NORMAL | EXC_RETIRED;
				SignalExecute(ano);
				break;
		#endif
		case KB_F7: {
			u32 bpAddr = 0;

			if (g_gba.GetCPU().GetInstructionSize() == 4) {
				u32 u = g_gba.GetMemory().GetD(g_gba.GetCPU().GetCurrentIP());

				if (((u & 0xf000000) == 0xb000000) ||
					((u & 0x0ffffff0) == 0x12fff10)) {
					bpAddr = g_gba.GetCPU().GetCurrentIP() + 4;
				}
			} else {
				u16 u = g_gba.GetMemory().GetW(g_gba.GetCPU().GetCurrentIP());

				if (((u & 0xf800) == 0xf000) ||
					((u & 0xff00) == 0x4700)) {
					bpAddr = g_gba.GetCPU().GetCurrentIP() + 4;
				}
			}
			if (bpAddr) {
				SetMode(MODE_APP);
				g_exceptionMask = EXC_NORMAL;
				// HEREHEREHERE this will not work for complex recursion.
				g_branchBP = g_gba.GetCPU().GetCurrentIP();
				g_retiredBP = bpAddr;
				g_branchDepth = 0;
				BreakpointSet(bpAddr, 1, 1, BP_EXEC, BP_INTERNAL);
				BreakpointSet(g_branchBP, 1, 1, BP_EXEC, BP_INTERNAL);
			} else {
				g_exceptionMask = EXC_NORMAL | EXC_RETIRED;
				g_retiredCheck = g_gba.GetCPU().GetCurrentIP();
			}
			SignalExecute(ano);
			break;
		}
		case KB_F8:
			SignalExecute(ano);
			g_exceptionMask = EXC_NORMAL | EXC_RETIRED;
			g_retiredCheck = g_gba.GetCPU().GetCurrentIP();
			break;
		case KB_F9:
			SignalExecute(ano);
			g_exceptionMask = EXC_NORMAL | EXC_ATOMIC;
			break;
		default:
			break;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Update the KB handler when we're in APP mode.  HEREHEREHERE we have to
// deal with reverse execution different (no outside stimuli).
static void UpdateAppKB(void) {
	Key k;
	static Key s_keys[] = {		// HEREHEREHERE make these remappable.
		KB_ALT,					// A
		KB_CTRL,				// B
		KB_SPACE,				// SELECT
		KB_ENTER,				// START
		KB_RIGHT,				// RIGHT
		KB_LEFT,				// LEFT
		KB_UP,					// UP
		KB_DOWN,				// DOWN
		KB_D,					// R
		KB_A,					// L
	};
	MType m;
	u16 n;
	u16 gk = 0x3ff;

	if (KeysThere()) {
		k = KeyPop();
		switch(k & KB_BASE_MASK) {
			case KB_P:
				VideoMakeScreenShot();
				Warn("Screenshot detected.\n");
				break;
			case KB_0:
				g_globalEnable ^= ENABLE_BG0;
				break;
			case KB_1:
				g_globalEnable ^= ENABLE_BG1;
				break;
			case KB_2:
				g_globalEnable ^= ENABLE_BG2;
				break;
			case KB_3:
				g_globalEnable ^= ENABLE_BG3;
				break;
			case KB_S:
				g_globalEnable ^= ENABLE_OAM;
				break;
			case KB_X:
				++(*g_grid);
				if (*g_grid == 3)
					*g_grid = 0;
				break;
			case KB_F5:
				if (k & KB_SHIFT) {
					#if !defined(RELEASE_BUILD)
						SignalExecute(HISTORY_BACKWARD);
					#else
						SignalExecute(HISTORY_FORWARD);
					#endif
				} else if (k & KB_CTRL) {
					g_execute = FALSE;
				} else if (g_execute) {
					// HEREHEREHERE this needs to be in its own function.
					if (g_retiredBP) {
						BreakpointSet(g_retiredBP, 1, 1, 0, BP_INTERNAL);
						BreakpointSet(g_branchBP, 1, 1, 0, BP_INTERNAL);
						g_retiredBP = 0;
					}
					SetMode(MODE_DEBUGGER);
					g_codeWindow->SingleStep();
					g_dataWindow->SingleStep();
					g_appWindow->SingleStep();
					g_execute = FALSE;
				} else {
					g_execute = TRUE;
				}
				break;
			default:
				break;
		}
	}
	// Check for keys that are presently down.
	for (m=0, n=1; m<10; ++m) {
		if (KeyDown(s_keys[m])) {
			gk &= ~n;
		}
		n += n;
	}
	// HEREHEREHERE are users required to set the bits after they're
	// sampled, or not?
	g_gba.GetMemory().SetW(0x4000130, gk);
}

/////////////////////////////////////////////////////////////////////////////
//  A simple handler for when we're sampling user input.
static void UpdateUserKB(void) {
	Key k;

	if (KeysThere()) {
		k = KeyPop();
		if (k & KB_SHIFT) {
			k &= KB_BASE_MASK;
			k = toupper(k);
		} else {
			k &= KB_BASE_MASK;
			k = tolower(k);
		}
		DialogProcessInput(k);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Update the app, subject to the constraints the user has imposed.
static void AppTick(void) {
	if (g_execute) {
		if (HistoryGetOrientation() == HISTORY_FORWARD)
			g_gba.Tick();
		else
			g_gba.TickBackward();
		if (g_exceptions & EXC_USER) {
			if (g_retiredBP) {
				BreakpointSet(g_retiredBP, 1, 1, 0, BP_INTERNAL);
				BreakpointSet(g_branchBP, 1, 1, 0, BP_INTERNAL);
				g_retiredBP = 0;
			}
			SetMode(MODE_DEBUGGER);
			g_codeWindow->SingleStep();
			g_dataWindow->SingleStep();
			g_appWindow->SingleStep();
			g_execute = FALSE;
		}
	}
}

Bool ProcessIPCMessage(u32 p1, u32 p2) {
	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////
//  This will handle the alt-enter key combination.
void AGDAltEnter(void) {
	*g_runAppWindow = !(*g_runAppWindow);
	if ((s_mode == MODE_DEBUGGER) || !(*g_runAppModeSwitch)) {
		g_monitor.SetMode(g_userWidth,
						  g_userHeight,
						  g_userDepth,
						  !*g_runAppWindow);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This is the main GBA thread.
u8 MainThread(void * dummy) {
	Window w3;
	FString msg1, msg2;
	MType m, n;
	char buf[512];
	Bool showSplash=TRUE;
	Bool runNow=FALSE;

	ThreadSuspend(g_garbageThread, TRUE);
	g_monitor.Clear(g_colors[BACKGROUND]);

	SafeNew(g_codeWindow, (MType)1, __FILE__, __LINE__);
	g_codeWindow->Activate();
	SafeNew(g_dataWindow, (MType)1, __FILE__, __LINE__);
	g_dataWindow->Activate();
	SafeNew(g_appWindow, (MType)1, __FILE__, __LINE__);
	g_appWindow->Activate();

	// Let's examine our command-line args.
	for (m=0; m<g_args.Count(); ++m) {
		const FString & f = g_args[m];

		if ((f[0] == 'A') || (f[0] == 'a')) {
			switch(f[1]) {
				case 'C':
				case 'c':
					{
						FILE * killer;

						if (killer = fopen("debug.txt", "w")) {
							fclose(killer);
						}
					}
					break;
				case 'R':
				case 'r':
					CopyToChar(f, buf, 512);
					g_gba.Load(buf + 2);
					showSplash = FALSE;
					break;
				case 'G':
				case 'g':
					runNow = TRUE;
					showSplash = FALSE;
					break;
				case 'B':
				case 'b':
					g_userString.Destroy();
					for (n=2; n<f.Length(); ++n) {
						if ((f[n] != 'r') &&
							(f[n] != 'R') &&
							(f[n] != 'w') &&
							(f[n] != 'W') &&
							(f[n] != 'x') &&
							(f[n] != 'X')) {
							break;
						}
						g_userString += f[n];
					}
					g_userString += ' ';
					for (; n<f.Length(); ++n) {
						g_userString += f[n];
					}
					SetBreakpoint();
					break;
			}
		}
	}

	g_gba.Reset();

	if (runNow) {
		g_exceptionMask = EXC_NORMAL;
		SignalExecute(HISTORY_FORWARD);
		SetMode(MODE_APP);
	} else {
		SetMode(MODE_DEBUGGER);
		ActiveGraphic::RedrawAll();
		g_monitor.Refresh();
		ActiveGraphic::RedrawAll();
	}
	if (showSplash) {
		msg1 = g_appTitle;
		msg2 = s_copyright;
		DialogDeliverMessage(msg1, msg2);
	}
	g_altEnterHandler = AGDAltEnter;

	if (*g_soundEnable) {
		GBASoundOn();
	}

	while (1) {
		switch(s_mode) {
			case MODE_DEBUGGER:
				if ((g_execute == FALSE) || (g_exceptions & EXC_USER)) {
					UpdateMouse();
					UpdateWindows();
					UpdateDebuggerKB();
				} else if (KeyDown(KB_F5)) {
					g_execute = FALSE;
				}
				AppTick();
				if ((g_execute == FALSE) || (g_exceptions & EXC_USER)) {
					g_monitor.Refresh();
				}
				break;
			case MODE_APP:
				UpdateAppKB();
				AppTick();
				break;
			case MODE_QUERY_USER:
				UpdateMouse();
				UpdateUserKB();
				DialogRender();
				g_monitor.Refresh();
				break;
		}
		ThreadYield();
	}
	return 0;
}
