/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	All of the THUMB decoders reside here.

****************************************************************************/

#include "gba.h"

#if !defined(RELEASE_BUILD)
	extern "C" {
		u32 g_debug;
	};
#endif

/////////////////////////////////////////////////////////////////////////////
//  "Move shifted register".  Shift a register by an immediate amount and put
// it in another register.  Set the flags register.
void ARM::DecTHMSR(u32 address,
				   u16 code,
				   ARM::DecodedInstruction & dest) {
	u32 c2;
	u32 rs = ((code >> 3) & 7);
	u32 rd = ((code >> 0) & 7);
	u32 imm = ((code >> 6) & 0x1f);
	u32 op = ((code >> 11) & 3);

	c2 = ((u32)0xe1b00000 |
		  (rs << 0) |
		  (op << 5) |
		  (imm << 7) |
		  (rd << 12));
	DecDataProcessing(address, c2, dest);
}

/////////////////////////////////////////////////////////////////////////////
//  Add or subtract one register from another.  Another BDP.
void ARM::DecTHAddSub(u32 address,
					  u16 code,
					  ARM::DecodedInstruction & dest) {
	u32 c2;
	Bool imm = ((code >> 10) & 1);
	Bool sub = ((code >> 9) & 1);
	u32 rn = ((code >> 6) & 7);
	u32 rs = ((code >> 3) & 7);
	u32 rd = ((code >> 0) & 7);

	c2 = ((u32)0xe0100000 |
		  (imm ? (u32)0x02000000 : (u32)0x00000000) |
		  (sub ? (u32)0x00400000 : (u32)0x00800000) |
		  (rs << 16) |
		  (rd << 12) |
		  (rn << 0));
	DecDataProcessing(address, c2, dest);
}

/////////////////////////////////////////////////////////////////////////////
//  Do mov/compare/add/subtract with immediate.
void ARM::DecTHOpImm(u32 address,
					 u16 code,
					 ARM::DecodedInstruction & dest) {
	u32 c2;
	u8 op = ((code >> 11) & 3);
	u32 rd = ((code >> 8) & 7);
	u32 offset = ((code >> 0) & 0xff);
	static u32 s_ops[] = {
		0x01a00000,
		0x01400000,
		0x00800000,
		0x00400000,
	};

	c2 = ((u32)0xe2100000 |
		  s_ops[op] |
		  (rd << 12) |
		  (rd << 16) |
		  (offset << 0));
	DecDataProcessing(address, c2, dest);
}

/////////////////////////////////////////////////////////////////////////////
//  Do any of the ALU ops normally afforded to us in ARM mode.
void ARM::DecTHALUOps(u32 address,
					  u16 code,
					  ARM::DecodedInstruction & dest) {
	u32 c2;
	u32 op = ((code >> 6) & 0xf);
	u32 rd = ((code >> 0) & 0x7);
	u32 rs = ((code >> 3) & 0x7);

	switch(op) {
		case 0:
		case 1:
		case 5:
		case 6:
		case 8:
		case 10:
		case 11:
		case 12:
		case 14:
		case 15:
			c2 = ((u32)0xe0100000 |
				  (op << 21) |
				  (rd << 12) |
				  (rd << 16) |
				  (rs << 0));
			DecDataProcessing(address, c2, dest);
			break;
		case 7:
			op = 5;
		case 2:
		case 3:
		case 4:
			op -= 2;
			c2 = ((u32)0xe1b00010 |
				  (rd << 12) |
				  (rd << 0) |
				  (op << 5) |
				  (rs << 8));
			DecDataProcessing(address, c2, dest);
			break;
		case 9:		// Negate:  rsb with 0.
			c2 = ((u32)0xe2700000 |
				  (rd << 12) |
				  (rs << 16));
			DecDataProcessing(address, c2, dest);
			break;
		case 13:	// Multiply.
			c2 = ((u32)0xe0100090 |
				  (rs << 0) |
				  (rd << 8) |
				  (rd << 16));
			DecMultiply(address, c2, dest);
			break;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Operations involving the hi registers.
void ARM::DecTHHiRegOps(u32 address,
						u16 code,
						ARM::DecodedInstruction & dest) {
	u32 c2 = 0;
	u32 rd = ((code >> 0) & 7) |
			 ((code >> 4) & 8);
	u32 rs = ((code >> 3) & 0xf);
	u8 op = ((code >> 8) & 3);
	u32 excSave = GetForcedExceptions();
	static u32 s_ops[] = {
		0x00800000,
		0x01400000,
		0x01a00000
	};

	if ((rs < 8) && (rd < 8) && (op < 3)) {
		ForceExceptions(EXC_INVALID_OPCODE);
	}
	switch(op) {
		case 1:
			c2 |= (u32)0x00100000;
		case 2:
		case 0:
			c2 |= s_ops[op];
			c2 |= ((u32)0xe0000000 |
				   (rd << 12) |
				   (rd << 16) |
				   (rs << 0));
			DecDataProcessing(address, c2, dest);
			break;
		case 3:
			c2 = ((u32)0xe0000000 |
				  rs);
			DecBX(address, c2, dest);
			break;
	}
	ClearForcedExceptions();
	ForceExceptions(excSave);
}

/////////////////////////////////////////////////////////////////////////////
//  Do a PC-relative load.
void ARM::DecTHPCRelLd(u32 address,
					   u16 code,
					   ARM::DecodedInstruction & dest) {
	u32 c2;
	u32 rd = ((code >> 8) & 0x7);
	u32 imm = ((code >> 0) & 0xff) << 2;

	c2 = ((u32)0xe59f0000 |
		  (rd << 12) |
		  imm);

	DecSingleDataTransfer(address, c2, dest);
}

/////////////////////////////////////////////////////////////////////////////
//  Do a register-indexed load or store.
void ARM::DecTHLdStRegOff(u32 address,
						  u16 code,
						  ARM::DecodedInstruction & dest) {
	u32 c2;
	u32 l = ((code >> 11) & 1);
	u32 b = ((code >> 10) & 1);
	u32 ro = ((code >> 6) & 0x7);
	u32 rb = ((code >> 3) & 0x7);
	u32 rd = ((code >> 0) & 0x7);

	c2 = ((u32)0xe7800000 |
		  (l << 20) |
		  (b << 22) |
		  (ro << 0) |
		  (rb << 16) |
		  (rd << 12));
	DecSingleDataTransfer(address, c2, dest);
}

void ARM::DecTHLdStSexBHW(u32 address,
						  u16 code,
						  ARM::DecodedInstruction & dest) {
	u32 c2;
	u32 sh = ((code >> 10) & 3);
	u32 ro = ((code >> 6) & 7);
	u32 rb = ((code >> 3) & 7);
	u32 rd = ((code >> 0) & 7);
	// H and S bit are flipped, for some reason.
	static u8 s_flips[] = {
		1,
		2,
		1,
		3
	};

	if (sh) {
		c2 = ((u32)1 << 20);
	} else {
		c2 = 0;
	}
	sh = s_flips[sh];
	c2 |= ((u32)0xe1800090 |
		   (ro << 0) |
		   (rb << 16) |
		   (rd << 12) |
		   (sh << 5));
	DecHalfwordReg(address, c2, dest);
}

/////////////////////////////////////////////////////////////////////////////
//  Direct-addressing load/store.
void ARM::DecTHLdStImm(u32 address,
					   u16 code,
					   ARM::DecodedInstruction & dest) {
	u32 c2;
	u32 b = ((code >> 12) & 1);
	u32 l = ((code >> 11) & 1);
	u32 imm = ((code >> 6) & 0x1f);
	u32 rb = ((code >> 3) & 7);
	u32 rd = ((code >> 0) & 7);

	if (!b) {
		imm <<= 2;
	}
	c2 = ((u32)0xe5800000 |
		  (b << 22) |
		  (l << 20) |
		  (imm << 0) |
		  (rb << 16) |
		  (rd << 12));
	DecSingleDataTransfer(address, c2, dest);
}

/////////////////////////////////////////////////////////////////////////////
//  Load a halfword from, or store a halfword to, memory.
void ARM::DecTHLdStHw(u32 address,
					  u16 code,
					  ARM::DecodedInstruction & dest) {
	u32 c2;
	u32 rd = ((code >> 0) & 7);
	u32 rb = ((code >> 3) & 7);
	u32 imm = ((code >> 6) & 0x1f) << 1;
	u32 l = ((code >> 11) & 1);

	c2 = ((u32)0xe1c000b0 |
		  (rd << 12) |
		  (rb << 16) |
		  ((imm & 0xf0) << 4) |
		  (imm & 0xf) |
		  (l << 20));
	DecHalfwordImm(address, c2, dest);
}

/////////////////////////////////////////////////////////////////////////////
//  Do an SP-relative load or store.
void ARM::DecTHSPRelLdSt(u32 address,
						 u16 code,
						 ARM::DecodedInstruction & dest) {
	u32 c2;
	u32 l = ((code >> 11) & 1);
	u32 rd = ((code >> 8) & 7);
	u32 imm = ((code >> 0) & 0xff) << 2;

	c2 = ((u32)0xe58d0000 |
		  (l << 20) |
		  (rd << 12) |
		  (imm));
	DecSingleDataTransfer(address, c2, dest);
}

void ARM::DecTHLEA(u32 address,
				   u16 code,
				   ARM::DecodedInstruction & dest) {
	u32 c2;
	Bool sp = ((code >> 11) & 1);
	u32 rd = ((code >> 8) & 7);
	u32 imm = ((code >> 0) & 0xff) << 2;

	c2 = ((u32)0xe6800000 |
		  (sp ? ((u32)0xd << 16) : ((u32)0xf << 16)) |
		  (rd << 12) |
		  (imm));
	DecDataProcessing(address, c2, dest);
}

/////////////////////////////////////////////////////////////////////////////
//  Add an offset to SP.
void ARM::DecTHAddOffSP(u32 address,
						u16 code,
						ARM::DecodedInstruction & dest) {
	u32 c2;
	Bool as = ((code >> 7) & 1);
	u32 imm = ((code >> 0) & 0x7f) << 2;

	c2 = ((u32)0xe60dd000 |
		  imm |
		  (as ? (u32)0x00400000 : 0x00800000));
	DecDataProcessing(address, c2, dest);
}

/////////////////////////////////////////////////////////////////////////////
//  Encode a Thumb push/pop.
void ARM::DecTHPushPop(u32 address,
					   u16 code,
					   ARM::DecodedInstruction & dest) {
	u32 c2;
	u32 rlist = ((code >> 0) & 0xff);
	Bool ld = ((code >> 11) & 1);
	Bool lrpc = ((code >> 8) & 1);

	c2 = (u32)0xe82d0000;
	if (ld) {
		c2 |= 0x00900000;
		if (lrpc) {
			rlist |= 0x8000;
		}
	} else {
		c2 |= 0x01000000;
		if (lrpc) {
			rlist |= 0x4000;
		}
	}
	c2 |= rlist;
	DecBlockDataTransfer(address, c2, dest);
}

/////////////////////////////////////////////////////////////////////////////
//  Multiple load/store.
void ARM::DecTHMLS(u32 address,
				   u16 code,
				   ARM::DecodedInstruction & dest) {
	u32 c2;
	u32 l = ((code >> 11) & 1);
	u32 rb = ((code >> 8) & 7);
	u32 rlist = ((code >> 0) & 0x7f);

	c2 = ((u32)0xe8a00000 |
		  (l << 20) |
		  (rb << 16) |
		  (rlist << 0));
	DecBlockDataTransfer(address, c2, dest);
}

/////////////////////////////////////////////////////////////////////////////
//  Execute a conditional branch.
void ARM::DecTHCondBr(u32 address,
					  u16 code,
					  ARM::DecodedInstruction & dest) {
	u32 c2;
	u32 cond = ((code >> 8) & 0xf);
	u32 imm = ((code >> 0) & 0xff);

	if (imm & 0x80) {
		imm |= 0xffff00;
	}
	c2 = ((u32)0x0a000000 |
		  (cond << 28) |
		  imm);
	DecBranch(address, c2, dest);
}

/////////////////////////////////////////////////////////////////////////////
//  Software interrupt.
void ARM::DecTHSWI(u32 address,
				   u16 code,
				   ARM::DecodedInstruction & dest) {
	u32 c2;
	u32 imm = ((code >> 0) & 0xff);

	c2 = ((u32)0xef000000 |
		  imm);
	DecSWI(address, c2, dest);
}

/////////////////////////////////////////////////////////////////////////////
//  Do an unconditional branch.
void ARM::DecTHUncondBranch(u32 address,
							u16 code,
							ARM::DecodedInstruction & dest) {
	u32 c2;
	u32 imm = ((code >> 0) & 0x7ff);

	if (imm & 0x400) {
		imm |= 0xfff800;
	}
	c2 = ((u32)0xea000000 |
		  imm);
	DecBranch(address, c2, dest);
}

/////////////////////////////////////////////////////////////////////////////
//  Effect a branch-and-link.
void ARM::DecTHLongBrLink(u32 address,
						  u16 code,
						  ARM::DecodedInstruction & dest) {
	u32 imm = ((code >> 0) & 0x7ff);
	Bool c1 = ((code >> 11) & 1);
	U16 * d;
	u32 leftOff;
	u32 exceptions = EXC_ATOMIC |
					 EXC_RETIRED |
					 EXC_CYCLE_BREAK |
					 GetForcedExceptions();
	u32 cycles;

	d = InsertConditionCodeCheck(address, 0xe, dest, leftOff);
	if (!c1) {
		u32 disp;

		exceptions &= ~EXC_RETIRED;
		if (imm & 0x400) {
			imm |= 0xfffffc00;
		}
		imm <<= 12;
		InsertImmediateToRMNoDisp(ARMEX_MemStore_Source,
								  address + 4 + imm);
		disp = GetRegAddress(14);
		InsertImmediateToReg(ARMEX_MemStore_Dest,
							 disp);
		InsertDispToRegRM(ARMEX_MemStore_DestCopy,
								  disp);
		Insert(dest, ARMEX_MemStore, ARMEX_MemStore_End);
		cycles = GetCycles(address, 1, 0, 0);		// HEREHEREHERE Unknown.
	} else {
		u32 r14 = GetRegAddress(14);
		u32 r15 = GetRegAddress(15);

		SaveHistory(dest, r14);
		SaveHistory(dest, r15);
		imm <<= 1;
		InsertDispToAcc(ARMEX_LongBranch, r14);
		InsertImmediateToAcc(ARMEX_LongBranch_Offset, imm);
		InsertDispToAcc(ARMEX_LongBranch_StoreR15, r15);
		InsertDispToRMImm(ARMEX_LongBranch_StoreR14, r14);
		InsertImmediateToRM(ARMEX_LongBranch_StoreR14, address + 3);
		Insert(dest, ARMEX_LongBranch, ARMEX_LongBranch_End);
		exceptions |= EXC_EXTERNAL_JUMP;
		cycles = GetCycles(address, 1, 0, 0);		// HEREHEREHERE Unknown.
	}
	InsertConditionCodePost(dest, d, leftOff, exceptions, address, cycles);
}

