/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	The implementation of the video processor.  When we transition from
hblank-on to hblank-off, we will render the next scanline.  Any rendering
that occurs after hblank-off to that scanline will be ignored.
	Okay, we're ready to start implementing the new rendering system.
Macroscopically, this is the algorithm:
	- Per scanline, get a priority-bucketed list of sprites and BG.
	- Render the items to the two layers according to their characteristics.
	- Accumulate the results.
	The optimal way to do each step is presently being debated.  I think for
now I will do a C++ implementation and let the profiler tell me where the
work needs to be done.

****************************************************************************/

#include "gba.h"

// HEREHEREHERE have to experimentally determine correct value.
#define Y_BECOMES_NEGATIVE				161
#define X_BECOMES_NEGATIVE				384

extern "C" {
	extern U8 * g_readMap[];
	extern U8 * g_writeMap[];
};

#define VIDEO_REGISTER_SEG			4
#define PALETTE_SEG					5
#define VRAM_SEG					6
#define OAM_SEG						7
#define HORIZONTAL_RENDER_PERIOD	1004
#define HORIZONTAL_TOTAL_PERIOD		1232
#define VBLANK_SCANLINE_START		160
#define VBLANK_TOTAL_SCANLINES		229

// OAM stuff.
#define PRIO_SHIFT					10
#define PRIO_MASK					3

Sticky < u8 > g_grid("grid", 0);

enum {
	BG_OFF,			// This BG disabled in this mode.
	BG_NRM,			// Normal tile mode.
	BG_ROT,			// Rotated/scaled tile mode.
	BG_PAL,			// Paletted (mode 0x13) tile mode.
	BG_DIR,			// Direct-color mode.
};

typedef SqrMatrix < S16, 2 > GBTForm;

struct BGInfo {
	BGInfo * next;
	GBTForm transform;
	u32 xMask;
	u32 yMask;
	u8 shift;
	U16 * mapAddress;
	U8 * dataAddress;
	u32 xBase;
	u32 yBase;
	u32 xVal;
	u32 yVal;
	u8 priority;
	u8 mode;
	Bool pal256;
	Bool transp;
	Bool rotate;
	Bool tiled;
};

static BGInfo s_backgrounds[4];
static BGInfo * s_BGBuckets[4];

// Globally enable/disable BG and OAM.
u16 g_globalEnable;
	
//  Do we switch to fullscreen when we run the app?
Sticky < Bool > g_runAppModeSwitch("Run app mode switch", TRUE);

//	Do we run in windowed mode?  This negates g_runAppModeSwitch.
Sticky < Bool > g_runAppWindow("Run app in windowed mode", FALSE);

static Bitmap s_dest;
VideoProcessor::GBSprite * VideoProcessor::s_sprites[128];

u16 g_ticksTilNextVideo;

//  These flags govern the appearance of the sprite image.
#define IS_ALPHA		0x80
#define IS_4BIT			0x40
#define IS_XFLIP		0x20
#define IS_YFLIP		0x10
#define IS_DOUBLE		0x08
#define IS_DISABLED		0x04
#define IS_IN_QUEUE		0x02

/////////////////////////////////////////////////////////////////////////////
//  Okay, I'm switching over to a fixed-point implementation, which hopefully
// means a static lookup table (no cache fills!)
struct InternalSprite {
	InternalSprite * next;
	InternalSprite * prev;
	u8 index;
	u16 startX;
	u8 startY;
	u8 flags;
	u8 pal;
	u8 priority;
	u8 sz;
	u8 width;
	u8 height;
	u8 tform;
	u16 tile;
	U8 * tileAddress;
};
//  Our actual internals!
static InternalSprite s_internals[128];
static VideoProcessor::GBSprite s_spriteData[128];
static InternalSprite * s_spritesEachLine[4][160];
// The indices below are priority, buffer, sprite index.  We're double-
// buffering.
static InternalSprite * s_currentSpriteList[4][2][128];
static u8 s_currentSpriteCount[4];
static u8 s_widths[] = {
	8, 16, 32, 64, 16, 32, 32, 64, 8, 8, 16, 32, 0, 0, 0, 0
};
static u8 s_heights[] = {
	8, 16, 32, 64, 8, 8, 16, 32, 16, 32, 32, 64, 0, 0, 0, 0
};
static u8 s_wShifts[] = {
	0, 1, 2, 3, 1, 2, 2, 3, 0, 0, 1, 2, 0, 0, 0, 0
};

struct SpriteImage {
	Map < U16 > data;
};

// The first index is for whether we're in 1D or 2D mode.
static SpriteImage s_spriteImages[2][16];
static GBTForm s_standardRotations[4];
static Bool s_filled0[240];
static Bool s_filled1[240];
static Bool s_alpha0[240];
static u16 s_scanline0[240];
static u16 s_scanline1[240];
static u16 s_pixelCount;
static u8 s_screenWidth, s_screenHeight;

/////////////////////////////////////////////////////////////////////////////
//  This will build the static tile-indirection tables.
void VideoProcessor::BuildSpriteIndirectionTables(void) {
	static Bool s_initialized;

	if (!s_initialized) {
		u8 u, x, y, m;

		for (m=0; m<2; ++m) {
			for (u=0; u<16; ++u) {
				SpriteImage & s = s_spriteImages[m][u];
				u8 w = s_widths[u];
				u8 h = s_heights[u];
				u8 ws;

				if (m) {
					ws = s_wShifts[u];
				} else {
					// ws = 5;
					ws = 4;
				}
		
				s.data.Allocate(w, h, __FILE__, __LINE__);
				s.data.Open(Map < U16 >::WRITE);
				for (y=0; y<h; ++y) {
					for (x=0; x<w; ++x) {
						s.data[y][x] = ((y >> 3) << (ws + 6)) +
									   ((y & 7) << 3) +
									   ((x >> 3) << 6) +
									   ((x & 7));
					}
				}
				s.data.Close();
			}
		}
		s_initialized = TRUE;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  De-initialize the video processor.
void VideoProcessor::DeInitialize(void) {
	MType m, u;

	for (m=0; m<2; ++m) {
		for (u=0; u<16; ++u) {
			s_spriteImages[m][u].data.Destroy();
		}
	}
	s_dest.Destroy();
}

/////////////////////////////////////////////////////////////////////////////
//  Remove the indicated sprite from all queues that it presently occupies.
inline void RemoveSpriteFromQueue(InternalSprite & is) {
	u8 y;

	if ((y=is.startY) >= Y_BECOMES_NEGATIVE)
		y = 0;
	if (!(is.flags & IS_DISABLED)) {
		if (is.prev) {
			is.prev->next = is.next;
		} else {
			s_spritesEachLine[is.priority][y] = is.next;
		}
		if (is.next) {
			is.next->prev = is.prev;
		}
		is.flags |= IS_DISABLED;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Insert the indicated sprite into all relevant queues.
inline void AddSpriteToQueue(InternalSprite & is) {
	InternalSprite * s, * p;
	u8 y;

	if ((y=is.startY) >= Y_BECOMES_NEGATIVE)
		y = 0;
	if (s = s_spritesEachLine[is.priority][y]) {
		p = NULL;
		while (s && (s->index < is.index)) {
			p = s;
			s = s->next;
		}
		if (p) {
			p->next = &is;
		} else {
			s_spritesEachLine[is.priority][y] = &is;
		}
		is.prev = p;
		if (s) {
			s->prev = &is;
		}
		is.next = s;
	} else {
		s_spritesEachLine[is.priority][y] = &is;
		is.next = is.prev = NULL;
	}
	is.flags &= ~IS_DISABLED;
}

/////////////////////////////////////////////////////////////////////////////
//  Replace the sprite in the relevant queues.
inline void ReplaceSpriteInQueue(InternalSprite & is,
								 u8 yy,
								 u8 scanline,
								 u8 priority) {
	MType m;
	u8 buf=scanline & 1;
	u8 disc;

	RemoveSpriteFromQueue(is);
	is.startY = yy;
	is.priority = priority;
	AddSpriteToQueue(is);

	if (yy < Y_BECOMES_NEGATIVE) {
		disc = scanline - yy;
	} else {
		disc = scanline + (256 - yy);
	}

	// If I'm in the current scanline queue, and don't deserve to be, then
	// remove me.  Else, insert me (or leave me there).
	if (is.flags & IS_IN_QUEUE) {
		if (disc >= is.height) {
			for (m=0; m<s_currentSpriteCount[priority]; ++m) {
				if (s_currentSpriteList[priority][buf][m] == &is)
					break;
			}
			CHECK(m != s_currentSpriteCount[priority]);
			--s_currentSpriteCount[priority];
			for (; m<s_currentSpriteCount[priority]; ++m) {
				s_currentSpriteList[priority][buf][m] =
					s_currentSpriteList[priority][buf][m + 1];
			}
			is.flags &= ~IS_IN_QUEUE;
		}
	} else if (disc && (disc < is.height)) {
		// If disc is equal to zero, it means this sprite will
		// be inserted in the next merge.
		MType n;

		for (m=0; m<s_currentSpriteCount[priority]; ++m) {
			if (s_currentSpriteList[priority][buf][m]->index >
				is.index)
				break;
		}

		for (n=s_currentSpriteCount[priority]; n>m; --n) {
			s_currentSpriteList[priority][buf][n] =
				s_currentSpriteList[priority][buf][n - 1];
		}
		s_currentSpriteList[priority][buf][m] =	&is;
		++s_currentSpriteCount[priority];
		is.flags |= IS_IN_QUEUE;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This function must be called to bind the video processor to the View.  I
// know, it's kind of sloppy, but the main libs are WIP anyway.
void VideoProcessor::Initialize(void) {
	u8 r, g, b;
	u16 a;
	u16 index;
	MType m, n, p;

	m_orientation = HISTORY_FORWARD;
	// First, let's build the color tables for the debug screen.
	g_monitor.SetMode(g_userWidth,
					  g_userHeight,
					  g_userDepth,
					  !*g_runAppWindow);
	for (b=0, index=0; b<32; ++b) {
		for (g=0; g<32; ++g) {
			for (r=0; r<32; ++r, ++index) {
				m_debugMap[index] = g_monitor.MapColor(r / (real)31,
													   g / (real)31,
													   b / (real)31);
			}
		}
	}
	m_debugMask = g_monitor.MapColor(0.4999, 0.4999, 0.4999);

	// Now, the color tables for the app screen.
	g_monitor.SetMode(EMU_WIDTH,
					  EMU_HEIGHT,
					  EMU_DEPTH);
	for (b=0, index=0; b<32; ++b) {
		for (g=0; g<32; ++g) {
			for (r=0; r<32; ++r, ++index) {
				m_appMap[index] = g_monitor.MapColor(r / (real)31,
													 g / (real)31,
													 b / (real)31);
			}
		}
	}
	m_appMask = g_monitor.MapColor(0.4999, 0.4999, 0.4999);

	// Okay, now it's time to build the GBA alpha-blending table.  We
	// pass in (alpha << 15) | (color), and get a color back.
	for (a=0, index=0; a<15; ++a) {
		for (b=0; b<32; ++b) {
			for (g=0; g<32; ++g) {
				for (r=0; r<32; ++r, ++index) {
					u16 rr, gg, bb;

					rr = (r * (a + 1)) >> 4;
					gg = (g * (a + 1)) >> 4;
					bb = (b * (a + 1)) >> 4;
					CHECK((rr < 0x20) && (gg < 0x20) && (bb < 0x20));
					m_alphaTable[index] = (bb << 10) |
										  (gg << 5) |
										  (rr);
				}
			}
		}
	}

	s_dest.Allocate(240, 160, __FILE__, __LINE__);
	s_dest.Clear(0);
	m_nextEvent = E_HBLANK;
	g_ticksTilNextVideo = HORIZONTAL_TOTAL_PERIOD - 
						  HORIZONTAL_RENDER_PERIOD;
	for (m=0; m<128; ++m) {
		s_internals[m].flags = IS_DISABLED;
		s_internals[m].index = m;
	}
	BuildSpriteIndirectionTables();
	m_scanline = 0;

	for (n=p=0; n<2; ++n) {
		for (m=0; m<2; ++m, ++p) {
			s_standardRotations[p][0][1] = 0;
			s_standardRotations[p][1][0] = 0;
			if (m) {
				s_standardRotations[p][0][0] = -0x100;
			} else {
				s_standardRotations[p][0][0] = 0x100;
			}
			if (n) {
				s_standardRotations[p][1][1] = -0x100;
			} else {
				s_standardRotations[p][1][1] = 0x100;
			}
		}
	}
	g_globalEnable = ENABLE_OAM |
					 ENABLE_BG0 |
					 ENABLE_BG1 |
					 ENABLE_BG2 |
					 ENABLE_BG3;
}

/////////////////////////////////////////////////////////////////////////////
//  Make a video processor.
void VideoProcessor::SetParent(GBA * parent) {
	MType m;
	m_parent = parent;

	m_vidreg = (U16 *)g_readMap[VIDEO_REGISTER_SEG];
	m_pal = (GBAColor *)g_readMap[PALETTE_SEG];
	m_vram = g_readMap[VRAM_SEG];
	m_sprites = (GBSprite *)g_readMap[OAM_SEG];
	m_spritePal = (GBAColor *)(g_readMap[PALETTE_SEG] + 0x200);
	m_spriteData = (U8 *)(g_readMap[VRAM_SEG] + 0x10000);

	// We will insertion-sort our list of sprites every scanline, in case
	// somebody's doing some kind of funky stuff at hblank.  Let's form
	// the table of pointers here.
	for (m=0; m<128; ++m) {
		s_sprites[m] = m_sprites + m;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This function will make a priority-sorted list of the BG layers.
void VideoProcessor::UpdateBGLayers(const u8 v) {
	// HEREHEREHERE in the future, we may consider a difference check.
	if (1) {
		MType m;
		u16 u;
		u16 w = m_vidreg[LCDC];
		u8 mode = w & 7;
		static u8 s_rotateMode[6][4] = {
			{ BG_NRM, BG_NRM, BG_NRM, BG_NRM },
			{ BG_NRM, BG_NRM, BG_ROT, BG_OFF },
			{ BG_OFF, BG_OFF, BG_ROT, BG_ROT },
			{ BG_OFF, BG_OFF, BG_DIR, BG_OFF },
			{ BG_OFF, BG_OFF, BG_PAL, BG_OFF },
			{ BG_OFF, BG_OFF, BG_DIR, BG_OFF }
		};
	
		s_BGBuckets[0] = s_BGBuckets[1] =
		s_BGBuckets[2] = s_BGBuckets[3] = NULL;
		for (m=3, u=(u16)1 << 11; TRUE; --m, u>>=1) {
			BGInfo * bg = s_backgrounds + m;
			S16 * inf;

			if (m == 2) {
				inf = (S16 *)(m_vidreg + 0x10);
			} else {
				inf = (S16 *)(m_vidreg + 0x18);
			}
			if (s_rotateMode[mode][m] == BG_ROT) {
				if (!v ||
					((*((S32 *)(inf + 4)) != bg->xBase) ||
					 (*((S32 *)(inf + 6)) != bg->yBase))) {
					bg->xVal = bg->xBase = *((S32 *)(inf + 4));
					bg->yVal = bg->yBase = *((S32 *)(inf + 6));
				}
			}
			if ((s_rotateMode[mode][m] > BG_OFF) &&
				(u & w & g_globalEnable)) {
				u16 bgStat = m_vidreg[BG0 + m];

				bg->pal256 = (bgStat >> 7) & 1;
				bg->mode = s_rotateMode[mode][m];
				bg->next = s_BGBuckets[bgStat & 3];
				s_BGBuckets[bgStat & 3] = bg;
				if (bg->mode < BG_PAL) {
					bg->mapAddress = (U16 *)(m_vram + ((bgStat & 0x1f00) << 3));
					bg->dataAddress = m_vram + ((bgStat & 0xc) << 12);
					if (bg->mode == BG_ROT) {
						bg->transform[0][0] = inf[0];
						bg->transform[0][1] = inf[1];
						bg->transform[1][0] = inf[2];
						bg->transform[1][1] = inf[3];
						bg->xMask =
						bg->yMask = (128 << ((bgStat >> 14) & 3)) - 1; 
						bg->shift = 7 + ((bgStat >> 14) & 3) - 3;
						bg->tiled = (bgStat >> 13) & 1;
						bg->rotate = TRUE;
					} else {
						u16 y = m_vidreg[SCY0 + m + m];
						u16 x = m_vidreg[SCX0 + m + m];

						bg->xMask = (256 << ((bgStat >> 14) & 1)) - 1;
						bg->yMask = (256 << ((bgStat >> 15) & 1)) - 1;
						bg->rotate = FALSE;
						bg->xBase = x & 0x1ff;
						bg->yBase = y & 0x1ff;
					}
					bg->transp = ((m_vidreg[BLDMOD] >> (8 + m)) & 1);
	
					// HEREHEREHERE in regards to mosaic, we have to find
					// out whether we're only dealing with powers of two!
				}
			}
			if (!m)
				break;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  My new approach to sprites relies on the fixed-point system for all
// sprites, as I think it can be done quickly enough.
void VideoProcessor::UpdateSprite(MType v) {
	InternalSprite & is = s_internals[v];
	GBSprite & s = *(s_sprites[v]);

	Bool pal256 = (s.ySHi >> 13) & 1;
	Bool alpha = (s.ySHi >> 10) & 1;
	Bool dble = (s.ySHi >> 9) & 1;
	Bool tform = (s.ySHi >> 8) & 1;
	u8 spal = s.prioTile >> 12;
	u16 sz = ((s.ySHi >> 14) << 2) | (s.xSLo >> 14);
	u16 xx = s.xSLo & 0x1ff;
	u16 yy = s.ySHi & 0xff;
	u16 prio = (s.prioTile >> PRIO_SHIFT) & PRIO_MASK;
	u16 tile = s.prioTile & 0x3ff;
	Bool xFlip = ((s.xSLo >> 12) & 1);
	Bool yFlip = ((s.xSLo >> 13) & 1);
	u8 rotation = ((s.xSLo >> 9) & 0x1f);

	// Quick checks for invisible sprites.
	if (((xx >= 240) && (xx < X_BECOMES_NEGATIVE)) ||
		((yy >= 160) && (yy < Y_BECOMES_NEGATIVE)) ||
		((yy >= Y_BECOMES_NEGATIVE) && (((yy + s_heights[sz] * (dble + 1)) & 0xff) >= Y_BECOMES_NEGATIVE)) ||
		((xx >= X_BECOMES_NEGATIVE) && (((xx + s_widths[sz] * (dble + 1)) & 0x1ff) >= X_BECOMES_NEGATIVE)) ||
		(dble && !tform)) {
		RemoveSpriteFromQueue(is);
		if (is.flags & IS_IN_QUEUE) {
			MType m;
			u8 buf = m_scanline & 1;

			for (m=0; m<s_currentSpriteCount[is.priority]; ++m) {
				if (s_currentSpriteList[is.priority][buf][m] == &is)
					break;
			}
			CHECK(m != s_currentSpriteCount[is.priority]);
			--s_currentSpriteCount[is.priority];
			for (; m<s_currentSpriteCount[is.priority]; ++m) {
				s_currentSpriteList[is.priority][buf][m] =
					s_currentSpriteList[is.priority][buf][m + 1];
			}
			is.flags &= ~IS_IN_QUEUE;
		}
		return;
	}

	is.flags &= IS_DISABLED;

	if (dble) {
		is.flags |= IS_DOUBLE;
		is.width = s_widths[sz] * 2;
		is.height = s_heights[sz] * 2;
	} else {
		is.flags &= ~IS_DOUBLE;
		is.width = s_widths[sz];
		is.height = s_heights[sz];
	}

	if (!pal256) {
		is.flags |= IS_4BIT;
	} else {
		is.flags &= ~IS_4BIT;
	}
	if (alpha) {
		is.flags |= IS_ALPHA;
	} else {
		is.flags &= ~IS_ALPHA;
	}
	is.sz = sz;
	if (tform) {
		is.tform = rotation;
	} else {
		is.tform = 0x80 + (yFlip << 1) + xFlip;
	}

	is.startX = xx;
	ReplaceSpriteInQueue(is, yy, m_scanline, prio);

	// These things do not mandate any kind of invalidation.
	is.pal = spal << 4;
	is.priority = prio;
	is.tile = tile;
	is.tileAddress = g_readMap[6] + 0x10000 + (tile << 5);
}

/////////////////////////////////////////////////////////////////////////////
//  We will update the OAM here.  My new approach does not rely on a cache.
void VideoProcessor::UpdateOAMLayers(const u8 v) {
	if (m_vidreg[LCDC] & 0x1000) {
		//  If we're bitwise identical, then nothing has changed.  HEREHEREHERE
		// we may be better off with a dirty bit. . .
		if (memcmp(s_spriteData, m_sprites, 8 * 128)) {
			MType m;

			for (m=0; m<128; ++m) {
				if (s_spriteData[m] != m_sprites[m]) {
					s_spriteData[m] = m_sprites[m];
					UpdateSprite(m);
				}
			}
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  We will combine this with the BG renderer to make an all-new renderer
// that goes priority-layer by priority-layer, rendering until we're all
// done.
void VideoProcessor::RenderItems(const u8 y) {
	u8 prio;
	GBAColor * spritePal = m_spritePal;
	MType n;
	MType k;
	GBAColor * sp;
	BGInfo * bg;
	S16 px, py;
	u16 bldmod = m_vidreg[BLDMOD];
	Bool alpha;
	s32 bgx, bgy;
	u8 sprtp = (m_vidreg[LCDC] >> 6) & 1;

	for (prio=0; prio<4; ++prio) {
		u8 curSprite = 0;
		u8 buf = y & 1;
		u8 newCount = 0;
		InternalSprite * is;
		InternalSprite * fromQueue = s_spritesEachLine[prio][y];
		GBTForm sq;
		SpriteImage * si;
		u8 wcomp, hcomp;
		u16 w2, h2;
		s16 tx, ty;
		u16 bx, by;
		u16 tile;
		U16 * mapAddress;
		U8 * rotAddress;
		u8 p;

		// First, render all sprites at this priority.  We do a merge-sort
		// here.
		while (1) {
			u16 w;
			u8 dstX;

			if (curSprite < s_currentSpriteCount[prio]) {
				is = s_currentSpriteList[prio][buf][curSprite];
			} else {
				is = NULL;
			}
			if (is) {
				if (fromQueue) {
					if (fromQueue->index < is->index) {
						is = fromQueue;
						fromQueue = fromQueue->next;
						is->flags |= IS_IN_QUEUE;
					} else {
						++curSprite;
					}
				} else {
					++curSprite;
				}
			} else {
				if (fromQueue) {
					is = fromQueue;
					fromQueue = fromQueue->next;
					is->flags |= IS_IN_QUEUE;
				} else {
					// All sprites from this line have been rendered into
					// the buffer.
					break;
				}
			}

			// Okay, we got the next sprite from the merge-sort.  Now, let's
			// render the sprite - if sprites are (globally) enabled.
			if (is->startY < Y_BECOMES_NEGATIVE)
				k = y - is->startY;
			else
				k = y + (256 - is->startY);
			if (m_vidreg[LCDC] & g_globalEnable & 0x1000) {
				sp = spritePal;
				if (is->flags & IS_4BIT) {
					sp += is->pal;
				}
				alpha = (is->flags & IS_ALPHA) ||
						((((bldmod >> 6) & 3) == 1) && (bldmod & 0x10));
	
				si = &s_spriteImages[sprtp][is->sz];
				wcomp = s_widths[is->sz];
				hcomp = s_heights[is->sz];
				w2 = (s_widths[is->sz] - 1) << 7;
				h2 = (s_heights[is->sz] - 1) << 7;
				si->data.Open(Map < U16 >::READ);
				w = is->width;
				py = (k - (is->height / 2)) * 2 + 1;
				px = -(w / 2) * 2 + 1;
				if (is->startX < X_BECOMES_NEGATIVE) {
					if (is->startX + w > 240) {
						w = 240 - is->startX;
					}
					dstX = is->startX;
				} else {
					px += (512 - is->startX) * 2;
					w -= 512 - is->startX;
					dstX = 0;
				}
				if (is->tform & 0x80) {
					sq = s_standardRotations[is->tform & 3];
				} else {
					sq[0][0] = m_sprites[is->tform * 4 + 0].scaleRot;
					sq[0][1] = m_sprites[is->tform * 4 + 1].scaleRot;
					sq[1][0] = m_sprites[is->tform * 4 + 2].scaleRot;
					sq[1][1] = m_sprites[is->tform * 4 + 3].scaleRot;
				}
				tx = (sq[0][0] * px + sq[0][1] * py) >> 1;
				ty = (sq[1][0] * px + sq[1][1] * py) >> 1;
				for (n=0; n<w; ++n, ++dstX) {
					u16 fx = (u16)(tx + w2) >> 8;
					u16 fy = (u16)(ty + h2) >> 8;
	
					if ((fx < wcomp) && (fy < hcomp)) {
						u16 ind = si->data[fy][fx];
	
						if (is->flags & IS_4BIT) {
							p = is->tileAddress[ind >> 1];
							if (ind & 1)
								p >>= 4;
							p &= 0xf;
						} else {
							p = is->tileAddress[ind];
						}
						if (p) {
							if (!s_filled0[dstX]) {
								s_filled0[dstX] = TRUE;
								++s_pixelCount;
								s_scanline0[dstX] = sp[p];
								if (alpha) {
									s_alpha0[dstX] = TRUE;
								} else {
									s_filled1[dstX] = TRUE;
									s_alpha0[dstX] = FALSE;
									++s_pixelCount;
								}
							} else if (!s_filled1[dstX]) {
								s_filled1[dstX] = TRUE;
								++s_pixelCount;
								s_scanline1[dstX] = sp[p];
							}
						}
					}
					tx += sq[0][0];
					ty += sq[1][0];
				}
				si->data.Close();
			}

			// That rendered the sprite.  Now, if the sprite is still active,
			// re-insert it into the next queue.
			if ((y < 159) && (k < is->height - 1)) {
				s_currentSpriteList[prio][buf^1][newCount++] = is;
			} else {
				is->flags &= ~IS_IN_QUEUE;
			}
		}
		s_currentSpriteCount[prio] = newCount;

		// The above rendered all sprites at this priority level, in their
		// proper order, and updated the sprite queue for the next scanline.
		// Now, we do all of the background layers!
		for (bg = s_BGBuckets[prio]; bg; bg=bg->next) {
			alpha = bg->transp;
			switch(bg->mode) {
				case BG_NRM:
					by = (y + bg->yBase) & bg->yMask;
					mapAddress = bg->mapAddress +
								 (((by >> 3) & 0x1f) << 5) +
								 (((by >> 3) & 0x20) << 6);
					by &= 7;
					for (n=0; n<240; ++n) {
						if (!s_filled0[n] || !s_filled1[n]) {
							u16 tpal;
							u16 bbx, bby;
		
							bx = (n + bg->xBase) & bg->xMask;
							bbx = bx >> 3;
							tile = mapAddress[((bbx & 0x20) << 5) +
											  (bbx & 0x1f)];
							bx &= 7;
							if ((tile >> 11) & 1) {
								bby = 7 - by;
							} else {
								bby = by;
							}
							if ((tile >> 10) & 1) {
								bx = 7 - bx;
							}
							if (bg->pal256) {
								tile = (tile & 0x3ff) << 6;	// Doesn't seem right. . .
								p = bg->dataAddress[tile + (bby << 3) + bx];
								tpal = 0;
							} else {
								tpal = (tile >> 8) & 0xf0;
		
								tile = (tile & 0x3ff) << 5;
								p = bg->dataAddress[tile + (bby << 2) + (bx >> 1)];
								if (bx & 1) {
									p >>= 4;
								}
								p &= 0xf;
							}
							if (p) {
								if (!s_filled0[n]) {
									s_filled0[n] = TRUE;
									if (alpha) {
										s_alpha0[n] = TRUE;
										++s_pixelCount;
									} else {
										s_filled1[n] = TRUE;
										s_alpha0[n] = FALSE;
										s_pixelCount += 2;
									}
									s_scanline0[n] = m_pal[p + tpal];
									if (s_pixelCount == s_screenWidth)
										goto EarlyOut;
								} else if (!s_filled1[n]) { // We know that s_filled1[n] must be FALSE.
									s_filled1[n] = TRUE;
									s_scanline1[n] = m_pal[p + tpal];
									++s_pixelCount;
									if (s_pixelCount == s_screenWidth)
										goto EarlyOut;
								}
							}
						}
					}
					break;
				case BG_ROT:
					// Here's how it works:  we apply the transform to
					// the current coordinates.  Then we see if tiling is
					// enabled.  If it is, we mask the coordinates and
					// proceed with our rendering.  If it's not, we first
					// check for clipping.
					bgx = bg->xVal;
					bgy = bg->yVal;
					bg->xVal += (s32)bg->transform[0][1];
					bg->yVal += (s32)bg->transform[1][1];
					for (n=0;
						 n<240;
						 ++n,
						 bgx += bg->transform[0][0],
						 bgy += bg->transform[1][0]) {
						if (!s_filled0[n] || !s_filled1[n]) {
							u16 bbx;

							bx = (bgx >> 8);
							by = (bgy >> 8);
							if (bg->tiled) {
								bx &= bg->xMask;
								by &= bg->yMask;
							} else {
								if ((bx >= bg->xMask) || (by >= bg->yMask))
									continue;
							}
							bbx = bx >> 3;

							rotAddress = (U8 *)bg->mapAddress + ((by >> 3) << bg->shift);
							tile = (u16)rotAddress[bbx & 0x7f] << 6;
							bx &= 7;
							by &= 7;
							p = bg->dataAddress[tile + (by << 3) + bx];
							if (p) {
								if (!s_filled0[n]) {
									s_filled0[n] = TRUE;
									if (alpha) {
										s_alpha0[n] = TRUE;
										++s_pixelCount;
									} else {
										s_filled1[n] = TRUE;
										s_alpha0[n] = FALSE;
										s_pixelCount += 2;
									}
									s_scanline0[n] = m_pal[p];
									if (s_pixelCount == s_screenWidth)
										goto EarlyOut;
								} else { // We know that s_filled1[n] must be FALSE.
									s_filled1[n] = TRUE;
									s_scanline1[n] = m_pal[p];
									++s_pixelCount;
									if (s_pixelCount == s_screenWidth)
										goto EarlyOut;
								}
							}
						}
					}
					break;
				case BG_PAL: {
					U8 * vram = (U8 *)m_vram;

					if (m_vidreg[LCDC] & 0x10)
						vram += 256 * 160;
					vram += (MType)y * 240;
					for (n=0; n<240; ++n, ++vram) {
						if (p = *vram) {
							if (!s_filled0[n]) {
								s_filled0[n] = TRUE;
								s_scanline0[n] = m_pal[p];
								s_alpha0[n] = FALSE;
								s_pixelCount += 2;
								if (s_pixelCount == s_screenWidth)
									goto EarlyOut;
							} else if (!s_filled1[n]) {
								s_filled1[n] = TRUE;
								s_scanline1[n] = m_pal[p];
								++s_pixelCount;
								if (s_pixelCount == s_screenWidth)
									goto EarlyOut;
							}
						}
					}
					break;
				}
				case BG_DIR: {
					U16 * vram = (U16 *)m_vram;
					u8 lb, ub;

					if ((m_vidreg[LCDC] & 7) == 5) {
						if (m_vidreg[LCDC] & 0x10)
							vram += 160 * 128;
						vram += (MType)y * 160;
						lb = 40;
						ub = 200;
					} else {
						vram += (MType)y * 240;
						lb = 0;
						ub = 240;
					}

					for (n=lb; n<ub; ++n, ++vram) {
						if (!s_filled0[n]) {
							s_filled0[n] = TRUE;
							s_scanline0[n] = *vram;
							s_alpha0[n] = FALSE;
							s_pixelCount += 2;
							if (s_pixelCount == s_screenWidth)
								goto EarlyOut;
						} else if (!s_filled1[n]) {
							s_filled1[n] = TRUE;
							s_scanline1[n] = *vram;
							++s_pixelCount;
							if (s_pixelCount == s_screenWidth)
								goto EarlyOut;
						}
					}
					break;
				}
			}
		}
	}
EarlyOut:
		;
}

/////////////////////////////////////////////////////////////////////////////
//  Finally:  we've done our background and our sprites, now it's time to
// accumulate the results.
void VideoProcessor::AccumulateResults(const u8 v) {
	Color * pal;
	u8 u;
	u32 j;
	u32 i = ((u32)m_vidreg[COLY] & 0x1f);
	u16 avgTo;

	if (m_target != TARGET_FULLSCREEN) {
		pal = m_debugMap;
	} else {
		pal = m_appMap;
	}
	switch ((m_vidreg[BLDMOD] >> 6) & 3) {
		case 0:
			i = 0;
			break;
		case 1:
			i = 0;
			break;
		case 2:
			avgTo = 0x7fff;
			break;
		case 3:
			avgTo = 0;
			break;
	}
	if (i >= 0x10) {
		for (u=0; u<240; ++u) {
			s_dest[v][u] = pal[avgTo];
		}
	} else if (i) {
		--i;
		j = (14 - i) << 15;
		i <<= 15;
		for (u=0; u<240; ++u) {
			u16 c = s_scanline0[u] & 0x7fff;

			if (avgTo) {
				s_dest[v][u] = pal[c + m_alphaTable[i + avgTo - c]];
			} else {
				s_dest[v][u] = pal[m_alphaTable[j + c]];
			}
		}
	} else {
		i = (u32)(m_vidreg[COLEV] >> 8) & 0x1f;
		if (!((m_vidreg[BLDMOD] >> 6) & 3))
			i = 0x10;
		// HEREHEREHERE have to deal with the case where i+j is not 16.
		if (i >= 0x10) {
			for (u=0; u<240; ++u) {
				s_dest[v][u] = pal[s_scanline0[u] & 0x7fff];
			}
		} else if (i) {
			--i;
			j = (14 - i) << 15;
			i <<= 15;
			for (u=0; u<240; ++u) {
				if (s_alpha0[u] && s_filled1[u]) {
					s_dest[v][u] = pal[m_alphaTable[j + (s_scanline0[u] & 0x7fff)] +
									   m_alphaTable[i + (s_scanline1[u] & 0x7fff)]];
				} else {
					s_dest[v][u] = pal[s_scanline0[u] & 0x7fff];
				}
			}
		} else {
			for (u=0; u<240; ++u) {
				if (s_alpha0[u] && s_filled1[u]) {
					s_dest[v][u] = pal[s_scanline1[u] & 0x7fff];
				} else {
					s_dest[v][u] = pal[s_scanline0[u] & 0x7fff];
				}
			}
		}
	}
}

//static Bool s_debug;

/////////////////////////////////////////////////////////////////////////////
//  This function is used to render modes 0, 1, and 2.
void VideoProcessor::RenderScanline(const u8 scanline) {
	u8 u;
	u8 curline = scanline;

//if (!s_debug)
//	return;

	if ((m_vidreg[LCDC] & 7) == 5) {
		if ((scanline < 16) || (scanline >= 144))
			return;
		curline -= 16;
	}
	s_dest.Open(Map < U16 >::WRITE);
	for (u=0; u<240; ++u) {
		s_filled0[u] = s_filled1[u] = s_alpha0[u] = FALSE;
		s_scanline0[u] = m_pal[0];
	}
	s_pixelCount = 0;
	switch(*g_grid) {
		case 0:
			break;
		case 1:
			if (!(scanline & 0xf)) {
				for (u=0; u<240; ++u) {
					s_filled0[u] = s_filled1[u] = TRUE;
					s_scanline0[u] = 0x1f;
					s_pixelCount = 480;
				}
			} else {
				for (u=0; u<240; u+=16) {
					s_filled0[u] = s_filled1[u] = TRUE;
					s_scanline0[u] = 0x1f;
					s_pixelCount = 30;
				}
			}
			break;
		case 2:
			if (!(scanline & 7)) {
				for (u=0; u<240; ++u) {
					s_filled0[u] = s_filled1[u] = TRUE;
					s_scanline0[u] = 0x1f;
					s_pixelCount = 480;
				}
			} else {
				for (u=0; u<240; u+=8) {
					s_filled0[u] = s_filled1[u] = TRUE;
					s_scanline0[u] = 0x1f;
					s_pixelCount = 60;
				}
			}
			break;
	}
	if (s_pixelCount != 480) {
		UpdateBGLayers(curline);
		UpdateOAMLayers(curline);
		RenderItems(curline);
	}
	AccumulateResults(scanline);
	s_dest.Close();
}

/////////////////////////////////////////////////////////////////////////////
//  Specify the rendering target.
void VideoProcessor::SetTarget(u8 target) {
	if ((m_target = target) == TARGET_WINDOW) {
		RenderOptions ro;

		s_dest.Render(g_appWindow->GetAppReceptor(),
					  0,
					  0,
					  ro);
		g_appWindow->GenerateContent(AppWindow::TAB_APP);
	}
}

/////////////////////////////////////////////////////////////////////////////
//	This is obsolete.
void VideoProcessor::Redraw(void) {
}

/////////////////////////////////////////////////////////////////////////////
//  Tick the video processor.  HEREHEREHERE when we set the LCY value, we
// have to go through the history system.
void VideoProcessor::Tick(void) {
	if (IsOn()) {
		u16 disp = m_vidreg[STAT];
		ARM & cpu = m_parent->GetCPU();

		disp &= ~(VIDEO_IN_COINCIDENCE |
				  VIDEO_IN_HBLANK |
				  VIDEO_IN_VBLANK);
		if (m_target == TARGET_WINDOW) {
			RenderOptions ro;

			switch(m_nextEvent) {
				case E_HBLANK:
					disp |= VIDEO_IN_HBLANK;
					while (!g_gba.GetDMA().Tick(2));
					if (disp & VIDEO_IRQ_HBLANK)
						cpu.IRQ(ARM::IRQ_HBLANK);
					m_nextEvent = E_HTOTAL;
					g_ticksTilNextVideo += HORIZONTAL_TOTAL_PERIOD -
										   HORIZONTAL_RENDER_PERIOD;
					break;
				case E_HTOTAL:
					++m_scanline;
					if (m_scanline == VBLANK_SCANLINE_START) {
						while (!g_gba.GetDMA().Tick(1));
						if (disp & VIDEO_IRQ_VBLANK)
							cpu.IRQ(ARM::IRQ_VBLANK);
					} else if (m_scanline == VBLANK_TOTAL_SCANLINES) {
						m_scanline = 0;
					}
					if (m_scanline >= VBLANK_SCANLINE_START) {
						disp |= VIDEO_IN_VBLANK;
						m_nextEvent = E_HTOTAL;
						g_ticksTilNextVideo += HORIZONTAL_TOTAL_PERIOD;
					} else {
						m_nextEvent = E_HBLANK;
						g_ticksTilNextVideo += HORIZONTAL_RENDER_PERIOD;
					}
					if (((disp >> 8) & 0xff) == m_scanline) {
						disp |= VIDEO_IN_COINCIDENCE;
						if (disp & VIDEO_IRQ_VCOUNT) {
							cpu.IRQ(ARM::IRQ_VCOUNT);
						}
					}
					break;
			}
			MType m;

			for (m=0; m<160; ++m) {
				RenderScanline(m);
			}
			s_dest.Render(g_appWindow->GetAppReceptor(),
						  0,
						  0,
						  ro);
			g_appWindow->GenerateContent(AppWindow::TAB_APP);
			m_vidreg[LCY] = m_scanline;
//s_debug = !s_debug;
		} else if (m_target == TARGET_FAST_WINDOW) {
			u8 fh = TextGetFontHeight(g_font);

			switch(m_nextEvent) {
				case E_HBLANK:
					disp |= VIDEO_IN_HBLANK;
					while (!g_gba.GetDMA().Tick(2));
					if (disp & VIDEO_IRQ_HBLANK)
						cpu.IRQ(ARM::IRQ_HBLANK);
					m_nextEvent = E_HTOTAL;
					g_ticksTilNextVideo += HORIZONTAL_TOTAL_PERIOD -
										   HORIZONTAL_RENDER_PERIOD;
					break;
				case E_HTOTAL:
					++m_scanline;
					if (m_scanline < VBLANK_SCANLINE_START) {
						RenderScanline(m_scanline);
					} else if (m_scanline == VBLANK_SCANLINE_START) {
						RenderOptions ro;
	
						while (!g_gba.GetDMA().Tick(1));
						if (disp & VIDEO_IRQ_VBLANK)
							cpu.IRQ(ARM::IRQ_VBLANK);
						s_dest.Render(g_monitor,
									  g_monitor.GetWidth() - EMU_WINDOW_WIDTH + 2,
									  g_monitor.GetHeight() - EMU_WINDOW_HEIGHT - STATUS_BAR_HEIGHT + fh,
									  ro);
						g_monitor.Refresh();
//s_debug = !s_debug;
						GBASoundUpdate();
					} else if (m_scanline == VBLANK_TOTAL_SCANLINES) {
						m_scanline = 0;
						RenderScanline(m_scanline);
					}
					if (m_scanline >= VBLANK_SCANLINE_START) {
						disp |= VIDEO_IN_VBLANK;
						m_nextEvent = E_HTOTAL;
						g_ticksTilNextVideo += HORIZONTAL_TOTAL_PERIOD;
					} else {
						m_nextEvent = E_HBLANK;
						g_ticksTilNextVideo += HORIZONTAL_RENDER_PERIOD;
					}
					if (((disp >> 8) & 0xff) == m_scanline) {
						disp |= VIDEO_IN_COINCIDENCE;
						if (disp & VIDEO_IRQ_VCOUNT) {
							cpu.IRQ(ARM::IRQ_VCOUNT);
						}
					}
					break;
			}
			m_vidreg[LCY] = m_scanline;
		} else {
			switch(m_nextEvent) {
				case E_HBLANK:
					disp |= VIDEO_IN_HBLANK;
					while (!g_gba.GetDMA().Tick(2));
					if (disp & VIDEO_IRQ_HBLANK)
						cpu.IRQ(ARM::IRQ_HBLANK);
					m_nextEvent = E_HTOTAL;
					g_ticksTilNextVideo += HORIZONTAL_TOTAL_PERIOD -
										   HORIZONTAL_RENDER_PERIOD;
					break;
				case E_HTOTAL:
					++m_scanline;
					if (m_scanline < VBLANK_SCANLINE_START) {
						RenderScanline(m_scanline);
					} else if (m_scanline == VBLANK_SCANLINE_START) {
						RenderOptions ro;
	
						while (!g_gba.GetDMA().Tick(1));
						if (disp & VIDEO_IRQ_VBLANK)
							cpu.IRQ(ARM::IRQ_VBLANK);
						s_dest.Render(g_monitor,
									  (320 - 240) / 2,
									  (240 - 160) / 2,
									  ro);
//s_debug = !s_debug;
						g_monitor.Refresh();
						GBASoundUpdate();
					} else if (m_scanline == VBLANK_TOTAL_SCANLINES) {
						m_scanline = 0;
						RenderScanline(m_scanline);
					}
					if (m_scanline >= VBLANK_SCANLINE_START) {
						disp |= VIDEO_IN_VBLANK;
						m_nextEvent = E_HTOTAL;
						g_ticksTilNextVideo += HORIZONTAL_TOTAL_PERIOD;
					} else {
						m_nextEvent = E_HBLANK;
						g_ticksTilNextVideo += HORIZONTAL_RENDER_PERIOD;
					}
					if (((disp >> 8) & 0xff) == m_scanline) {
						disp |= VIDEO_IN_COINCIDENCE;
						if (disp & VIDEO_IRQ_VCOUNT) {
							cpu.IRQ(ARM::IRQ_VCOUNT);
						}
					}
					break;
			}
			m_vidreg[LCY] = m_scanline;
		}
		m_vidreg[STAT] = disp;
	} else {
		m_vidreg[STAT] &= ~(VIDEO_IN_COINCIDENCE |
							VIDEO_IN_HBLANK |
							VIDEO_IN_VBLANK);

		// HEREHEREHERE we will have to handle this carefully when
		// it is off (when we update g_ticksTilNextVideo).
		g_ticksTilNextVideo = 0xffff;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Tick the video processor backwords.  HEREHEREHERE do we have to worry
// about the transition from off-on?  HEREHEREHERE we have to worry about
// proper cycle updating in the reverse-executor.
void VideoProcessor::TickBackwards(void) {
	if (IsOn()) {
		if (m_target == TARGET_WINDOW) {
			RenderOptions ro;

			switch(m_nextEvent) {
				case E_HBLANK:
					m_nextEvent = E_HTOTAL;
					g_ticksTilNextVideo = HORIZONTAL_RENDER_PERIOD;
					break;
				case E_HTOTAL:
					if (!m_scanline--) {
						m_scanline = VBLANK_TOTAL_SCANLINES - 1;
					}
					if (m_scanline < VBLANK_SCANLINE_START) {
						m_nextEvent = E_HBLANK;
						g_ticksTilNextVideo = HORIZONTAL_TOTAL_PERIOD -
											  HORIZONTAL_RENDER_PERIOD;
					} else {
						m_nextEvent = E_HTOTAL;
						g_ticksTilNextVideo = HORIZONTAL_TOTAL_PERIOD;
					}
					break;
			}
			MType m;

			for (m=0; m<160; ++m) {
				RenderScanline(m);
			}
			s_dest.Render(g_appWindow->GetAppReceptor(),
						  0,
						  0,
						  ro);
			g_appWindow->GenerateContent(AppWindow::TAB_APP);
		} else {
			u8 next = VBLANK_SCANLINE_START;

			switch(m_nextEvent) {
				case E_HBLANK:
					m_nextEvent = E_HTOTAL;
					g_ticksTilNextVideo = HORIZONTAL_RENDER_PERIOD;
					next = m_scanline + 1;
					break;
				case E_HTOTAL:
					if (!m_scanline--) {
						m_scanline = VBLANK_TOTAL_SCANLINES - 1;
						next = 0;
					}
					if (m_scanline < VBLANK_SCANLINE_START) {
						m_nextEvent = E_HBLANK;
						g_ticksTilNextVideo = HORIZONTAL_TOTAL_PERIOD -
											  HORIZONTAL_RENDER_PERIOD;
					} else {
						m_nextEvent = E_HTOTAL;
						g_ticksTilNextVideo = HORIZONTAL_TOTAL_PERIOD;
					}
					break;
			}
			if (next < VBLANK_SCANLINE_START) {
				RenderScanline(next);
				if (!next) {
					RenderOptions ro;

					s_dest.Render(g_monitor,
								  (320 - 240) / 2,
								  (240 - 160) / 2,
								  ro);
					g_monitor.Refresh();
				}
			}
		}
	} else {
		g_ticksTilNextVideo = 0xffff;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Inform the video processor that we're switching our time orientation.
void VideoProcessor::SetOrientation(u8 orientation) {
	if (m_orientation != orientation) {
		if (orientation == HISTORY_FORWARD) {
			switch(m_nextEvent) {
				case E_HBLANK:
					m_nextEvent = E_HTOTAL;
					g_ticksTilNextVideo = HORIZONTAL_TOTAL_PERIOD -
										   HORIZONTAL_RENDER_PERIOD -
										   g_ticksTilNextVideo;
					break;
				case E_HTOTAL:
					if (m_scanline >= VBLANK_SCANLINE_START) {
						m_nextEvent = E_HTOTAL;
						g_ticksTilNextVideo = HORIZONTAL_TOTAL_PERIOD -
											   g_ticksTilNextVideo;
					} else {
						m_nextEvent = E_HBLANK;
						g_ticksTilNextVideo = HORIZONTAL_RENDER_PERIOD -
											   g_ticksTilNextVideo;
					}
					break;
			}
		} else {
			switch(m_nextEvent) {
				case E_HBLANK:
					m_nextEvent = E_HTOTAL;
					g_ticksTilNextVideo = HORIZONTAL_RENDER_PERIOD -
										   g_ticksTilNextVideo;
					break;
				case E_HTOTAL:
					if (m_scanline >= VBLANK_SCANLINE_START) {
						m_nextEvent = E_HTOTAL;
						g_ticksTilNextVideo = HORIZONTAL_TOTAL_PERIOD -
											   g_ticksTilNextVideo;
					} else {
						m_nextEvent = E_HBLANK;
						g_ticksTilNextVideo = HORIZONTAL_TOTAL_PERIOD -
											   HORIZONTAL_RENDER_PERIOD -
											   g_ticksTilNextVideo;
					}
					break;
			}
		}
		m_orientation = orientation;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This function makes a screenshot.  For now, we just do one, HEREHEREHERE
// we have to make this auto-increment and we should probably make a save-
// frames option for movie-making.
void VideoMakeScreenShot(void) {
	s_dest.Export("screen.bmp");
}
