/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	The implementation of our first window.

****************************************************************************/

#include "gba.h"

static TString s_helpFile("help_eng.txt");
static TString s_disassembly("Disassembly");
static TString s_source("Source");
static TString s_load("Load");
static TString s_options("Options");
static TString s_warnings("Warnings");
static TString s_user("User");
static TString s_help("Help");

static Sticky < FString > s_curDir("ROM search directory", "./");
Sticky < u8 > g_cursorDisplayOption("Cursor display option", 0);
Sticky < Bool > g_debugLog("Log user messages to debug.txt", FALSE);

/////////////////////////////////////////////////////////////////////////////
//  Make the code window.
CodeWindow::CodeWindow(void) {
	SetXYWH(0,
			0,
			g_monitor.GetWidth(),
			g_monitor.GetHeight() - EMU_WINDOW_HEIGHT - STATUS_BAR_HEIGHT);

	m_help.Open(s_helpFile);

	AddPanel(s_disassembly, &m_disassembly);
	AddPanel(s_load, 		&m_load);
	AddPanel(s_options,		&m_options);
	AddPanel(s_warnings,	&m_warnings);
	AddPanel(s_user,		&m_user);
	AddPanel(s_help, 		&m_help);
}

/////////////////////////////////////////////////////////////////////////////
//  Mark any text windows as being completely dirty upon entry.
void CodeWindow::SetTab(u8 tab) {
	switch(tab) {
		case TAB_DISASSEMBLY:
			m_disassembly.Goto(g_gba.GetCPU().GetCurrentIP());
			break;
		case TAB_LOAD:
			m_load.MakeDirty();
			break;
		case TAB_OPTIONS:
			m_options.MakeDirty();
			break;
		case TAB_WARNINGS:
			m_warnings.MakeDirty();
			break;
		case TAB_USER:
			m_user.MakeDirty();
			break;
		case TAB_HELP:
			m_help.MakeDirty();
			break;
	}
	Window::SetTab(tab);
}

/////////////////////////////////////////////////////////////////////////////
//  The user wants to single-step the CPU.
void CodeWindow::SingleStep(void) {
	m_disassembly.Goto(g_gba.GetCPU().GetCurrentIP());
}

/////////////////////////////////////////////////////////////////////////////
//  The user wants to go to the indicated location.
void CodeWindow::Goto(u32 line) {
	switch(m_selection) {
		case TAB_DISASSEMBLY:
			m_disassembly.Goto(line);
			break;
		case TAB_WARNINGS:
			m_warnings.Goto(line);
			break;
		case TAB_USER:
			m_user.Goto(line);
			break;
		case TAB_HELP:
			m_help.Goto(line);
			break;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Figure out the IP that the cursor presently points to.
u32 CodeWindow::GetCurrentCursorIP(void) {
	switch(m_selection) {
		case TAB_DISASSEMBLY:
			return m_disassembly.GetLine();
			break;
		default:
			return 0xffffffff;
			break;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  A text generator that displays the contents of a file.
void FileTextPanel::Open(const FString & filename) {
	FILE * f;
	char buf[1024];

	m_data.Destroy();
	DirFormFilename(buf, filename);
	if (!(f = fopen(buf, "r"))) {
		Warn("Unable to open %s.\n", buf);
	} else {
		while (!feof(f)) {
			m_data += fgets(buf, 1024, f);
		}
		fclose(f);
	}
	m_lineNumber = 0;
	GenerateContent();
}

/////////////////////////////////////////////////////////////////////////////
//  Redraw the file text panel.
void FileTextPanel::Redraw(void) {
	MType m;
	u16 x = m_rect.x;
	u16 inc = TextGetFontWidth(g_font) * 4;

	for (m=0; m<TEXT_NUM_TABS; ++m) {
		TextSetTab(m, x);
		x += inc;
	}
	TextPanel::Redraw();
}

/////////////////////////////////////////////////////////////////////////////
//  This function will build a list of all of the files in the indicated
// directory.
void BrowserPanel::FindAllFiles(const FString & s) {
	FString t = *s_curDir;

	if (s[0] == DIR_DRIVE) {
		if (DirFindAll(m_data, s)) {
			*s_curDir = s;
			m_lineNumber = 0;
			GenerateContent();
		}
	} else {
		DirAppend(t, s);
		if (DirFindAll(m_data, t)) {
			*s_curDir = t;
			m_lineNumber = 0;
			GenerateContent();
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This gives the user a list of files/directories to browse.
BrowserPanel::BrowserPanel(void) {
	if (!DirExists(*s_curDir)) {
		*s_curDir = DirGetAppPath();
	}
	FindAllFiles(".");
}

/////////////////////////////////////////////////////////////////////////////
//  Handle clicking in the browser panel.
void BrowserPanel::Click(u16 x, u16 y) {
	MType m;

	y -= m_rect.y;
	y /= TextGetFontHeight(g_font);
	m = y;
	if (m_content[m][0] == DIR_DIRECTORY) {
		FindAllFiles(m_content[m]);
	} else if (m_content[m][0] == DIR_DRIVE) {
		FindAllFiles(m_content[m]);
	} else {
		FString f = *s_curDir;

		DirAppend(f, m_content[m]);
		g_gba.Load(f);
		m_parent->SetTab(CodeWindow::TAB_DISASSEMBLY);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  When the user clicks in the disassembly panel, we toggle a breakpoint.
void DisassemblyPanel::Click(u16 x, u16 y) {
	u32 addr;
	u32 v = g_gba.GetCPU().GetInstructionSize();

	addr = y - m_rect.y;
	addr /= TextGetFontHeight(g_font);
	addr *= v;
	addr += m_address;
	switch(*g_cursorDisplayOption) {
		case 0:
			break;
		case 1:
			addr -= 10 * v;
			break;
		case 2:
			addr -= (m_content.Count() / 2) * v;
			break;
	}

	BreakpointToggle(addr);
}

/////////////////////////////////////////////////////////////////////////////
//  We have to set a tab stop for the disassembly panel.
void DisassemblyPanel::Redraw(void) {
	u8 v = g_gba.GetCPU().GetInstructionSize();
	u16 fw = TextGetFontWidth(g_font);
	u16 indent = fw * (25 + v * 2);

	TextSetTab(0, indent);
	TextPanel::Redraw();
}

/////////////////////////////////////////////////////////////////////////////
//  Make the disassembly panel.
void DisassemblyPanel::SetDims(u16 x, u16 y, u16 w, u16 h) {
	TextPanel::SetDims(x, y, w, h);
	m_address = g_gba.GetCPU().GetCurrentIP();
	m_bottom = m_address + m_content.Count() * 4;
}

/////////////////////////////////////////////////////////////////////////////
//  Generate content for the disassembly window.
void DisassemblyPanel::GenerateContent(void) {
	MType m;
	u32 u = m_address;
	u8 v = g_gba.GetCPU().GetInstructionSize();

	switch(*g_cursorDisplayOption) {
		case 0:
			break;
		case 1:
			u -= 10 * v;
			break;
		case 2:
			u -= (m_content.Count() / 2) * v;
			break;
	}

	for (m=0; m<m_content.Count(); ++m, u+=v) {
		char buf[32];
		Bool varyColor = TRUE;

		if (u == m_address) {
			if (BreakpointIsActive(u, BP_EXEC)) {
				m_content[m] = TEXT_COLOR_PURPLE;
			} else {
				m_content[m] = TEXT_COLOR_YELLOW;
			}
			varyColor = FALSE;
		} else {
			if (BreakpointIsActive(u, BP_EXEC)) {
				m_content[m] = TEXT_COLOR_BLUE;
				varyColor = FALSE;
			} else {
				m_content[m] = TEXT_COLOR_WHITE;
			}
		}
		g_gba.GetSymbol(m_content[m], u);
		m_content[m] += ":  ";
		if (v == 2) {
			if (varyColor) {
				m_content[m] += (wchar_t)TEXT_COLOR_GREEN;
			}
			sprintf(buf, "%04lx  ", (unsigned long)g_gba.GetMemory().GetW(u));
		} else {
			if (varyColor) {
				m_content[m] += (wchar_t)TEXT_COLOR_RED;
			}
			sprintf(buf, "%08lx  ", (unsigned long)g_gba.GetMemory().GetD(u));
		}
		m_content[m] += buf;
		g_gba.GetCPU().Decode(m_content[m], u);
	}
	m_bottom = u;
	MakeDirty();
}

/////////////////////////////////////////////////////////////////////////////
//  Move the cursor up in the disassembly window.
void DisassemblyPanel::CursorUp(void) {
	m_address -= g_gba.GetCPU().GetInstructionSize();
	GenerateContent();
}

/////////////////////////////////////////////////////////////////////////////
//  Move the cursor down.
void DisassemblyPanel::CursorDown(void) {
	m_address += g_gba.GetCPU().GetInstructionSize();
	GenerateContent();
}

/////////////////////////////////////////////////////////////////////////////
//  Move up one page.
void DisassemblyPanel::PageUp(void) {
	u32 old = m_address;

	m_address -= g_gba.GetCPU().GetInstructionSize() *
				 (m_content.Count() - 4);
	GenerateContent();
	while (m_bottom < old) {
		CursorDown();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Move down one page.
void DisassemblyPanel::PageDown(void) {
	m_address = m_bottom;
	GenerateContent();
}

/////////////////////////////////////////////////////////////////////////////
//  Jump to the indicated position.
void DisassemblyPanel::Goto(u32 line) {
	m_address = line;
	GenerateContent();
}

/////////////////////////////////////////////////////////////////////////////
//  Create an options panel.
OptionsPanel::OptionsPanel(void) {
	m_curRow = 2;
	m_curCol = 4;
	m_maxWidth = 0;
	#if !defined(RELEASE_BUILD)
		AddGUIElement(&m_language);
	#endif
	AddGUIElement(&m_registerFormat);
	AddGUIElement(&m_lineCount);
	AddGUIElement(&m_memoryFormat);
	AddGUIElement(&m_runAppModeSwitch);
	AddGUIElement(&m_debugLog);
	AddGUIElement(&m_runAppWindow);
	AddGUIElement(&m_skipStartup);
	AddGUIElement(&m_useThumbSyntax);
	AddGUIElement(&m_useMnemonic);
	AddGUIElement(&m_soundEnable);
	AddGUIElement(&m_resolution);
}

/////////////////////////////////////////////////////////////////////////////
//  The options panel deals with GUI elements in a special way.
void OptionsPanel::ResizeGUIElement(GUIElement * g) {
	u16 mh = m_rect.h;
	u16 fh = TextGetFontHeight(g_font);

	if (m_curRow + g->GetHeight() + fh > mh) {
		m_curCol += m_maxWidth + 2 * TextGetFontWidth(g_font);
		m_maxWidth = 0;
		m_curRow = 2;
	}
	g->SetX(m_rect.x + m_curCol);
	g->SetY(m_rect.y + m_curRow);
	m_curRow += g->GetHeight() + fh;
	if (g->GetWidth() > m_maxWidth)
		m_maxWidth = g->GetWidth();
	Panel::AddGUIElement(g);
}

/////////////////////////////////////////////////////////////////////////////
//  When we set the OptionsPanel size, we have to resize our GUI elements.
void OptionsPanel::SetDims(u16 x, u16 y, u16 w, u16 h) {
	Panel::SetDims(x, y, w, h);
	#if !defined(RELEASE_BUILD)
		ResizeGUIElement(&m_language);
	#endif
	ResizeGUIElement(&m_registerFormat);
	ResizeGUIElement(&m_lineCount);
	ResizeGUIElement(&m_memoryFormat);
	ResizeGUIElement(&m_runAppModeSwitch);
	ResizeGUIElement(&m_debugLog);
	ResizeGUIElement(&m_runAppWindow);
	ResizeGUIElement(&m_skipStartup);
	ResizeGUIElement(&m_useThumbSyntax);
	ResizeGUIElement(&m_useMnemonic);
	ResizeGUIElement(&m_soundEnable);
	ResizeGUIElement(&m_resolution);
}

/////////////////////////////////////////////////////////////////////////////
//  HEREHEREHERE we should allow a cursor to move around from one GUI element
// to the next.
void OptionsPanel::CursorUp(void) {
}

/////////////////////////////////////////////////////////////////////////////
//  HEREHEREHERE we should allow a cursor to move around from one GUI element
// to the next.
void OptionsPanel::CursorDown(void) {
}

/////////////////////////////////////////////////////////////////////////////
//  HEREHEREHERE we should allow a cursor to move around from one GUI element
// to the next.
void OptionsPanel::PageUp(void) {
}

/////////////////////////////////////////////////////////////////////////////
//  HEREHEREHERE we should allow a cursor to move around from one GUI element
// to the next.
void OptionsPanel::PageDown(void) {
}

/////////////////////////////////////////////////////////////////////////////
//  No content outside of the GUI elements.
void OptionsPanel::GenerateContent(void) {
	MakeDirty();
}

/////////////////////////////////////////////////////////////////////////////
//  Redraw the options panel.
void OptionsPanel::Redraw(void) {
	if (IsDirty()) {
		g_monitor.Clear(g_colors[BACKGROUND], &m_rect);
		CleanFrame();
	}
	Panel::Redraw();
}

#if !defined(RELEASE_BUILD)
	#define NUM_LANGUAGES			8
	
	static TString s_languageTitle("LANGUAGE");
	static TString s_languageEnglish("English");
	static TString s_languageFrench("French");
	static TString s_languageGerman("German");
	static TString s_languageItalian("Italian");
	static TString s_languageRussian("Russian");
	static TString s_languageChinese("Chinese");
	static TString s_languageJapanese("Japanese");
	static TString s_languageArabic("Arabic");
	u8 g_languageChoice;
	u8 g_language;
	/////////////////////////////////////////////////////////////////////////////
	//  Make the Language radio box.
	OptionsPanel::Language::Language(void) {
		u16 fh = TextGetFontHeight(g_font);
	
		SetTitle(s_languageTitle);
		SetDims(5, 5 + fh, NUM_LANGUAGES);
		SetCaption(0, s_languageEnglish);
		SetCaption(1, s_languageFrench);
		SetCaption(2, s_languageGerman);
		SetCaption(3, s_languageItalian);
		SetCaption(4, s_languageRussian);
		SetCaption(5, s_languageChinese);
		SetCaption(6, s_languageJapanese);
		SetCaption(7, s_languageArabic);
		m_selection = g_languageChoice = g_language;
	}
	
	/////////////////////////////////////////////////////////////////////////////
	//  Which language is presently selected?  Note:  not necessarily the active
	// language, just the language that will be used on startup.  We may want to
	// give the user the option to restart in the new language.
	u8 OptionsPanel::Language::QuerySelect(void) {
		return g_languageChoice;
	}
	
	/////////////////////////////////////////////////////////////////////////////
	//  Select the specified language.
	void OptionsPanel::Language::Select(u8 item) {
		CHECK(item < NUM_LANGUAGES);
		g_languageChoice = item;
	}
	
	/////////////////////////////////////////////////////////////////////////////
	//  No special behavior needed for a deselect.
	void OptionsPanel::Language::Deselect(u8 item) {
	}
#endif

#define NUM_RESOLUTIONS					5
static TString s_resolutionTitle("RESOLUTION");
static TString s_resolution640x480("640x480");
static TString s_resolution800x600("800x600");
static TString s_resolution1024x768("1024x768");
static TString s_resolution1280x1024("1280x1024");
static TString s_resolution1600x1200("1600x1200");
OptionsPanel::Resolution::Resolution(void) {
	u16 fh = TextGetFontHeight(g_font);

	SetTitle(s_resolutionTitle);
	SetDims(5, 5 + fh, NUM_RESOLUTIONS);
	SetCaption(0, s_resolution640x480);
	SetCaption(1, s_resolution800x600);
	SetCaption(2, s_resolution1024x768);
	SetCaption(3, s_resolution1280x1024);
	SetCaption(4, s_resolution1600x1200);
	m_selection = QuerySelect();
}

/////////////////////////////////////////////////////////////////////////////
//  Determine the current selection.
u8 OptionsPanel::Resolution::QuerySelect(void) {
	switch(*g_userVideoWidth) {
		default:
		case 640:
			return 0;
			break;
		case 800:
			return 1;
			break;
		case 1024:
			return 2;
			break;
		case 1280:
			return 3;
			break;
		case 1600:
			return 4;
			break;
	}
}

static TString s_resWontChange1("Note:  Resolution changes won't take");
static TString s_resWontChange2("place until you restart AGD.");

/////////////////////////////////////////////////////////////////////////////
//  Select the indicated item.
void OptionsPanel::Resolution::Select(u8 item) {
	if (item != m_selection) {
		CHECK(item < NUM_RESOLUTIONS);
		switch(item) {
			case 0:
				*g_userVideoWidth = 640;
				*g_userVideoHeight = 480;
				break;
			case 1:
				*g_userVideoWidth = 800;
				*g_userVideoHeight = 600;
				break;
			case 2:
				*g_userVideoWidth = 1024;
				*g_userVideoHeight = 768;
				break;
			case 3:
				*g_userVideoWidth = 1280;
				*g_userVideoHeight = 1024;
				break;
			case 4:
				*g_userVideoWidth = 1600;
				*g_userVideoHeight = 1200;
				break;
		}
		DialogDeliverMessage(s_resWontChange1,
							 s_resWontChange2);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Don't do anything in a deselect.
void OptionsPanel::Resolution::Deselect(u8 item) {
}

static TString s_lineCountTitle("SOURCE/DISASSEMBLY CURSOR");
static TString s_lcOption1("Display no lines before cursor");
static TString s_lcOption2("Display 10 lines before cursor");
static TString s_lcOption3("Display cursor in middle of page");

/////////////////////////////////////////////////////////////////////////////
//  Make the line-count radio box.
OptionsPanel::LineCount::LineCount(void) {
	SetTitle(s_lineCountTitle);
	SetDims(5, 124, 3);
	SetCaption(0, s_lcOption1);
	SetCaption(1, s_lcOption2);
	SetCaption(2, s_lcOption3);
	m_selection = QuerySelect();
}

/////////////////////////////////////////////////////////////////////////////
//  Which code display option is active.
u8 OptionsPanel::LineCount::QuerySelect(void) {
	return *g_cursorDisplayOption;
}

/////////////////////////////////////////////////////////////////////////////
//  Select the indicated option.
void OptionsPanel::LineCount::Select(u8 option) {
	*g_cursorDisplayOption = option;
}

/////////////////////////////////////////////////////////////////////////////
//  Deselect the indicated option.
void OptionsPanel::LineCount::Deselect(u8 option) {
}

#define NUM_REGISTER_FORMATS			7
#define NUM_MEMORY_FORMATS				21
static TString s_registerFormatTitle("REGISTER FORMAT");
static TString s_memoryFormatTitle("MEMORY FORMAT");
static TString s_fbHex("Unsigned Byte Hex");
static TString s_fbDec("Unsigned Byte Decimal");
static TString s_fbBin("Unsigned Byte Binary");
static TString s_fhHex("Unsigned Halfword Hex");
static TString s_fhDec("Unsigned Halfword Decimal");
static TString s_fhBin("Unsigned Halfword Binary");
static TString s_fHex("Unsigned Word Hex");
static TString s_fDec("Unsigned Word Decimal");
static TString s_fBin("Unsigned Word Binary");
static TString s_fsbHex("Signed Byte Hex");
static TString s_fsbDec("Signed Byte Decimal");
static TString s_fsbBin("Signed Byte Binary");
static TString s_fshHex("Signed Halfword Hex");
static TString s_fshDec("Signed Halfword Decimal");
static TString s_fshBin("Signed Halfword Binary");
static TString s_fsHex("Signed Word Hex");
static TString s_fsDec("Signed Word Decimal");
static TString s_fsBin("Signed Word Binary");
static TString s_fText("ASCII Text");
static TString s_fwText("Wide-character Text");
static TString s_fFloat("32-bit float");
/////////////////////////////////////////////////////////////////////////////
//  Make the register format radio box.
OptionsPanel::RegisterFormat::RegisterFormat(void) {
	u16 fh = TextGetFontHeight(g_font);

	SetTitle(s_registerFormatTitle);
	SetDims(5, 124 + fh, NUM_REGISTER_FORMATS);
	SetCaption(0, s_fHex);
	SetCaption(1, s_fDec);
	SetCaption(2, s_fBin);
	SetCaption(3, s_fsHex);
	SetCaption(4, s_fsDec);
	SetCaption(5, s_fsBin);
	SetCaption(6, s_fFloat);
	m_selection = QuerySelect();
}

/////////////////////////////////////////////////////////////////////////////
//  Which language is presently selected?  Note:  not necessarily the active
// language, just the language that will be used on startup.  We may want to
// give the user the option to restart in the new language.
u8 OptionsPanel::RegisterFormat::QuerySelect(void) {
	u8 retval;

	if (*g_registerFormat == MFLOAT) {
		retval = 6;
	} else {
		switch(*g_registerFormat & (MDECIMAL | MBINARY)) {
			default:
			case 0:
				retval = 0;
				break;
			case MDECIMAL:
				retval = 1;
				break;
			case MBINARY:
				retval = 2;
				break;
		}
		if (*g_registerFormat & MSIGNED) {
			retval += 3;
		}
	}
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Select the specified language.
void OptionsPanel::RegisterFormat::Select(u8 item) {
	static u8 s_formats[] = {
		MWORD,
		MWORD | MDECIMAL,
		MWORD | MBINARY,
		MWORD | MSIGNED,
		MWORD | MDECIMAL | MSIGNED,
		MWORD | MBINARY | MSIGNED,
		MFLOAT
	};
	CHECK(item < NUM_REGISTER_FORMATS);
	if (s_formats[item] != *g_registerFormat) {
		*g_registerFormat = s_formats[item];
		g_appWindow->GenerateContent(AppWindow::TAB_REGISTERS);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  No special behavior needed for a deselect.
void OptionsPanel::RegisterFormat::Deselect(u8 item) {
}

/////////////////////////////////////////////////////////////////////////////
//  Make the register format radio box.
OptionsPanel::MemoryFormat::MemoryFormat(void) {
	u16 fh = TextGetFontHeight(g_font);

	SetTitle(s_memoryFormatTitle);
	SetDims(5, 230 + fh, NUM_MEMORY_FORMATS);
	SetCaption(0, s_fbHex);
	SetCaption(1, s_fbDec);
	SetCaption(2, s_fbBin);
	SetCaption(3, s_fhHex);
	SetCaption(4, s_fhDec);
	SetCaption(5, s_fhBin);
	SetCaption(6, s_fHex);
	SetCaption(7, s_fDec);
	SetCaption(8, s_fBin);
	SetCaption(9, s_fsbHex);
	SetCaption(10, s_fsbDec);
	SetCaption(11, s_fsbBin);
	SetCaption(12, s_fshHex);
	SetCaption(13, s_fshDec);
	SetCaption(14, s_fshBin);
	SetCaption(15, s_fsHex);
	SetCaption(16, s_fsDec);
	SetCaption(17, s_fsBin);
	SetCaption(18, s_fText);
	SetCaption(19, s_fwText);
	SetCaption(20, s_fFloat);
	m_selection = QuerySelect();
}

/////////////////////////////////////////////////////////////////////////////
//  Which language is presently selected?  Note:  not necessarily the active
// language, just the language that will be used on startup.  We may want to
// give the user the option to restart in the new language.
u8 OptionsPanel::MemoryFormat::QuerySelect(void) {
	u8 retval;
	u8 w = *g_memoryFormat;

	switch(w) {
		case MTEXT:
			retval = 18;
			break;
		case MWTEXT:
			retval = 19;
			break;
		case MFLOAT:
			retval = 20;
			break;
		default:
			if (w & MSIGNED) {
				retval = 9;
			} else {
				retval = 0;
			}
			retval += (w >> 3) & 3;
			retval += (w & 3) * 3;
			break;
	}
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Select the specified language.
void OptionsPanel::MemoryFormat::Select(u8 item) {
	static u8 s_formats[] = {
		MBYTE,
		MBYTE | MDECIMAL,
		MBYTE | MBINARY,
		MHALFWORD,
		MHALFWORD | MDECIMAL,
		MHALFWORD | MBINARY,
		MWORD,
		MWORD | MDECIMAL,
		MWORD | MBINARY,
		MBYTE | MSIGNED,
		MBYTE | MDECIMAL | MSIGNED,
		MBYTE | MBINARY | MSIGNED,
		MHALFWORD | MSIGNED,
		MHALFWORD | MDECIMAL | MSIGNED,
		MHALFWORD | MBINARY | MSIGNED,
		MWORD | MSIGNED,
		MWORD | MDECIMAL | MSIGNED,
		MWORD | MBINARY | MSIGNED,
		MTEXT,
		MWTEXT,
		MFLOAT
	};
	CHECK(item < NUM_MEMORY_FORMATS);
	if (s_formats[item] != *g_memoryFormat) {
		*g_memoryFormat = s_formats[item];
		g_dataWindow->GenerateContent(DataWindow::TAB_MEMORY);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  No special behavior needed for a deselect.
void OptionsPanel::MemoryFormat::Deselect(u8 item) {
}

static TString s_appModeSwitchTitle("Run app full-screen");

/////////////////////////////////////////////////////////////////////////////
//  The AppModeSwitch checkbox specifies whether we want to do a mode switch
// when we run the app.
OptionsPanel::AppModeSwitch::AppModeSwitch(void) {
	u16 fh = TextGetFontHeight(g_font);

	SetTitle(s_appModeSwitchTitle);
	SetSpacing(1);
}

/////////////////////////////////////////////////////////////////////////////
//  Is the box checked?
Bool OptionsPanel::AppModeSwitch::QuerySelect(void) {
	return *g_runAppModeSwitch;
}

/////////////////////////////////////////////////////////////////////////////
//  Check the box.
void OptionsPanel::AppModeSwitch::Select(void) {
	*g_runAppModeSwitch = TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//  Uncheck the box.
void OptionsPanel::AppModeSwitch::Deselect(void) {
	*g_runAppModeSwitch = FALSE;
}

static TString s_debugLogTitle("Log user debug output");

/////////////////////////////////////////////////////////////////////////////
//  The DebugLog checkbox specifies whether we want to do a mode switch
// when we run the app.
OptionsPanel::DebugLog::DebugLog(void) {
	u16 fh = TextGetFontHeight(g_font);

	SetTitle(s_debugLogTitle);
	SetSpacing(1);
}

/////////////////////////////////////////////////////////////////////////////
//  Is the box checked?
Bool OptionsPanel::DebugLog::QuerySelect(void) {
	return *g_debugLog;
}

/////////////////////////////////////////////////////////////////////////////
//  Check the box.
void OptionsPanel::DebugLog::Select(void) {
	*g_debugLog = TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//  Uncheck the box.
void OptionsPanel::DebugLog::Deselect(void) {
	*g_debugLog = FALSE;
}

static TString s_appWindowedModeTitle("Run AGD in a window");

/////////////////////////////////////////////////////////////////////////////
//  Do we want to run AGD in a window.
OptionsPanel::AppModeWindow::AppModeWindow(void) {
	u16 fh = TextGetFontHeight(g_font);

	SetTitle(s_appWindowedModeTitle);
	SetSpacing(1);
}

/////////////////////////////////////////////////////////////////////////////
//  Is this box checked?
Bool OptionsPanel::AppModeWindow::QuerySelect(void) {
	return *g_runAppWindow;
}

/////////////////////////////////////////////////////////////////////////////
//  Check the box.
void OptionsPanel::AppModeWindow::Select(void) {
	*g_runAppWindow = TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//  Uncheck it.
void OptionsPanel::AppModeWindow::Deselect(void) {
	*g_runAppWindow = FALSE;
}

static TString s_skipStartupTitle("Skip BIOS boot sequence");

/////////////////////////////////////////////////////////////////////////////
//  Do we wish to skip the BIOS boot sequence?
OptionsPanel::SkipStartup::SkipStartup(void) {
	u16 fh = TextGetFontHeight(g_font);

	SetTitle(s_skipStartupTitle);
	SetSpacing(1);
}

/////////////////////////////////////////////////////////////////////////////
//  Is the box checked?
Bool OptionsPanel::SkipStartup::QuerySelect(void) {
	return *g_skipStartup;
}

/////////////////////////////////////////////////////////////////////////////
//  Check the box.
void OptionsPanel::SkipStartup::Select(void) {
	*g_skipStartup = TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//  Uncheck the box.
void OptionsPanel::SkipStartup::Deselect(void) {
	*g_skipStartup = FALSE;
}

static TString s_useThumbSyntaxTitle("Use Hx for hi-regs in Thumb disassembly");

/////////////////////////////////////////////////////////////////////////////
//  Should we use thumb-syntax, or ARM syntax, for hi-regs.
OptionsPanel::UseThumbSyntax::UseThumbSyntax(void) {
	SetTitle(s_useThumbSyntaxTitle);
	SetSpacing(1);
}

/////////////////////////////////////////////////////////////////////////////
//  Is the box checked?
Bool OptionsPanel::UseThumbSyntax::QuerySelect(void) {
	return *g_useThumbSyntax;
}

/////////////////////////////////////////////////////////////////////////////
//  Check the box.
void OptionsPanel::UseThumbSyntax::Select(void) {
	*g_useThumbSyntax = TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//  Uncheck the box.
void OptionsPanel::UseThumbSyntax::Deselect(void) {
	*g_useThumbSyntax = FALSE;
}

static TString s_useMnemonicTitle("Use mnemonic format for LDM/STM");

/////////////////////////////////////////////////////////////////////////////
//  Should we use thumb-syntax, or ARM syntax, for hi-regs.
OptionsPanel::UseMnemonic::UseMnemonic(void) {
	SetTitle(s_useMnemonicTitle);
	SetSpacing(1);
}

/////////////////////////////////////////////////////////////////////////////
//  Is the box checked?
Bool OptionsPanel::UseMnemonic::QuerySelect(void) {
	return *g_useMnemonic;
}

/////////////////////////////////////////////////////////////////////////////
//  Check the box.
void OptionsPanel::UseMnemonic::Select(void) {
	*g_useMnemonic = TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//  Uncheck the box.
void OptionsPanel::UseMnemonic::Deselect(void) {
	*g_useMnemonic = FALSE;
}

static TString s_soundEnableTitle("Enable sound");

/////////////////////////////////////////////////////////////////////////////
//  Should we use thumb-syntax, or ARM syntax, for hi-regs.
OptionsPanel::SoundEnable::SoundEnable(void) {
	SetTitle(s_soundEnableTitle);
	SetSpacing(1);
}

/////////////////////////////////////////////////////////////////////////////
//  Is the box checked?
Bool OptionsPanel::SoundEnable::QuerySelect(void) {
	return *g_soundEnable;
}

/////////////////////////////////////////////////////////////////////////////
//  Check the box.
void OptionsPanel::SoundEnable::Select(void) {
	*g_soundEnable = TRUE;
	GBASoundOn();
}

/////////////////////////////////////////////////////////////////////////////
//  Uncheck the box.
void OptionsPanel::SoundEnable::Deselect(void) {
	*g_soundEnable = FALSE;
	GBASoundOff();
}

/////////////////////////////////////////////////////////////////////////////
//  Add a warning to the list of warnings.
void UserWarningPanel::AddWarning(const IString & f) {
	m_data += f;
	Goto(0x80000000);
}

/////////////////////////////////////////////////////////////////////////////
//  Issue some sort of warning to the user.
void WarnUser(const IString & f) {
	IString i = TEXT_COLOR_YELLOW;

	i += f;
	g_codeWindow->AddWarning(i);
	g_codeWindow->SetTab(CodeWindow::TAB_WARNINGS);
}

static TString s_debugTextFailure("Unable to open debug.txt for appending.\n");

/////////////////////////////////////////////////////////////////////////////
//  Add a user message to the user panel.
void CodeWindow::AddUserMessage(const FString & f) {
	if (*g_debugLog) {
		FILE * g = fopen("debug.txt", "a");

		if (g) {
			char buf[256];

			CopyToChar(f, buf, 256);
			fprintf(g, "%s\n", buf);
			fclose(g);
		} else {
			WarnUser(s_debugTextFailure);
		}
	}
	m_user.InsertString(f);
}

