/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	The implementation of the data window.  The data window displays all of
the GBA data, including RAM, hardware internals, etc.

****************************************************************************/

#include "gba.h"

static char s_scratch[16];

Sticky < u8 > g_memoryFormat("Memory Format", MWORD);

#define KNOWN_STRINGS_COLOR				TEXT_COLOR_WHITE

/////////////////////////////////////////////////////////////////////////////
//  Figure out the width of the representation of this format.
u8 MemoryGetWidth(u8 format) {
	u8 retval;

	if (format < MTEXT) {
		switch (format & (MDECIMAL | MBINARY)) {
			case 0:
				switch(format & 3) {
					case MBYTE:
						retval = 2 + 2;
						break;
					case MHALFWORD:
						retval = 4 + 2;
						break;
					case MWORD:
						retval = 8 + 2;
						break;
				}
				break;
			case MDECIMAL:
				switch(format & 3) {
					case MBYTE:
						retval = 3 + 2;
						break;
					case MHALFWORD:
						retval = 5 + 2;
						break;
					case MWORD:
						retval = 10 + 2;
						break;
				}
				break;
			case MBINARY:
				switch(format & 3) {
					case MBYTE:
						retval = 8 + 2;
						break;
					case MHALFWORD:
						retval = 16 + 2;
						break;
					case MWORD:
						retval = 32 + 2;
						break;
				}
				break;
		}
		if (format & MSIGNED) {
			++retval;
		}
	} else switch(format) {
		case MTEXT:
			retval = 1;
			break;
		case MWTEXT:
			retval = 1;
			break;
		case MFLOAT:
			retval = 8 + 2;
			break;
	}
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Dump a binary representation of the indicated value into the indicated
// wchar_t buffer.
static void BDump(char * dest, u32 value, u8 nDigits) {
	value <<= 32 - nDigits;

	while(nDigits--) {
		*dest++ = (value & 0x80000000) ? '1' : '0';
		value += value;
	}
	*dest = 0;
}

/////////////////////////////////////////////////////////////////////////////
//  Give me a 32-bit value and a desired format, and I'll append the
// translation of the value into the format to the string.
void MemoryTranslate(FString & f, u32 value, u8 format) {
	char buf[64];
	Bool addSpaces = TRUE;
	u8 c;
	Bool dec;

	switch(format) {
		case MBYTE:
			sprintf(buf, "%02x", value);
			break;
		case MHALFWORD:
			sprintf(buf, "%04x", value);
			break;
		case MWORD:
			sprintf(buf, "%08x", value);
			break;
		case MSIGNED | MBYTE:
			if (value & 0x80) {
				value = -value & 0xff;
				sprintf(buf, "-%02x", value);
			} else {
				sprintf(buf, "+%02x", value);
			}
			break;
		case MSIGNED | MHALFWORD:
			if (value & 0x8000) {
				value = -value & 0xffff;
				sprintf(buf, "-%04x", value);
			} else {
				sprintf(buf, "+%04x", value);
			}
			break;
		case MSIGNED | MWORD:
			if (value & 0x80000000) {
				value = -value;
				sprintf(buf, "-%08x", value);
			} else {
				sprintf(buf, "+%08x", value);
			}
			break;
		case MDECIMAL | MBYTE:
			sprintf(buf, "%03u", value);
			break;
		case MDECIMAL | MHALFWORD:
			sprintf(buf, "%05u", value);
			break;
		case MDECIMAL | MWORD:
			sprintf(buf, "%010u", value);
			break;
		case MDECIMAL | MSIGNED | MBYTE:
			if (value & 0x80) {
				value = -value & 0xff;
				sprintf(buf, "-%03u", value);
			} else {
				sprintf(buf, "+%03u", value);
			}
			break;
		case MDECIMAL | MSIGNED | MHALFWORD:
			if (value & 0x8000) {
				value = -value & 0xffff;
				sprintf(buf, "-%05u", value);
			} else {
				sprintf(buf, "+%05u", value);
			}
			break;
		case MDECIMAL | MSIGNED | MWORD:
			if (value & 0x80000000) {
				value = -value;
				sprintf(buf, "-%010u", value);
			} else {
				sprintf(buf, "+%010u", value);
			}
			break;
		case MBINARY | MBYTE:
			BDump(buf, value, 8);
			break;
		case MBINARY | MHALFWORD:
			BDump(buf, value, 16);
			break;
		case MBINARY | MWORD:
			BDump(buf, value, 32);
			break;
		case MBINARY | MSIGNED | MBYTE:
			if (value & 0x80) {
				value = -value & 0xff;
				buf[0] = '-';
			} else {
				buf[0] = '+';
			}
			BDump(buf + 1, value, 8);
			break;
		case MBINARY | MSIGNED | MHALFWORD:
			if (value & 0x8000) {
				value = -value & 0xffff;
				buf[0] = '-';
			} else {
				buf[0] = '+';
			}
			BDump(buf + 1, value, 16);
			break;
		case MBINARY | MSIGNED | MWORD:
			if (value & 0x80000000) {
				value = -value;
				buf[0] = '-';
			} else {
				buf[0] = '+';
			}
			BDump(buf + 1, value, 32);
			break;
		case MWTEXT:
		case MTEXT:
			addSpaces = FALSE;
			if (value < ' ') {
				value = '.';
			}
			buf[0] = value;
			buf[1] = '\0';
			break;
		case MFLOAT:
			sprintf(buf, "%lg", *((float *)&value));
			dec = FALSE;
			for (c=0; buf[c] && (c<8); ++c) {
				if (buf[c] == '.')
					dec = TRUE;
			}
			// Number was too large to fit.
			if ((c == 8) && (buf[c]) && (!dec)) {
				buf[6] = '.';
				buf[7] = '.';
			}
			for (; c<8; ++c) {
				buf[c] = ' ';
			}
			buf[c] = '\0';
			break;
	}
	f += buf;
	if (addSpaces) {
		f += "  ";
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Set the dimensions of the memory panel.
void MemoryPanel::SetDims(u16 x, u16 y, u16 w, u16 h) {
	TextPanel::SetDims(x, y, w, h);
	m_address = 0x6000000;
	GenerateContent();
}

static TString s_LCDC("[LCDC     ");
static TString s_videoMode("Mode=");
static TString s_videoPage("Page=");
static TString s_hblankUpdate("OAM/HBL");
static TString s_1dSprites("1D sprites");
static TString s_2dSprites("2D sprites");
static TString s_bg0("BG0");
static TString s_bg1("BG1");
static TString s_bg2("BG2");
static TString s_bg3("BG3");
static TString s_OAM("OAM");
static TString * s_lcdcStrings[] = {
	&s_hblankUpdate,
	NULL,
	NULL,
	&s_bg0,
	&s_bg1,
	&s_bg2,
	&s_bg3,
	&s_OAM,
};

static TString s_LCDStat("[LCDSTAT  ");
static TString s_VBlank("VBL");
static TString s_HBlank("HBL");
static TString s_coincidence("LYC");
static TString s_VBlankIRQ("VBL IRQ");
static TString s_HBlankIRQ("HBL IRQ");
static TString s_coincidenceIRQ("LYC IRQ");
static TString s_coincidenceLine("LYC=");
static TString * s_lcdStatStrings[] = {
	&s_VBlank,
	&s_HBlank,
	&s_coincidence,
	&s_VBlankIRQ,
	&s_HBlankIRQ,
	&s_coincidenceIRQ,
};

static TString s_scanline("[Scanline ");

static TString s_BG0Name("[BG0      ");
static TString s_BG1Name("[BG1      ");
static TString s_BG2Name("[BG2      ");
static TString s_BG3Name("[BG3      ");
static TString * s_bgNames[] = {
	&s_BG0Name,
	&s_BG1Name,
	&s_BG2Name,
	&s_BG3Name,
};
static TString s_priority("PRI=");
static TString s_dataAddress("DAT=");
static TString s_mosaic("MOS");
static TString * s_bgStrings[] = {
	&s_mosaic,
};
static TString s_8BitColor("8BIT");
static TString s_4BitColor("4BIT");
static TString s_mapAddress("MAP=");
static TString s_tile("TILE");
static TString s_transp("TRANS");
static TString * s_rotStrings[] = {
	&s_tile,
};
static TString * s_normalStrings[] = {
	&s_transp,
};
static TString s_sizeString("SZ=");
static TString s_normal0("256x256");
static TString s_normal1("512x256");
static TString s_normal2("256x512");
static TString s_normal3("512x512");
static TString * s_normalSizes[] = {
	&s_normal0,
	&s_normal1,
	&s_normal2,
	&s_normal3,
};
static TString s_rot0("128x128");
static TString s_rot1("256x256");
static TString s_rot2("512x512");
static TString s_rot3("1024x1024");
static TString * s_rotSizes[] = {
	&s_rot0,
	&s_rot1,
	&s_rot2,
	&s_rot3,
};
static TString s_bg0SCX("[BG0 SCX  ");
static TString s_bg0SCY("[BG0 SCY  ");
static TString s_bg1SCX("[BG1 SCX  ");
static TString s_bg1SCY("[BG1 SCY  ");
static TString s_bg2SCX("[BG2 SCX  ");
static TString s_bg2SCY("[BG2 SCY  ");
static TString s_bg3SCX("[BG3 SCX  ");
static TString s_bg3SCY("[BG3 SCY  ");
static TString * s_bgScrollNames[] = {
	&s_bg0SCX,
	&s_bg0SCY,
	&s_bg1SCX,
	&s_bg1SCY,
	&s_bg2SCX,
	&s_bg2SCY,
	&s_bg3SCX,
	&s_bg3SCY,
};
static TString s_bg2Matrix("[BG2 TFORM");
static TString s_bg3Matrix("[BG3 TFORM");
static TString s_bg2XValue("[BG2 X    ");
static TString s_bg2YValue("[BG2 Y    ");
static TString s_bg3XValue("[BG3 X    ");
static TString s_bg3YValue("[BG3 Y    ");
static TString * s_bgXYNames[] = {
	&s_bg2XValue,
	&s_bg2YValue,
	&s_bg3XValue,
	&s_bg3YValue,
};
static TString s_mosaicName("[MOSAIC   ");
static TString s_horizBG("BG H-RPT=");
static TString s_vertBG("BG V-RPT=");
static TString s_horizOAM("OAM H-RPT=");
static TString s_vertOAM("OAM V-RPT=");

static TString s_bldmodName("[BLDMOD   ");
static TString s_bldmodBG0P0("B0");
static TString s_bldmodBG1P0("B1");
static TString s_bldmodBG2P0("B2");
static TString s_bldmodBG3P0("B3");
static TString s_bldmodOAMP0("SP");
static TString s_bldmodBD0("BD");
static TString s_pixel0("P0: ");
static TString s_pixel1("P1: ");
static TString * s_bldmodStrings[] = {
	&s_bldmodBG0P0,
	&s_bldmodBG1P0,
	&s_bldmodBG2P0,
	&s_bldmodBG3P0,
	&s_bldmodOAMP0,
	&s_bldmodBD0,
};
static TString s_bldmodNoEffect("NO FX");
static TString s_bldmodAlpha("BLEND");
static TString s_bldmodBrtInc("BRT+");
static TString s_bldmodBrtDec("BRT-");
static TString * s_bldmodType[] = {
	&s_bldmodNoEffect,
	&s_bldmodAlpha,
	&s_bldmodBrtInc,
	&s_bldmodBrtDec,
};
static TString s_colevName("[COLEV    ");
static TString s_alpha0Name("A0=");
static TString s_alpha1Name("A1=");
static TString s_colyName("[COLY     ");
static TString s_dmaS0("[DMA0 SRC ");
static TString s_dmaS1("[DMA1 SRC ");
static TString s_dmaS2("[DMA2 SRC ");
static TString s_dmaS3("[DMA3 SRC ");
static TString * s_dmaSourceAddress[] = {
	&s_dmaS0,
	&s_dmaS1,
	&s_dmaS2,
	&s_dmaS3,
};
static TString s_dmaD0("[DMA0 DEST");
static TString s_dmaD1("[DMA1 DEST");
static TString s_dmaD2("[DMA2 DEST");
static TString s_dmaD3("[DMA3 DEST");
static TString * s_dmaDestAddress[] = {
	&s_dmaD0,
	&s_dmaD1,
	&s_dmaD2,
	&s_dmaD3,
};
static TString s_dmaC0("[DMA0 CTRL");
static TString s_dmaC1("[DMA1 CTRL");
static TString s_dmaC2("[DMA2 CTRL");
static TString s_dmaC3("[DMA3 CTRL");
static TString * s_dmaCtrlAddress[] = {
	&s_dmaC0,
	&s_dmaC1,
	&s_dmaC2,
	&s_dmaC3,
};
static TString s_dmaCount("CNT=");
static TString s_srcTfr0("S+");
static TString s_srcTfr1("S-");
static TString s_srcTfr2("S=");
static TString * s_dmaSrcTfr[] = {
	&s_srcTfr0,
	&s_srcTfr1,
	&s_srcTfr2,
	&s_srcTfr2,
};
static TString s_dstTfr0("D+");
static TString s_dstTfr1("D-");
static TString s_dstTfr2("D=");
static TString * s_dmaDstTfr[] = {
	&s_dstTfr0,
	&s_dstTfr1,
	&s_dstTfr2,
	&s_dstTfr0,
};
static TString s_dmaPd0("IMM");
static TString s_dmaPd1("VBL");
static TString s_dmaPd2("HBL");
static TString s_dmaPd3("???");
static TString * s_prdString[] = {
	&s_dmaPd0,
	&s_dmaPd1,
	&s_dmaPd2,
	&s_dmaPd0,
};
static TString s_dmaRpt("RPT");
static TString * s_dmaStrings[] = {
	&s_dmaRpt,
};
static TString s_dwordString("[32]");
static TString s_wordString("[16]");
static TString s_enableString("ON");
static TString s_disableString("OFF");
static u32 s_srcMask[] = {
	0x7ffffff,
	0xfffffff,
	0xfffffff,
	0xfffffff,
};

static u32 s_dstMask[] = {
	0x7ffffff,
	0x7ffffff,
	0xfffffff,
	0xfffffff,
};
static u32 s_countMask[] = {
	0x7fff,
	0x7fff,
	0x7fff,
	0x1ffff
};
static real s_timeFactors[] = {
	1,
	64,
	256,
	1024,
};
static TString s_td0("[TMR0 CNT ");
static TString s_td1("[TMR1 CNT ");
static TString s_td2("[TMR2 CNT ");
static TString s_td3("[TMR3 CNT ");
static TString * s_timerData[] = {
	&s_td0,
	&s_td1,
	&s_td2,
	&s_td3,
};
static TString s_cascade("CASC");
static TString s_tc0("[TMR0 CTRL");
static TString s_tc1("[TMR1 CTRL");
static TString s_tc2("[TMR2 CTRL");
static TString s_tc3("[TMR3 CTRL");
static TString * s_timerCtrl[] = {
	&s_tc0,
	&s_tc1,
	&s_tc2,
	&s_tc3,
};
static TString s_freq0("1/CYC");
static TString s_freq1("64/CYC");
static TString s_freq2("256/CYC");
static TString s_freq3("1024/CYC");
static TString * s_frequencies[] = {
	&s_freq0,
	&s_freq1,
	&s_freq2,
	&s_freq3,
};
static TString s_irq("IRQ");
static TString * s_timerStrings[] = {
	&s_irq,
};
static TString s_milleSeconds("mSEC");
static TString s_keyA("A");
static TString s_keyB("B");
static TString s_keySel("SEL");
static TString s_keyStart("START");
static TString s_keyRight("RIGHT");
static TString s_keyLeft("LEFT");
static TString s_keyUp("UP");
static TString s_keyDown("DOWN");
static TString s_keyRShoulder("R");
static TString s_keyLShoulder("L");
static TString * s_keyStrings[] = {
	&s_keyA,
	&s_keyB,
	&s_keySel,
	&s_keyStart,
	&s_keyRight,
	&s_keyLeft,
	&s_keyUp,
	&s_keyDown,
	&s_keyRShoulder,
	&s_keyLShoulder,
};
static TString s_irqVBL("VB");
static TString s_irqHBL("HB");
static TString s_irqLYC("LYC");
static TString s_irqTim0("T0");
static TString s_irqTim1("T1");
static TString s_irqTim2("T2");
static TString s_irqTim3("T3");
static TString s_irqComm("COM");
static TString s_irqDMA0("D0");
static TString s_irqDMA1("D1");
static TString s_irqDMA2("D2");
static TString s_irqDMA3("D3");
static TString s_irqKey("KEY");
static TString s_irqCart("CART");
static TString * s_irqStrings[] = {
	&s_irqVBL,
	&s_irqHBL,
	&s_irqLYC,
	&s_irqTim0,
	&s_irqTim1,
	&s_irqTim2,
	&s_irqTim3,
	&s_irqComm,
	&s_irqDMA0,
	&s_irqDMA1,
	&s_irqDMA2,
	&s_irqDMA3,
	&s_irqKey,
	&s_irqCart,
};
static TString s_keyString("[KEYS     ");
static TString s_ieString("[IE       ");
static TString s_ifString("[IF       ");
static TString s_imeString("[IME      ");

/////////////////////////////////////////////////////////////////////////////
//  A little helper to scan a bitvector and insert the appropriate strings.
static void InsertStrings(FString & dest,
						  TString ** src,
						  u32 mask,
						  u32 low,
						  u32 high) {
	u8 cnt;
	u32 it;
	Bool b = FALSE;

	for (cnt=0, it=low; it && (it<high); it+=it, ++cnt) {
		if (src[cnt]) {
			if (b) {
				dest += "; ";
			}
			b = TRUE;
			if (mask & it) {
				dest += (wchar_t)TEXT_COLOR_GREEN;
			} else {
				dest += (wchar_t)TEXT_COLOR_RED;
			}
			dest += *(src[cnt]);
			dest += (wchar_t)KNOWN_STRINGS_COLOR;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This determines if the indicated memory address is translated (into high-
// level text).
static Bool IsTranslated(u32 u) {
	if ((u >> 24) == 4) {
		u &= g_memMask[4];
		switch(u) {
			case 0:
			case 4:
			case 6:
			case 8:
			case 0xa:
			case 0xc:
			case 0xe:
			case 0x10:
			case 0x12:
			case 0x14:
			case 0x16:
			case 0x18:
			case 0x1a:
			case 0x1c:
			case 0x1e:
			case 0x20:
			case 0x30:
			case 0x28:
			case 0x2c:
			case 0x38:
			case 0x3c:
			case 0x4c:
			case 0x50:
			case 0x52:
			case 0x54:
			case 0xb0:
			case 0xb4:
			case 0xb8:
			case 0xbc:
			case 0xc0:
			case 0xc4:
			case 0xc8:
			case 0xcc:
			case 0xd0:
			case 0xd4:
			case 0xd8:
			case 0xdc:
			case 0x100:
			case 0x102:
			case 0x104:
			case 0x106:
			case 0x108:
			case 0x10a:
			case 0x10c:
			case 0x10e:
			case 0x130:
			case 0x200:
			case 0x202:
			case 0x208:
				return TRUE;
				break;
		}
	}
	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////
//  Generate content for the memory panel.
void MemoryPanel::GenerateContent(void) {
	u32 u = m_address;
	MType m;
	u16 fw = TextGetFontWidth(g_font);
	u16 itemWidth = MemoryGetWidth(*g_memoryFormat) * fw;
	MType w, sw;
	u16 tw;
	u32 val, v2;
	Bool rot;
	u8 mode;
	Bool forceBreak;

	mode = g_gba.GetMemory().GetW(0x4000000) & 7;
	// Let's figure out the maximum power of 2 that we can display per-
	// scanline.
	for (sw=1, tw=fw*11 + itemWidth, m=itemWidth;
		 tw + m < m_rect.w;
		 tw += m, m+=m) {
		sw += sw;
	}

	forceBreak = FALSE;
	for (m=0; m<m_content.Count(); ++m) {
		if (IsTranslated(u)) {
			m_content[m] = KNOWN_STRINGS_COLOR;
		} else {
			m_content[m] = TEXT_COLOR_ORANGE;
		}
		sprintf(s_scratch, "%08lx ", (unsigned long)u);
		m_content[m] += s_scratch;

		// If this is a "special" address (for now, I/O), then we display
		// specific information for that location.
		if (IsTranslated(u)) {
			u16 mc = g_gba.GetMemory().GetW(u);

			rot = FALSE;
			switch(u & g_memMask[4]) {
				case 0:
					m_content[m] += s_LCDC;
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					m_content[m] += s_videoMode;
					ConvertFromHex(m_content[m], mc & 7, 1);
					m_content[m] += "; ";
					if ((mc & 7) > 3) {
						m_content[m] += s_videoPage;
						if (mc & 0x10) {
							m_content[m] += "1; ";
						} else {
							m_content[m] += "0; ";
						}
					}
					if (mc & 0x40) {
						m_content[m] += s_1dSprites;
					} else {
						m_content[m] += s_2dSprites;
					}
					m_content[m] += "; ";
					InsertStrings(m_content[m],
								  s_lcdcStrings,
								  mc,
								  0x20,
								  0x2000);
					u += 2;
					break;
				case 4:
					m_content[m] += s_LCDStat;
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					InsertStrings(m_content[m],
								  s_lcdStatStrings,
								  mc,
								  0x1,
								  0x40);
					if (mc & 0x20) {
						m_content[m] += "; ";
						m_content[m] += s_coincidenceLine;
						ConvertFromHex(m_content[m], (mc >> 8) & 0xff);
					}
					u += 2;
					break;
				case 6:
					m_content[m] += s_scanline;
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					ConvertFromHex(m_content[m], mc, 4);
					u += 2;
					break;
				case 0xc:
					if ((mode > 0) && (mode < 3)) {
						rot = TRUE;
					}
				case 0xe:
					if (mode == 2) {
						rot = TRUE;
					}
				case 8:
				case 0xa:
					m_content[m] += *(s_bgNames[((u & g_memMask[4]) - 8) / 2]);
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					m_content[m] += s_priority;
					ConvertFromHex(m_content[m], mc & 3, 1);
					m_content[m] += "; ";
					m_content[m] += s_dataAddress;
					ConvertFromHex(m_content[m],
								   0x6000000 + ((u32)((mc >> 2) & 3) << 14),
								   7);
					m_content[m] += "; ";
					InsertStrings(m_content[m],
								  s_bgStrings,
								  mc,
								  0x40,
								  0x80);
					m_content[m] += "; ";
					if (mc & 0x80) {
						m_content[m] += s_8BitColor;
					} else {
						m_content[m] += s_4BitColor;
					}
					m_content[m] += "; ";
					m_content[m] += s_mapAddress;
					ConvertFromHex(m_content[m],
								   0x6000000 + ((u32)(mc & 0x1f00) << 3),
								   7);
					m_content[m] += "; ";

					if (rot) {
						InsertStrings(m_content[m],
									  s_rotStrings,
									  mc,
									  0x2000,
									  0x4000);
						m_content[m] += "; ";
						m_content[m] += s_sizeString;
						m_content[m] += *(s_rotSizes[(mc >> 14) & 3]);
					} else {
						InsertStrings(m_content[m],
									  s_normalStrings,
									  mc ^ 0x2000,
									  0x2000,
									  0x4000);
						m_content[m] += "; ";
						m_content[m] += s_sizeString;
						m_content[m] += *(s_normalSizes[(mc >> 14) & 3]);
					}
					u += 2;
					break;
				case 0x10:
				case 0x12:
				case 0x14:
				case 0x16:
				case 0x18:
				case 0x1a:
				case 0x1c:
				case 0x1e:
					m_content[m] += *(s_bgScrollNames[((u & g_memMask[4]) - 0x10) / 2]);
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					ConvertFromHex(m_content[m], mc & 0x3ff, 3);
					u += 2;
					break;
				case 0x20:
				case 0x30:
					if (m + 1 == m_content.Count()) {
						forceBreak = TRUE;
					} else {
						S16 mat[4];

						if ((u & g_memMask[4]) == 0x20) {
							m_content[m] += s_bg2Matrix;
						} else {
							m_content[m] += s_bg3Matrix;
						}
						mat[0] = g_gba.GetMemory().GetW(u);
						mat[1] = g_gba.GetMemory().GetW(u + 2);
						mat[2] = g_gba.GetMemory().GetW(u + 4);
						mat[3] = g_gba.GetMemory().GetW(u + 6);
						m_content[m] += "        ] : ";
						m_content[m] += (char)(0x18);
						m_content[m] += " ";
						ConvertFromHex(m_content[m], mat[0], 4);
						m_content[m] += "  ";
						ConvertFromHex(m_content[m], mat[1], 4);
						m_content[m] += (char)(0x19);
						m_content[m] += "  ";
						m_content[m] += (char)(0x18);
						m_content[m] += " ";
						ConvertFromFloat(m_content[m], mat[0] / 256.0, 8);
						m_content[m] += "  ";
						ConvertFromFloat(m_content[m], mat[1] / 256.0, 8);
						m_content[m] += (char)(0x19);

						// That was the first row of the matrix, now let's
						// do the second row.
						m_content[m + 1] = KNOWN_STRINGS_COLOR;
						m_content[m + 1] += "                               ";
						m_content[m + 1] += (char)(0x1a);
						m_content[m + 1] += " ";
						ConvertFromHex(m_content[m + 1], mat[2], 4);
						m_content[m + 1] += "  ";
						ConvertFromHex(m_content[m + 1], mat[3], 4);
						m_content[m + 1] += (char)(0x1b);
						m_content[m + 1] += "  ";
						m_content[m + 1] += (char)(0x1a);
						m_content[m + 1] += " ";
						ConvertFromFloat(m_content[m + 1], mat[2] / 256.0, 8);
						m_content[m + 1] += "  ";
						ConvertFromFloat(m_content[m + 1], mat[3] / 256.0, 8);
						m_content[m + 1] += (char)(0x1b);
						++m;
						u += 8;
					}
					break;
				case 0x28:
				case 0x2c:
				case 0x38:
				case 0x3c:
					val = u & g_memMask[4];

					if (val == 0x28)
						val = 0;
					else if (val == 0x2c)
						val = 1;
					else if (val == 0x38)
						val = 2;
					else
						val = 3;
					m_content[m] += *(s_bgXYNames[val]);
					val = g_gba.GetMemory().GetD(u);
					ConvertFromHex(m_content[m], val);
					m_content[m] += "] : ";
					ConvertFromHex(m_content[m], val);
					m_content[m] += " = ";
					ConvertFromFloat(m_content[m], val / 256.0, 8);
					u += 4;
					break;
				case 0x4c:
					m_content[m] += s_mosaicName;
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					m_content[m] += s_horizBG;
					ConvertFromHex(m_content[m], (mc & 0xf), 1);
					m_content[m] += "; ";
					m_content[m] += s_vertBG;
					ConvertFromHex(m_content[m], ((mc >> 4) & 0xf), 1);
					m_content[m] += "; ";
					m_content[m] += s_horizOAM;
					ConvertFromHex(m_content[m], ((mc >> 8) & 0xf), 1);
					m_content[m] += "; ";
					m_content[m] += s_vertOAM;
					ConvertFromHex(m_content[m], ((mc >> 12) & 0xf), 1);
					u += 2;
					break;
				case 0x50:
					m_content[m] += s_bldmodName;
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					m_content[m] += s_pixel0;
					InsertStrings(m_content[m],
								  s_bldmodStrings,
								  mc,
								  0x1,
								  0x40);
					m_content[m] += "; ";
					m_content[m] += s_pixel1;
					InsertStrings(m_content[m],
								  s_bldmodStrings,
								  mc,
								  0x100,
								  0x4000);
					m_content[m] += "; ";
					m_content[m] += *(s_bldmodType[(mc >> 6) & 3]);
					u += 2;
					break;
				case 0x52:
					m_content[m] += s_colevName;
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					m_content[m] += s_alpha0Name;
					ConvertFromHex(m_content[m], mc & 0x1f, 2);
					m_content[m] += " (";
					if ((mc & 0x1f) >= 0x10) {
						ConvertFromFloat(m_content[m], 1.0, 5);
					} else {
						ConvertFromFloat(m_content[m], (real)(mc & 0x1f) / 16.0, 5);
					}
					m_content[m] += "); ";
					m_content[m] += s_alpha1Name;
					ConvertFromHex(m_content[m], (mc >> 8) & 0x1f, 2);
					m_content[m] += " (";
					if (((mc >> 8) & 0x1f) >= 0x10) {
						ConvertFromFloat(m_content[m], 1.0, 5);
					} else {
						ConvertFromFloat(m_content[m], (real)((mc >> 8) & 0x1f) / 16.0, 5);
					}
					m_content[m] += ")";
					u += 2;
					break;
				case 0x54:
					m_content[m] += s_colyName;
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					ConvertFromHex(m_content[m], mc & 0x1f, 2);
					m_content[m] += " (";
					if ((mc & 0x1f) >= 0x10) {
						ConvertFromFloat(m_content[m], 1.0, 5);
					} else {
						ConvertFromFloat(m_content[m], (real)(mc & 0x1f) / 16.0, 5);
					}
					m_content[m] += ")";
					u += 2;
					break;
				case 0xb0:
				case 0xbc:
				case 0xc8:
				case 0xd4:
					v2 = ((u & 0xff) - 0xb0) / 12;
					m_content[m] += *(s_dmaSourceAddress[v2]);
					val = g_gba.GetMemory().GetD(u);
					ConvertFromHex(m_content[m], val);
					m_content[m] += "] : ";
					ConvertFromHex(m_content[m], val & s_srcMask[v2]);
					u += 4;
					break;
				case 0xb4:
				case 0xc0:
				case 0xcc:
				case 0xd8:
					v2 = ((u & 0xff) - 0xb4) / 12;
					m_content[m] += *(s_dmaDestAddress[v2]);
					val = g_gba.GetMemory().GetD(u);
					ConvertFromHex(m_content[m], val);
					m_content[m] += "] : ";
					ConvertFromHex(m_content[m], val & s_dstMask[v2]);
					u += 4;
					break;
				case 0xb8:
				case 0xc4:
				case 0xd0:
				case 0xdc:
					v2 = ((u & 0xff) - 0xb8) / 12;
					m_content[m] += *(s_dmaCtrlAddress[v2]);
					val = g_gba.GetMemory().GetD(u);
					ConvertFromHex(m_content[m], val);
					m_content[m] += "] : ";
					m_content[m] += s_dmaCount;
					ConvertFromHex(m_content[m], val & s_countMask[v2], 5);
					m_content[m] += "; ";
					m_content[m] += *(s_dmaSrcTfr[(val >> 0x17) & 3]);
					m_content[m] += "; ";
					m_content[m] += *(s_dmaDstTfr[(val >> 0x15) & 3]);
					m_content[m] += "; ";
					m_content[m] += *(s_prdString[(val >> 0x1c) & 3]);
					m_content[m] += "; ";
					InsertStrings(m_content[m],
								  s_dmaStrings,
								  val,
								  0x2000000,
								  0x4000000);
					m_content[m] += "; ";
					if ((val >> 0x1a) & 1) {
						m_content[m] += s_dwordString;
					} else {
						m_content[m] += s_wordString;
					}
					m_content[m] += "; ";
					if ((val >> 0x1f) & 1) {
						m_content[m] += s_enableString;
					} else {
						m_content[m] += s_disableString;
					}
					u += 4;
					break;
				case 0x100:
				case 0x104:
				case 0x108:
				case 0x10c:
					v2 = (u & 0xf) / 4;
					m_content[m] += *(s_timerData[v2]);
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += " (";
					val = g_gba.GetMemory().GetW(u + 2);
					if ((u & 0xf) && (val & 0x4)) {
						m_content[m] += s_cascade;
					} else {
						real r = 0x10000 - mc;

						r *= s_timeFactors[val & 3];
						r *= 1.0 / (16777.216);
						ConvertFromFloat(m_content[m], r, 8);
						m_content[m] += " ";
						m_content[m] += s_milleSeconds;
					}
					m_content[m] += ")";
					u += 2;
					break;
				case 0x102:
				case 0x106:
				case 0x10a:
				case 0x10e:
					v2 = ((u & 0xf) - 2) / 4;
					m_content[m] += *(s_timerCtrl[v2]);
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					if (((u & 0xf) == 2) & (mc & 0x4)) {
						m_content[m] += s_cascade;
					} else {
						m_content[m] += *(s_frequencies[mc & 3]);
					}
					m_content[m] += "; ";
					InsertStrings(m_content[m],
								  s_timerStrings,
								  mc,
								  0x40,
								  0x80);
					m_content[m] += "; ";
					if (mc & 0x80) {
						m_content[m] += s_enableString;
					} else {
						m_content[m] += s_disableString;
					}
					u += 2;
					break;
				case 0x130:
					m_content[m] += s_keyString;
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					InsertStrings(m_content[m],
								  s_keyStrings,
								  ~mc,
								  0x1,
								  0x400);
					u += 2;
					break;
				case 0x200:
					m_content[m] += s_ieString;
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					InsertStrings(m_content[m],
								  s_irqStrings,
								  mc,
								  0x1,
								  0x4000);
					u += 2;
					break;
				case 0x202:
					m_content[m] += s_ifString;
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					InsertStrings(m_content[m],
								  s_irqStrings,
								  mc,
								  0x1,
								  0x4000);
					u += 2;
					break;
				case 0x208:
					m_content[m] += s_imeString;
					ConvertFromHex(m_content[m], mc, 4);
					m_content[m] += "    ] : ";
					if (mc) {
						m_content[m] += s_enableString;
					} else {
						m_content[m] += s_disableString;
					}
					u += 2;
					break;
			}
			if (forceBreak)
				break;
			continue;
		}
		for (w=0; w<sw; ++w) {
			u8 size;
			u32 value;

			if ((u >= 0x5000000) && (u < 0x5000400)) {
				u16 col = g_gba.GetMemory().GetW(u) & 0x7fff;

				m_content[m] += (wchar_t)TEXT_BGCOLOR_SPECIFIED;
				m_content[m] += (wchar_t)(col);
				m_content[m] += (wchar_t)TEXT_COLOR_SPECIFIED;
				m_content[m] += (wchar_t)(~col & 0x7fff);
			}
			switch(*g_memoryFormat) {
				case MTEXT:
					size = 1;
					break;
				case MWTEXT:
					size = 2;
					break;
				case MFLOAT:
					size = 4;
					break;
				default:
					size = 1 << (*g_memoryFormat & 3);
					break;
			}
			switch(size) {
				case 1:
					value = g_gba.GetMemory().GetB(u);
					u += 1;
					break;
				case 2:
					value = g_gba.GetMemory().GetW(u);
					u += 2;
					break;
				case 4:
					value = g_gba.GetMemory().GetD(u);
					u += 4;
					break;
			}
			MemoryTranslate(m_content[m], value, *g_memoryFormat);
			if (IsTranslated(u))
				break;
		}
	}
	m_bottom = u;
	MakeDirty();
}

/////////////////////////////////////////////////////////////////////////////
//  Move the cursor up in the memory panel.
void MemoryPanel::CursorUp(void) {
	m_address -= m_step * 4;
	GenerateContent();
}

/////////////////////////////////////////////////////////////////////////////
//  Move the cursor down.
void MemoryPanel::CursorDown(void) {
	m_address += m_step * 4;
	GenerateContent();
}

/////////////////////////////////////////////////////////////////////////////
//  Move up one page.
void MemoryPanel::PageUp(void) {
	m_address -= (m_bottom - m_address);
	GenerateContent();
}

/////////////////////////////////////////////////////////////////////////////
//  Move down one page.
void MemoryPanel::PageDown(void) {
	m_address = m_bottom;
	GenerateContent();
}

/////////////////////////////////////////////////////////////////////////////
//  Jump to the indicated position.
void MemoryPanel::Goto(u32 line) {
	m_address = line;
	GenerateContent();
}

/////////////////////////////////////////////////////////////////////////////
//  Specify the mask that this checkbox uses.
void BreakpointPanel::BreakCheck::SetMask(u8 mask) {
	m_mask = mask;
}

/////////////////////////////////////////////////////////////////////////////
//  Specify which breakpoint I'm interested in.
void BreakpointPanel::BreakCheck::SetBP(MType bp) {
	m_bp = bp;
}

/////////////////////////////////////////////////////////////////////////////
//  Is this item checked?
Bool BreakpointPanel::BreakCheck::QuerySelect(void) {
	CHECK(m_bp < g_userBreakpoints.Count());
	return g_userBreakpoints[m_bp].options & m_mask;
}

/////////////////////////////////////////////////////////////////////////////
//  Select this item.
void BreakpointPanel::BreakCheck::Select(void) {
	u8 u;
	Breakpoint & b = g_userBreakpoints[m_bp];

	u = b.options | m_mask;
//	BreakpointSet(b.base, b.range, b.countdown, u, BP_USER);
	BreakpointReset(m_bp, b.base, b.range, b.countdown, u);
	g_dataWindow->GenerateContent(DataWindow::TAB_BREAKPOINTS);
}

/////////////////////////////////////////////////////////////////////////////
//  Deselect this item.
void BreakpointPanel::BreakCheck::Deselect(void) {
	u8 u;
	Breakpoint & b = g_userBreakpoints[m_bp];

	u = b.options & ~m_mask;
//	BreakpointSet(b.base, b.range, b.countdown, u, BP_USER);
	BreakpointReset(m_bp, b.base, b.range, b.countdown, u);
	g_dataWindow->GenerateContent(DataWindow::TAB_BREAKPOINTS);
}

/////////////////////////////////////////////////////////////////////////////
//  Initialize my index.
BreakpointPanel::BreakpointPanel(void) {
	m_currentBreakpoint = 0;
}

/////////////////////////////////////////////////////////////////////////////
//  When the dims are specified, we'll know how many checkboxes we need.
void BreakpointPanel::SetDims(u16 x, u16 y, u16 w, u16 h) {
	MType m, n;
	u16 fw = TextGetFontWidth(g_font);
	u16 fh = TextGetFontHeight(g_font);
	u16 xx;
	u16 yy;

	TextPanel::SetDims(x, y, w, h);
	xx = m_rect.x;
	yy = m_rect.y;
	for (n=0; n<3; ++n) {
		m_checks[n].Make(m_content.Count() - 1, __FILE__, __LINE__);
	}
	for (m=0; m<m_content.Count() - 1; ++m) {
		m_checks[0][m].SetTitle("R");
		m_checks[0][m].SetMask(BP_READ);
		m_checks[1][m].SetTitle("W");
		m_checks[1][m].SetMask(BP_WRITE);
		m_checks[2][m].SetTitle("X");
		m_checks[2][m].SetMask(BP_EXEC);
		for (n=0; n<3; ++n) {
			m_checks[n][m].SetDims(xx + 3 * fw * n, yy + fh);
			AddGUIElement(&m_checks[n][m]);
			m_checks[n][m].Hide();
		}
		yy += fh;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Right-click on a breakpoint to delete it.
void BreakpointPanel::RClick(u16 x, u16 y) {
	u16 fh = TextGetFontHeight(g_font);
	MType m;

	m = (y - m_rect.y) / fh + m_currentBreakpoint;
	if (m && (m - 1 < g_userBreakpoints.Count())) {
		--m;
		BreakpointClearUser(m);
		GenerateContent();
		g_codeWindow->GenerateContent(CodeWindow::TAB_DISASSEMBLY);
	}
}

static MType s_cur;

/////////////////////////////////////////////////////////////////////////////
//  The user has specified a new range for the indicated breakpoint.
static void GetRange(void) {
	u32 u;

	if (u = ConvertToHex(g_userString)) {
		Breakpoint & b = g_userBreakpoints[s_cur];

		BreakpointReset(s_cur,
						b.base,
						u,
						b.countdown,
						b.options);
		g_dataWindow->GenerateContent(DataWindow::TAB_BREAKPOINTS);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  The user has specified a new range for the indicated breakpoint.
static void GetMaxCount(void) {
	u32 u;

	if (u = ConvertToHex(g_userString)) {
		Breakpoint & b = g_userBreakpoints[s_cur];

		BreakpointReset(s_cur,
						b.base,
						b.range,
						u,
						b.options);
		g_dataWindow->GenerateContent(DataWindow::TAB_BREAKPOINTS);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  The user has specified a new address for the indicated breakpoint.
static void GetAddress(void) {
	u32 u;
	Breakpoint & b = g_userBreakpoints[s_cur];

	u = g_gba.GetAddress(g_userString);
	BreakpointReset(s_cur,
					u,
					b.range,
					b.countdown,
					b.options);
	g_dataWindow->GenerateContent(DataWindow::TAB_BREAKPOINTS);
}

static TString s_getRange("Specify the range of this breakpoint.");
static TString s_getMaxCount("Specify the countdown for this breakpoint.");
static TString s_getAddress("Specify the address for this breakpoint.");

/////////////////////////////////////////////////////////////////////////////
//  Left-clicking on the breakpoint window enables you to edit its fields.
void BreakpointPanel::Click(u16 x, u16 y) {
	u16 fh = TextGetFontHeight(g_font);
	u16 fw = TextGetFontWidth(g_font);
	MType m;

	Panel::Click(x, y);
	m = (y - m_rect.y) / fh + m_currentBreakpoint;
	if (m && (m - 1 < g_userBreakpoints.Count())) {
		s_cur = --m;
		x = (x - m_rect.x) / fw;
		if ((x >= 9) && (x < 18)) {
			DialogGetString(s_getRange, GetRange);
		} else if ((x >= 27) && (x < 38)) {
			DialogGetString(s_getMaxCount, GetMaxCount);
		} else if ((x >= 38)) {
			DialogGetString(s_getAddress, GetAddress);
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Add a breakpoint to the breakpoint panel.  All user-generated breakpoints
// will originate here.
void BreakpointPanel::AddBreakpoint(u32 address, u32 range, u8 conditions) {
	BreakpointSet(address, range, 1, conditions, BP_USER);
	GenerateContent();
	g_codeWindow->GenerateContent(CodeWindow::TAB_DISASSEMBLY);
}

/////////////////////////////////////////////////////////////////////////////
//  Scroll up the list of breakpoints.
void BreakpointPanel::CursorUp(void) {
	if (m_currentBreakpoint) {
		--m_currentBreakpoint;
		GenerateContent();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Scroll down.
void BreakpointPanel::CursorDown(void) {
	if (m_currentBreakpoint + m_content.Count() - 1 <
		g_userBreakpoints.Count()) {
		++m_currentBreakpoint;
		GenerateContent();
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Go up a whole page.
void BreakpointPanel::PageUp(void) {
	if (m_currentBreakpoint > m_content.Count() - 1) {
		m_currentBreakpoint -= m_content.Count() - 1;
	} else {
		m_currentBreakpoint = 0;
	}
	GenerateContent();
}

/////////////////////////////////////////////////////////////////////////////
//  Go down a whole page.
void BreakpointPanel::PageDown(void) {
	if (m_currentBreakpoint + (m_content.Count() - 1) * 2 <
		g_userBreakpoints.Count()) {
		m_currentBreakpoint += m_content.Count() - 1;
		GenerateContent();
	} else if (g_userBreakpoints.Count() >= m_content.Count() - 1) {
		m_currentBreakpoint = g_userBreakpoints.Count() -
							  m_content.Count() +
							  1;
		GenerateContent();
	}
}

static TString s_breakpointHeader("CONDS:   RANGE:   COUNT:   MAX COUNT: ADDRESS:");

/////////////////////////////////////////////////////////////////////////////
//  Generate the breakpoint window's content.
void BreakpointPanel::GenerateContent(void) {
	MType cnt = g_userBreakpoints.Count();
	MType bph = m_content.Count() - 1;
	MType m, n;

	if ((m_currentBreakpoint + bph > cnt) &&
		(cnt >= bph)) {
		m_currentBreakpoint = cnt - bph;
	}
	m_content[0] = TEXT_COLOR_GREEN;
	m_content[0] += s_breakpointHeader;
	for (m=0; m<bph; ++m) {
		if (m + m_currentBreakpoint < cnt) {
			Breakpoint & b = g_userBreakpoints[m + m_currentBreakpoint];

			m_content[m + 1] = "         ";
			for (n=0; n<3; ++n) {
				m_checks[n][m].SetBP(m + m_currentBreakpoint);
				m_checks[n][m].Show();
			}
			ConvertFromHex(m_content[m + 1], b.range);
			m_content[m + 1] += " ";
			ConvertFromHex(m_content[m + 1], b.curCount);
			m_content[m + 1] += " ";
			ConvertFromHex(m_content[m + 1], b.countdown);
			m_content[m + 1] += "   ";
			g_gba.GetSymbol(m_content[m + 1], b.base);
		} else {
			m_content[m + 1].Destroy();
			for (n=0; n<3; ++n) {
				m_checks[n][m].Hide();
			}
		}
	}
	MakeDirty();
}

/////////////////////////////////////////////////////////////////////////////
//  Redraw the breakpoint window.
void BreakpointPanel::Redraw(void) {
	u16 fw = TextGetFontWidth(g_font);
	color c = g_colors[BORDER];
	Bool b = IsDirty();

	TextPanel::Redraw();
	if (b) {
		DrawLine(g_monitor,
				 c,
				 m_rect.x + 9 * fw - 2,
				 m_rect.y,
				 m_rect.x + 9 * fw - 2,
				 m_rect.y + m_rect.h);
		DrawLine(g_monitor,
				 c,
				 m_rect.x + 18 * fw - 2,
				 m_rect.y,
				 m_rect.x + 18 * fw - 2,
				 m_rect.y + m_rect.h);
		DrawLine(g_monitor,
				 c,
				 m_rect.x + 27 * fw - 2,
				 m_rect.y,
				 m_rect.x + 27 * fw - 2,
				 m_rect.y + m_rect.h);
		DrawLine(g_monitor,
				 c,
				 m_rect.x + 38 * fw - 2,
				 m_rect.y,
				 m_rect.x + 38 * fw - 2,
				 m_rect.y + m_rect.h);
		DrawLine(g_monitor,
				 c,
				 m_rect.x,
				 m_rect.y + TextGetFontHeight(g_font) - 1,
				 m_rect.x + m_rect.w,
				 m_rect.y + TextGetFontHeight(g_font) - 1);
	}
}

/////////////////////////////////////////////////////////////////////////////
//  The panel that lets us look at the sprites.
OAMPanel::OAMPanel(void) {
	m_currentSprite = 0;
}

/////////////////////////////////////////////////////////////////////////////
//  We don't really need this.
void OAMPanel::GenerateContent(void) {
}

static TString s_spriteNumber("Sprite #:  ");
static TString s_spritePos("Location:  ");

/////////////////////////////////////////////////////////////////////////////
//  Redraw the client area.
void OAMPanel::Redraw(void) {
	u16 x = m_rect.x;
	u16 y = m_rect.y;
	u16 fh = TextGetFontHeight(g_font);
	FString f;
	u32 addr = 0x7000000 + m_currentSprite * 8;
	MemoryDispatcher & m = g_gba.GetMemory();

	g_monitor.Clear(g_colors[BACKGROUND], &m_rect);

	f = s_spriteNumber;
	ConvertFromHex(f, m_currentSprite, 2);
	TextWrite(g_monitor, g_font, x, y, g_colors[TEXT], f);
	y += fh;
	f = s_spritePos;
	// HEREHEREHERE have to deal with negatives.
	ConvertFromHex(f, m.GetW(addr + 2) & 0x1ff, 3);
	f += ", ";
	// HEREHEREHERE have to deal with negatives.
	ConvertFromHex(f, m.GetW(addr) & 0xff, 3);
	TextWrite(g_monitor, g_font, x, y, g_colors[TEXT], f);

	
}

/////////////////////////////////////////////////////////////////////////////
//  Up arrow = previous sprite.
void OAMPanel::CursorUp(void) {
	++m_currentSprite;
	m_currentSprite &= 0x7f;
}

/////////////////////////////////////////////////////////////////////////////
//  Down arrow = next sprite.
void OAMPanel::CursorDown(void) {
	--m_currentSprite;
	m_currentSprite &= 0x7f;
}

/////////////////////////////////////////////////////////////////////////////
//  PgUp = go back by 16.
void OAMPanel::PageUp(void) {
	m_currentSprite += 0x10;
	m_currentSprite &= 0x7f;
}

/////////////////////////////////////////////////////////////////////////////
//  PgDn = go forward by 16.
void OAMPanel::PageDown(void) {
	m_currentSprite -= 0x10;
	m_currentSprite &= 0x7f;
}

//  Messy, but better than other alternatives.
static TString s_watches("Watches");
static TString s_memory("Memory");
static TString s_breakpoints("Breakpoints");
static TString s_oam("OAM");

/////////////////////////////////////////////////////////////////////////////
//  Build the data window.
DataWindow::DataWindow(void) {
	SetXYWH(0,
			g_monitor.GetHeight() - EMU_WINDOW_HEIGHT - STATUS_BAR_HEIGHT,
			g_monitor.GetWidth() - EMU_WINDOW_WIDTH,
			EMU_WINDOW_HEIGHT);
	#if !defined(RELEASE_BUILD)
		AddPanel(s_watches, &m_watches);
	#endif
	AddPanel(s_memory, &m_memory);
	AddPanel(s_breakpoints, &m_breakpoints);
	AddPanel(s_oam, &m_oam);
	SetTab(TAB_MEMORY);
}

/////////////////////////////////////////////////////////////////////////////
//  Activate the indicated tab.
void DataWindow::SetTab(u8 tab) {
	switch(tab) {
		#if !defined(RELEASE_BUILD)
			case TAB_WATCHES:
				m_watches.MakeDirty();
				break;
		#endif
		case TAB_MEMORY:
			m_memory.MakeDirty();
			break;
		case TAB_BREAKPOINTS:
			m_breakpoints.MakeDirty();
			break;
		case TAB_OAM:
			m_oam.MakeDirty();
			break;
	}
	Window::SetTab(tab);
}

/////////////////////////////////////////////////////////////////////////////
//  Respond to a single-step instruction.
void DataWindow::SingleStep(void) {
	m_memory.GenerateContent();
	m_breakpoints.GenerateContent();
	m_oam.GenerateContent();
}

/////////////////////////////////////////////////////////////////////////////
//  Channel this through to the breakpoint window.
void DataWindow::AddBreakpoint(u32 address, u32 range, u8 conditions) {
	m_breakpoints.AddBreakpoint(address, range, conditions);
}

/////////////////////////////////////////////////////////////////////////////
//  Respond to a user's goto request.
void DataWindow::Goto(u32 u) {
	switch(m_selection) {
		case TAB_MEMORY:
			m_memory.Goto(u);
			break;
		case TAB_BREAKPOINTS:
			m_breakpoints.Goto(u);
			break;
	}
}
